const char config_applier_box_cfg_lua[] =
"local fiber = require('fiber')\n"
"local log = require('internal.config.utils.log')\n"
"local instance_config = require('internal.config.instance_config')\n"
"local snapshot = require('internal.config.utils.snapshot')\n"
"local schedule_task = fiber._internal.schedule_task\n"
"local mkversion = require('internal.mkversion')\n"
"\n"
"local function peer_uris(configdata)\n"
"    local peers = configdata:peers()\n"
"    if #peers <= 1 then\n"
"        return nil\n"
"    end\n"
"\n"
"    local names = configdata:names()\n"
"    local err_msg_prefix = ('replication.peers construction for instance %q ' ..\n"
"        'of replicaset %q of group %q'):format(names.instance_name,\n"
"        names.replicaset_name, names.group_name)\n"
"\n"
"    -- Is there a peer in our replicaset with an URI suitable to\n"
"    -- connect (except ourself)\?\n"
"    local has_upstream = false\n"
"\n"
"    local uris = {}\n"
"    for _, peer_name in ipairs(peers) do\n"
"        local iconfig_def = configdata._peers[peer_name].iconfig_def\n"
"        local is_anon = instance_config:get(iconfig_def, 'replication.anon')\n"
"        -- Don't use anonymous replicas as upstreams.\n"
"        --\n"
"        -- An anonymous replica can't be an upstream for a\n"
"        -- non-anonymous instance.\n"
"        --\n"
"        -- An anonymous replica can be an upstream for another\n"
"        -- anonymous replica, but only non-anonymous peers are set\n"
"        -- as upstreams by default.\n"
"        --\n"
"        -- A user may configure a custom data flow using\n"
"        -- `replication.peers` option.\n"
"        if not is_anon then\n"
"            local uri = instance_config:instance_uri(iconfig_def, 'peer')\n"
"            if uri == nil then\n"
"                log.info('%s: instance %q has no iproto.advertise.peer or ' ..\n"
"                    'iproto.listen URI suitable to create a client socket',\n"
"                    err_msg_prefix, peer_name)\n"
"            end\n"
"            if uri ~= nil and peer_name ~= names.instance_name then\n"
"                has_upstream = true\n"
"            end\n"
"            table.insert(uris, uri)\n"
"        end\n"
"    end\n"
"\n"
"    if not has_upstream then\n"
"        error(('%s: no suitable peer URIs found'):format(err_msg_prefix), 0)\n"
"    end\n"
"\n"
"    return uris\n"
"end\n"
"\n"
"local function log_destination(log)\n"
"    if log.to == 'stderr' or log.to == 'devnull' then\n"
"        return box.NULL\n"
"    elseif log.to == 'file' then\n"
"        return ('file:%s'):format(log.file)\n"
"    elseif log.to == 'pipe' then\n"
"        return ('pipe:%s'):format(log.pipe)\n"
"    elseif log.to == 'syslog' then\n"
"        local res = ('syslog:identity=%s,facility=%s'):format(\n"
"            log.syslog.identity,\n"
"            log.syslog.facility)\n"
"        -- TODO: Syslog's URI format is different from tarantool's\n"
"        -- one for Unix domain sockets: `unix:/path/to/socket` vs\n"
"        -- `unix/:/path/to/socket`. We expect syslog's format\n"
"        -- here, but maybe it worth to accept our own one (or even\n"
"        -- just path) and transform under hood.\n"
"        if log.syslog.server ~= nil then\n"
"            res = res .. (',server=%s'):format(log.syslog.server)\n"
"        end\n"
"        return res\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"--------------------------------------------------------------------------------\n"
"--  Instance/replicaset names\n"
"--------------------------------------------------------------------------------\n"
"\n"
"local names_state = {\n"
"    -- Used to access confidata and drop alerts.\n"
"    config = nil,\n"
"    -- Shows, whether triggers for setting names are configured.\n"
"    is_configured = false,\n"
"    -- Waiting for rw status.\n"
"    rw_watcher = nil,\n"
"    -- Shows, whether schema upgrade trigger should be deleted.\n"
"    is_upgrade_wait = false,\n"
"}\n"
"\n"
"-- Make alerts for all names, which are missing (from snapshot or\n"
"-- from box.info) and which have not be alerted already.\n"
"local function names_alert_missing(config, missing_names)\n"
"    local msg = 'box_cfg.apply: name %s for %s uuid is missing from the ' ..\n"
"                'snapshot. It will be automatically set when possible.'\n"
"    local replicaset_name = config._configdata._replicaset_name\n"
"    if missing_names[replicaset_name] ~= nil and\n"
"       config._alerts[replicaset_name] == nil then\n"
"        local replicaset_uuid = missing_names[replicaset_name]\n"
"        local warning = msg:format(replicaset_name, replicaset_uuid)\n"
"        config:_alert(replicaset_name, {type = 'warn', message = warning})\n"
"    end\n"
"\n"
"    local unknown_msg = 'box_cfg.apply: instance %s is unknown. Possibly ' ..\n"
"                        'instance_name is not set in database and UUID is ' ..\n"
"                        'not specified. Or instance have not joined yet.'\n"
"    for name, uuid in pairs(missing_names._peers) do\n"
"        local warning\n"
"        if uuid == 'unknown' then\n"
"            warning = unknown_msg:format(name)\n"
"        else\n"
"            warning = msg:format(name, uuid)\n"
"        end\n"
"        -- Alert may be done with 'unknown' uuid. If it was so, update alert.\n"
"        local alert = config._alerts[name]\n"
"        if alert == nil or alert.message ~= warning then\n"
"            config:_alert(name, {type = 'warn', message = warning})\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Formward declaration of function, which cleans triggers.\n"
"local names_check_and_clean\n"
"\n"
"local function names_box_cfg(cfg)\n"
"    local ok, err = pcall(function()\n"
"        box.cfg(cfg)\n"
"    end)\n"
"    if not ok then\n"
"        log.warn('Failed to apply name in box.cfg: %s.', err)\n"
"    end\n"
"    return ok\n"
"end\n"
"\n"
"local function names_on_name_set(name)\n"
"    names_state.config:_alert_drop(name)\n"
"\n"
"    local config_rs_name = names_state.config._configdata._replicaset_name\n"
"    local config_inst_name = names_state.config._configdata._instance_name\n"
"    if name == config_rs_name then\n"
"        names_box_cfg({replicaset_name = name})\n"
"    elseif name == config_inst_name then\n"
"        names_box_cfg({instance_name = name})\n"
"    end\n"
"\n"
"    names_check_and_clean()\n"
"end\n"
"\n"
"local function names_schema_on_replace(old, new)\n"
"    if old == nil and new and new[1] == 'replicaset_name' then\n"
"        box.on_commit(function()\n"
"            schedule_task(names_on_name_set, new[2])\n"
"        end)\n"
"    end\n"
"end\n"
"\n"
"local function missing_names_is_empty(missing_names, replicaset_name)\n"
"    return not missing_names[replicaset_name] and\n"
"           table.equals(missing_names._peers, {})\n"
"end\n"
"\n"
"local function names_try_set_missing()\n"
"    local configdata = names_state.config._configdata\n"
"    local replicaset_name = configdata._replicaset_name\n"
"    local missing_names = configdata:missing_names()\n"
"\n"
"    if box.info.ro or missing_names_is_empty(missing_names,\n"
"                                             replicaset_name) then\n"
"        -- Somebody have done work for us, nothing to update or we're not\n"
"        -- a rw, which is possible if the function was invoked after reload.\n"
"        return\n"
"    end\n"
"\n"
"    box.begin()\n"
"    -- Set replicaset_name.\n"
"    if missing_names[replicaset_name] ~= nil then\n"
"        box.space._schema:insert{'replicaset_name', replicaset_name}\n"
"    end\n"
"\n"
"    -- Set names for all instances in the replicaset.\n"
"    for name, uuid in pairs(missing_names._peers) do\n"
"        if uuid ~= 'unknown' then\n"
"            local tuple = box.space._cluster.index.uuid:get(uuid)\n"
"            -- Tuple may be nil, when instance has not joined yet. Alert\n"
"            -- will be dropped, when instance is joined, nothing to do.\n"
"            if tuple ~= nil then\n"
"                box.space._cluster:update(tuple[1], {{'=', 3, name}})\n"
"            end\n"
"        end\n"
"    end\n"
"    box.commit()\n"
"end\n"
"\n"
"local function names_rw_watcher()\n"
"    return box.watch('box.status', function(_, status)\n"
"        -- It's ok, if names_try_set_missing will be triggered\n"
"        -- several times. It's NoOp after first execution.\n"
"        if status.is_ro == false then\n"
"            schedule_task(names_try_set_missing)\n"
"        end\n"
"    end)\n"
"end\n"
"\n"
"local function names_schema_upgrade_on_replace(old, new)\n"
"    if old == nil or new == nil then\n"
"        return\n"
"    end\n"
"\n"
"    local latest_version = mkversion.get_latest()\n"
"    local old_version = mkversion.from_tuple(old)\n"
"    local new_version = mkversion.from_tuple(new)\n"
"    if old_version < latest_version and new_version >= latest_version then\n"
"        -- We cannot do it inside on_replace trigger, as the version\n"
"        -- is not considered set yet and we may try to set names, when\n"
"        -- schema is not yet updated, which will fail as DDL is\n"
"        -- prohibited before schema upgrade.\n"
"        box.on_commit(function()\n"
"            names_state.rw_watcher = names_rw_watcher()\n"
"        end)\n"
"    end\n"
"end\n"
"\n"
"local function names_cluster_on_replace(old, new)\n"
"    local instance_name = nil\n"
"    if new ~= nil then\n"
"        instance_name = new[3]\n"
"    elseif old ~= nil then\n"
"        instance_name = names_state.config._configdata:peer_name_by_uuid(old[2])\n"
"    end\n"
"\n"
"    if instance_name ~= nil then\n"
"        box.on_commit(function()\n"
"            schedule_task(names_on_name_set, instance_name)\n"
"        end)\n"
"    end\n"
"end\n"
"\n"
"local function no_missing_names_alerts(config)\n"
"    local configdata = config._configdata\n"
"    if config._alerts[configdata._replicaset_name] ~= nil then\n"
"        return false\n"
"    end\n"
"\n"
"    local peers = configdata:peers()\n"
"    for _, name in ipairs(peers) do\n"
"        if config._alerts[name] ~= nil then\n"
"            return false\n"
"        end\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"names_check_and_clean = function()\n"
"    if no_missing_names_alerts(names_state.config) then\n"
"        names_state.is_configured = false\n"
"\n"
"        box.space._schema:on_replace(nil, names_schema_on_replace)\n"
"        box.space._cluster:on_replace(nil, names_cluster_on_replace)\n"
"\n"
"        if names_state.is_upgrade_wait then\n"
"            names_state.is_upgrade_wait = false\n"
"            box.space._schema:on_replace(nil, names_schema_upgrade_on_replace)\n"
"        end\n"
"\n"
"        if names_state.rw_watcher ~= nil then\n"
"            names_state.rw_watcher:unregister()\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function names_apply(config, missing_names, schema_version)\n"
"    -- names_state.config should be updated, as all triggers rely on configdata,\n"
"    -- saved in it. configdata may be different from the one, we already have.\n"
"    names_state.config = config\n"
"\n"
"    -- Even if everything is configured we try to make alerts one\n"
"    -- more time, as new instances without names may be found.\n"
"    names_alert_missing(config, missing_names)\n"
"    -- Don't wait for box.status to change, we may be already rw, set names\n"
"    -- on reload, if it's possible and needed.\n"
"    if schema_version and schema_version >= mkversion.get_latest() then\n"
"        names_try_set_missing()\n"
"    end\n"
"\n"
"    if names_state.is_configured then\n"
"        -- All triggers are already configured, nothing to do, but wait.\n"
"        return\n"
"    end\n"
"\n"
"    box.space._schema:on_replace(names_schema_on_replace)\n"
"    box.space._cluster:on_replace(names_cluster_on_replace)\n"
"\n"
"    -- Wait for rw state. If schema version is nil, bootstrap is\n"
"    -- going to be done.\n"
"    if schema_version and schema_version < mkversion.get_latest() then\n"
"        box.space._schema:on_replace(names_schema_upgrade_on_replace)\n"
"        names_state.is_upgrade_wait = true\n"
"    else\n"
"        names_state.rw_watcher = names_rw_watcher()\n"
"    end\n"
"\n"
"    names_state.is_configured = true\n"
"end\n"
"\n"
"local function get_schema_version_before_cfg(config)\n"
"    local snap_path = snapshot.get_path(config._configdata._iconfig_def)\n"
"    if snap_path ~= nil then\n"
"        return mkversion.from_tuple(snapshot.get_schema_version(snap_path))\n"
"    end\n"
"    -- Bootstrap, config not found\n"
"    return nil\n"
"end\n"
"\n"
"-- Returns nothing or {needs_retry = true}.\n"
"local function apply(config)\n"
"    local configdata = config._configdata\n"
"    local box_cfg = configdata:filter(function(w)\n"
"        return w.schema.box_cfg ~= nil\n"
"    end, {use_default = true}):map(function(w)\n"
"        return w.schema.box_cfg, w.data\n"
"    end):tomap()\n"
"\n"
"    -- Explicitly set box_cfg.listen to box.NULL if iproto.listen is not\n"
"    -- provided.\n"
"    -- TODO: drop this when default for array value will be supported.\n"
"    if configdata:get('iproto.listen') == nil then\n"
"        box_cfg.listen = box.NULL\n"
"    end\n"
"\n"
"    -- Construct box_cfg.replication.\n"
"    if box_cfg.replication == nil then\n"
"        box_cfg.replication = peer_uris(configdata)\n"
"    end\n"
"\n"
"    -- Construct logger destination (box_cfg.log) and log modules.\n"
"    --\n"
"    -- `log.nonblock`, `log.level`, `log.format`, 'log.modules'\n"
"    -- options are marked with the `box_cfg` annotations and so\n"
"    -- they're already added to `box_cfg`.\n"
"    local cfg_log = configdata:get('log', {use_default = true})\n"
"    box_cfg.log = log_destination(cfg_log)\n"
"\n"
"    -- Construct audit logger destination and audit filter (box_cfg.audit_log\n"
"    -- and audit_filter).\n"
"    --\n"
"    -- `audit_log.nonblock` and 'audit_log.filter' options are marked with the\n"
"    -- `box_cfg` annotations and so they're already added to `box_cfg`.\n"
"    local audit_log = configdata:get('audit_log', {use_default = true})\n"
"    if audit_log ~= nil and next(audit_log) ~= nil then\n"
"        box_cfg.audit_log = log_destination(audit_log)\n"
"        if audit_log.filter ~= nil then\n"
"            assert(type(audit_log.filter) == 'table')\n"
"            box_cfg.audit_filter = table.concat(audit_log.filter, ',')\n"
"        else\n"
"            box_cfg.audit_filter = 'compatibility'\n"
"        end\n"
"    end\n"
"\n"
"    -- The startup process may need a special handling and differs\n"
"    -- from the configuration reloading process.\n"
"    local is_startup = type(box.cfg) == 'function'\n"
"\n"
"    local failover = configdata:get('replication.failover',\n"
"        {use_default = true})\n"
"    local is_anon = configdata:get('replication.anon', {use_default = true})\n"
"\n"
"    -- Read-only or read-write\?\n"
"    if failover == 'off' then\n"
"        -- 'rw' and 'ro' mean itself.\n"
"        --\n"
"        -- The default is determined depending of amount of\n"
"        -- instances in the given replicaset.\n"
"        --\n"
"        -- * 1 instance: read-write.\n"
"        -- * >1 instances: read-only.\n"
"        --\n"
"        -- NB: configdata.lua verifies that there is at least one\n"
"        -- non-anonymous instance. So, an anonymous replica is\n"
"        -- read-only by default.\n"
"        --\n"
"        -- NB: configdata.lua also verifies that read-write mode\n"
"        -- is not enabled for an anonymous replica in the config.\n"
"        local mode = configdata:get('database.mode', {use_default = true})\n"
"        if mode == 'ro' then\n"
"            box_cfg.read_only = true\n"
"        elseif mode == 'rw' then\n"
"            box_cfg.read_only = false\n"
"        elseif #configdata:peers() == 1 then\n"
"            assert(mode == nil)\n"
"            box_cfg.read_only = false\n"
"        elseif #configdata:peers() > 1 then\n"
"            assert(mode == nil)\n"
"            box_cfg.read_only = true\n"
"        else\n"
"            assert(false)\n"
"        end\n"
"    elseif failover == 'manual' then\n"
"        -- Set RO/RW based on the 'leader' replicaset option.\n"
"        --\n"
"        -- NB: If there is no configured leader, all the instances\n"
"        -- of the given replicaset are configured as read-only.\n"
"        --\n"
"        -- NB: configdata.lua verifies that an anonymous replica\n"
"        -- is not set as a leader.\n"
"        box_cfg.read_only = not configdata:is_leader()\n"
"    elseif failover == 'election' then\n"
"        -- Enable leader election on non-anonymous instances.\n"
"        if box_cfg.election_mode == nil then\n"
"            box_cfg.election_mode = is_anon and 'off' or 'candidate'\n"
"        end\n"
"\n"
"        -- An anonymous replica can be configured with\n"
"        -- `election_mode: off`, but not other modes.\n"
"        --\n"
"        -- The validation is performed in configdata.lua for all\n"
"        -- the peers of the given replicaset.\n"
"        if is_anon then\n"
"            assert(box_cfg.election_mode == 'off')\n"
"        end\n"
"\n"
"        -- A particular instance may participate in the replicaset\n"
"        -- in different roles.\n"
"        --\n"
"        -- The default role is 'candidate' -- it votes and may be\n"
"        -- elected as a leader. The RO/RW mode is determined by\n"
"        -- the election results. box.cfg({read_only = false})\n"
"        -- means that we don't impose any extra restrictions.\n"
"        --\n"
"        -- 'voter' only votes for a new leader, but can't be\n"
"        -- elected. The actual database mode is RO (it is\n"
"        -- controlled by the underneath logic, so we pass\n"
"        -- read_only = false here).\n"
"        --\n"
"        -- An instance in the 'off' election mode neither votes,\n"
"        -- nor can be elected. It just fetches all the data.\n"
"        -- This is the only election mode, where the read-only\n"
"        -- database mode is forced by the configuration applying\n"
"        -- code.\n"
"        --\n"
"        -- An instance in the 'manual' election mode acts as a\n"
"        -- voter until box.ctl.promote() is called on it. After\n"
"        -- that it starts a new election round and acts as a\n"
"        -- candidate during this (and only this) round. No extra\n"
"        -- RO/RW restriction is performed there (controlled by the\n"
"        -- underneath logic).\n"
"        if box_cfg.election_mode == 'off' then\n"
"            box_cfg.read_only = true  -- forced RO\n"
"        elseif box_cfg.election_mode == 'voter' then\n"
"            box_cfg.read_only = false -- means no restrictions\n"
"        elseif box_cfg.election_mode == 'manual' then\n"
"            box_cfg.read_only = false -- means no restrictions\n"
"        elseif box_cfg.election_mode == 'candidate' then\n"
"            box_cfg.read_only = false -- means no restrictions\n"
"        else\n"
"            assert(false)\n"
"        end\n"
"    elseif failover == 'supervised' then\n"
"        -- The startup flow in the 'supervised' failover mode is\n"
"        -- the following.\n"
"        --\n"
"        -- * Look over the peer names of the replicaset and choose\n"
"        --   the minimal name across all non-anonymous instances\n"
"        --   (compare them lexicographically).\n"
"        -- * The instance with the minimal name starts in the RW\n"
"        --   mode to be able to bootstrap the replicaset if there\n"
"        --   is no local snapshot. Otherwise, it starts as usual,\n"
"        --   in RO.\n"
"        -- * All the other instances are started in RO.\n"
"        -- * The instance that is started in RW re-checks, whether\n"
"        --   it was a bootstrap leader (box.info.id == 1) and, if\n"
"        --   not, goes to RO.\n"
"        --\n"
"        -- The algorithm leans on the replicaset bootstrap process\n"
"        -- that chooses an RW instance as the bootstrap leader.\n"
"        -- bootstrap_strategy 'auto' fits this criteria.\n"
"        --\n"
"        -- As result, all the instances are in RO or one is in RW.\n"
"        -- The supervisor (failover agent) manages the RO/RW mode\n"
"        -- during the lifetime of the replicaset.\n"
"        --\n"
"        -- The 'minimal name' criteria is chosen with idea that\n"
"        -- instances are often names like 'storage-001',\n"
"        -- 'storage-002' and so on, so a probability that a newly\n"
"        -- added instance has the minimal name is low.\n"
"        --\n"
"        -- The RO/RW mode remains unchanged on reload.\n"
"        local am_i_bootstrap_leader = false\n"
"        if is_startup then\n"
"            local instance_name = configdata:names().instance_name\n"
"            am_i_bootstrap_leader =\n"
"                snapshot.get_path(configdata._iconfig_def) == nil and\n"
"                instance_name == configdata:bootstrap_leader_name()\n"
"            box_cfg.read_only = not am_i_bootstrap_leader\n"
"        end\n"
"\n"
"        -- It is possible that an instance with the minimal\n"
"        -- name (configdata:bootstrap_leader_name()) is added to\n"
"        -- the configuration, when the replicaset is already\n"
"        -- bootstrapped.\n"
"        --\n"
"        -- Ideally, we shouldn't make this instance RW. However,\n"
"        -- we can't differentiate this situation from one, when\n"
"        -- the replicaset bootstrap is needed, before a first\n"
"        -- box.cfg().\n"
"        --\n"
"        -- However, after leaving box.cfg() or from a background\n"
"        -- fiber after box.ctl.wait_rw() we can check that the\n"
"        -- instance was a bootstrap leader using the\n"
"        -- box.info.id == 1 condition.\n"
"        --\n"
"        -- If box.info.id != 1, then the replicaset was already\n"
"        -- bootstrapped and so we should go to RO.\n"
"        if am_i_bootstrap_leader then\n"
"            fiber.new(function()\n"
"                local name = 'config_set_read_only_if_not_bootstrap_leader'\n"
"                fiber.self():name(name, {truncate = true})\n"
"                box.ctl.wait_rw()\n"
"                if box.info.id ~= 1 then\n"
"                    -- Not really a bootstrap leader, just a\n"
"                    -- replica. Go to RO.\n"
"                    box.cfg({read_only = true})\n"
"                end\n"
"            end)\n"
"        end\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"\n"
"    assert(type(box.cfg) == 'function' or type(box.cfg) == 'table')\n"
"    if type(box.cfg) == 'table' then\n"
"        local box_cfg_nondynamic = configdata:filter(function(w)\n"
"            return w.schema.box_cfg ~= nil and w.schema.box_cfg_nondynamic\n"
"        end, {use_default = true}):map(function(w)\n"
"            return w.schema.box_cfg, w.data\n"
"        end):tomap()\n"
"        box_cfg_nondynamic.log = box_cfg.log\n"
"        box_cfg_nondynamic.audit_log = box_cfg.audit_log\n"
"        box_cfg_nondynamic.audit_filter = box_cfg.audit_filter\n"
"        for k, v in pairs(box_cfg_nondynamic) do\n"
"            if v ~= box.cfg[k] then\n"
"                local warning = 'box_cfg.apply: non-dynamic option '..k..\n"
"                    ' will not be set until the instance is restarted'\n"
"                config:_alert('box_cfg_apply_non_dynamic',\n"
"                              {type = 'warn', message = warning})\n"
"                box_cfg[k] = nil\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    -- Persist an instance name to protect a user from accidental\n"
"    -- attempt to run an instance from a snapshot left by another\n"
"    -- instance.\n"
"    --\n"
"    -- Persist a replicaset name to protect a user from attempt to\n"
"    -- mix instances with data from different replicasets into one\n"
"    -- replicaset.\n"
"    local names = configdata:names()\n"
"    box_cfg.instance_uuid = names.instance_uuid\n"
"    box_cfg.replicaset_uuid = names.replicaset_uuid\n"
"\n"
"    -- Names are applied only if they're already in snap file.\n"
"    -- Otherwise, master must apply names after box.cfg asynchronously.\n"
"    --\n"
"    -- Note: an anonymous replica has no an entry in the _cluster\n"
"    -- system space, so it can't have a persistent instance name.\n"
"    -- It is never returned by :missing_names().\n"
"    local missing_names = configdata:missing_names()\n"
"    if not is_anon and not missing_names._peers[names.instance_name] then\n"
"        box_cfg.instance_name = names.instance_name\n"
"    end\n"
"    if not missing_names[names.replicaset_name] then\n"
"        box_cfg.replicaset_name = names.replicaset_name\n"
"    end\n"
"    if not missing_names_is_empty(missing_names, names.replicaset_name) then\n"
"        if is_startup then\n"
"            local on_schema_init\n"
"            -- schema version may be nil, if bootstrap is done.\n"
"            local version = get_schema_version_before_cfg(config)\n"
"            -- on_schema init trigger isn't deleted, as it's executed only once.\n"
"            on_schema_init = box.ctl.on_schema_init(function()\n"
"                -- missing_names cannot be gathered inside on_schema_init\n"
"                -- trigger, as box.cfg is already considered configured but\n"
"                -- box.info is still not properly initialized.\n"
"                names_apply(config, missing_names, version)\n"
"                box.ctl.on_schema_init(nil, on_schema_init)\n"
"            end)\n"
"        else\n"
"            -- Note, that we try to find new missing names on every reload.\n"
"            names_apply(config, missing_names, mkversion.get())\n"
"        end\n"
"    end\n"
"\n"
"    -- Set bootstrap_leader option.\n"
"    box_cfg.bootstrap_leader = configdata:bootstrap_leader()\n"
"\n"
"    -- Set metrics option.\n"
"    local include = configdata:get('metrics.include', {use_default = true})\n"
"    local exclude = configdata:get('metrics.exclude', {use_default = true})\n"
"    local labels = configdata:get('metrics.labels', {use_default = true})\n"
"\n"
"    box_cfg.metrics = {\n"
"        include = include or { 'all' },\n"
"        exclude = exclude or { },\n"
"        labels = labels or { alias = names.instance_name },\n"
"    }\n"
"\n"
"    -- First box.cfg() call.\n"
"    --\n"
"    -- Force the read-only mode if:\n"
"    --\n"
"    -- * the startup may take a long time, and\n"
"    -- * the instance is not the only one in its replicaset, and\n"
"    -- * there is an existing snapshot (otherwise we wouldn't able\n"
"    --   to assign a bootstrap leader).\n"
"    --\n"
"    -- The reason is that the configured master may be switched\n"
"    -- while it is starting. In this case it is undesirable to set\n"
"    -- RW mode if the actual configuration marks the instance as\n"
"    -- RO.\n"
"    --\n"
"    -- The main startup code calls this applier second time if\n"
"    -- {needs_retry = true} is returned after re-reading of the\n"
"    -- configuration.\n"
"    --\n"
"    -- Automatic reapplying of the post-startup configuration is\n"
"    -- performed for all the cases, when the startup may take a\n"
"    -- long time.\n"
"    --\n"
"    -- This logic is disabled if the automatic leader election is\n"
"    -- in effect. The configuration has no leadership information,\n"
"    -- in this case, so there is no much sense to re-read it after\n"
"    -- startup.\n"
"    --\n"
"    -- The same for the supervised failover mode.\n"
"    if is_startup and failover ~= 'election' and failover ~= 'supervised' then\n"
"        local configured_as_rw = not box_cfg.read_only\n"
"        local in_replicaset = #configdata:peers() > 1\n"
"        local has_snap = snapshot.get_path(configdata._iconfig_def) ~= nil\n"
"\n"
"        -- Require at least one writable instance in the\n"
"        -- replicaset if the instance is to be bootstrapped (has\n"
"        -- no existing snapshot). Otherwise there is no one who\n"
"        -- would register the instance in the _cluster system\n"
"        -- space.\n"
"        --\n"
"        -- TODO: If an instance is configured with known\n"
"        -- instance_uuid that is already in _cluster, it may be\n"
"        -- re-bootstrapped. But we don't know from the\n"
"        -- configuration whether the given instance_uuid is\n"
"        -- already registered. We should decide what to do in the\n"
"        -- case.\n"
"        --\n"
"        -- Note: an anonymous replica is not to be registered in\n"
"        -- _cluster. So, it can subscribe to such a replicaset.\n"
"        if in_replicaset and not has_snap and not is_anon then\n"
"            local has_rw = false\n"
"            if failover == 'off' then\n"
"                for _, peer_name in ipairs(configdata:peers()) do\n"
"                    local opts = {peer = peer_name, use_default = true}\n"
"                    local mode = configdata:get('database.mode', opts)\n"
"                    -- NB: The default is box.NULL that is\n"
"                    -- interpreted as 'ro' for a replicaset with\n"
"                    -- more than one instance.\n"
"                    if mode == 'rw' then\n"
"                        has_rw = true\n"
"                        break\n"
"                    end\n"
"                end\n"
"            elseif failover == 'manual' then\n"
"                has_rw = configdata:leader() ~= nil\n"
"            else\n"
"                assert(false)\n"
"            end\n"
"            if not has_rw then\n"
"                error(('Startup failure.\\nNo leader to register new ' ..\n"
"                    'instance %q. All the instances in replicaset %q of ' ..\n"
"                    'group %q are configured to the read-only mode.'):format(\n"
"                    names.instance_name, names.replicaset_name,\n"
"                    names.group_name), 0)\n"
"            end\n"
"        end\n"
"\n"
"        -- Reading a snapshot may take a long time, fetching it\n"
"        -- from a remote master is even longer.\n"
"        local startup_may_be_long = has_snap or\n"
"            (not has_snap and in_replicaset)\n"
"\n"
"        -- Start to construct the force read-only mode condition.\n"
"        local force_read_only = configured_as_rw\n"
"\n"
"        -- It only has sense to force RO if we expect that the\n"
"        -- configured RO/RW mode has a good chance to change\n"
"        -- during the startup.\n"
"        force_read_only = force_read_only and startup_may_be_long\n"
"\n"
"        -- If the instance is the only one in its replicaset,\n"
"        -- there is no another one that would take the leadership.\n"
"        --\n"
"        -- It is OK to just start in RW.\n"
"        force_read_only = force_read_only and in_replicaset\n"
"\n"
"        -- Can't set RO for an instance in a replicaset without an\n"
"        -- existing snapshot, because one of the cases is when the\n"
"        -- replicaset is to be created from scratch. There should\n"
"        -- be at least one RW instance to act as a bootstrap\n"
"        -- leader.\n"
"        --\n"
"        -- We don't know, whether other instances have snapshots,\n"
"        -- so the best that we can do is to lean on the user\n"
"        -- provided configuration regarding RO/RW.\n"
"        force_read_only = force_read_only and has_snap\n"
"\n"
"        if force_read_only then\n"
"            box_cfg.read_only = true\n"
"        end\n"
"\n"
"        log.debug('box_cfg.apply (startup)')\n"
"        box.cfg(box_cfg)\n"
"\n"
"        -- NB: needs_retry should be true when force_read_only is\n"
"        -- true. It is so by construction.\n"
"        return {needs_retry = startup_may_be_long}\n"
"    end\n"
"\n"
"    -- If it is reload, just apply the new configuration.\n"
"    log.debug('box_cfg.apply')\n"
"    box.cfg(box_cfg)\n"
"end\n"
"\n"
"return {\n"
"    name = 'box_cfg',\n"
"    apply = apply,\n"
"}\n"
""
;
