const char config_applier_credentials_lua[] =
"local expression = require('internal.config.utils.expression')\n"
"local log = require('internal.config.utils.log')\n"
"local digest = require('digest')\n"
"local fiber = require('fiber')\n"
"\n"
"-- Var is set with the first apply() call.\n"
"local config\n"
"\n"
"-- {{{ Sync helpers\n"
"\n"
"--[[\n"
"This intermediate representation is a formatted set of\n"
"all permissions for a user or role. It is required to\n"
"standardize diff functions. All the validation is done\n"
"by config or box.info(), so neither the format nor the\n"
"helper function don't perform it. Below you can find two\n"
"converters to this representation, from box format and\n"
"config schema format, accordingly `privileges_{box,config}_convert()`.\n"
"\n"
"[obj_type][obj_name] = {\n"
"    read = true,\n"
"    write = true,\n"
"    ...\n"
"}\n"
"\n"
"obj_types:\n"
" - 'user'\n"
" - 'role'\n"
" - 'space'\n"
" - 'function'\n"
" - 'sequence'\n"
" - 'universe'\n"
" - 'lua_eval'\n"
" - 'lua_call'\n"
" - 'sql'\n"
"\n"
"Note that 'lua_eval', 'lua_call', 'sql' and 'universe' are special,\n"
"they don't allow obj_name specialisation.\n"
"\n"
"obj_names:\n"
" - mostly user defined strings, provided by config or box\n"
" - special value '', when there is no obj_name, e.g. for\n"
"   'universe' obj_type or for granting permission for all\n"
"   objects of a type.\n"
"\n"
"privs:\n"
" - read\n"
" - write\n"
" - execute\n"
" - session\n"
" - usage\n"
" - create\n"
" - drop\n"
" - alter\n"
" - reference\n"
" - trigger\n"
" - insert\n"
" - update\n"
" - delete\n"
"\n"
"Examples:\n"
"- - box.schema.user.grant('myuser', 'execute', 'function', 'myfunc')\n"
"  - ['function']['myfunc']['execute'] = true\n"
"  - grant execute of myfunc\n"
"\n"
"- - box.schema.user.grant('myuser', 'execute', 'function')\n"
"  - ['function']['']['execute'] = true\n"
"  - grant execute of all registered functions\n"
"\n"
"- - box.schema.user.grant('myuser', 'read', 'universe')\n"
"  - ['universe']['']['read'] = true\n"
"  - grant read to universe\n"
"\n"
"- - box.schema.user.grant('myuser', 'execute', 'role', 'super')\n"
"  - ['role']['super']['execute'] = true\n"
"  - equivalent to granting a role to myuser\n"
"\n"
"]]--\n"
"\n"
"local function privileges_from_box(privileges)\n"
"    privileges = privileges or {}\n"
"    assert(type(privileges) == 'table')\n"
"\n"
"    local res = {\n"
"        ['user'] = {},\n"
"        ['role'] = {},\n"
"        ['space'] = {},\n"
"        ['function'] = {},\n"
"        ['sequence'] = {},\n"
"        ['universe'] = {},\n"
"        ['lua_eval'] = {},\n"
"        ['lua_call'] = {},\n"
"        ['sql'] = {},\n"
"    }\n"
"\n"
"    for _, priv in ipairs(privileges) do\n"
"        local perms, obj_type, obj_name = unpack(priv)\n"
"        obj_name = obj_name or ''\n"
"\n"
"        res[obj_type][obj_name] = res[obj_type][obj_name] or {}\n"
"\n"
"        for _, perm in ipairs(perms:split(',')) do\n"
"            res[obj_type][obj_name][perm] = true\n"
"        end\n"
"    end\n"
"\n"
"    return res\n"
"end\n"
"\n"
"-- Note that 'all' is considered a special value, meaning all objects of\n"
"-- obj_type will be granted this permission. Don't use this function if it\n"
"-- may occur in any other meaning, e.g. user defined name.\n"
"--\n"
"-- Note: `obj_names` can be either an array with objects names or a string\n"
"--       with a single one. It could also be `nil`, meaning \"do nothing\".\n"
"local function privileges_add_perm(obj_type, obj_names, perm, intermediate)\n"
"    if obj_names == nil then\n"
"        return\n"
"    end\n"
"    if type(obj_names) == 'string' then\n"
"        obj_names = {obj_names}\n"
"    end\n"
"\n"
"    for _, obj_name  in ipairs(obj_names) do\n"
"        if obj_name == 'all' then\n"
"            -- '' is a special value, meaning all objects of this obj_type.\n"
"            obj_name = ''\n"
"        end\n"
"        intermediate[obj_type][obj_name] =\n"
"            intermediate[obj_type][obj_name] or {}\n"
"        intermediate[obj_type][obj_name][perm] = true\n"
"    end\n"
"end\n"
"\n"
"local function privileges_from_config(config_data)\n"
"    local privileges = config_data.privileges or {}\n"
"    assert(type(privileges) == 'table')\n"
"\n"
"    local intermediate = {\n"
"        ['user'] = {},\n"
"        ['role'] = {},\n"
"        ['space'] = {},\n"
"        ['function'] = {},\n"
"        ['sequence'] = {},\n"
"        ['universe'] = {},\n"
"        ['lua_eval'] = {},\n"
"        ['lua_call'] = {},\n"
"        ['sql'] = {},\n"
"    }\n"
"\n"
"    for _, priv in ipairs(privileges) do\n"
"        for _, perm in ipairs(priv.permissions) do\n"
"            if priv.universe then\n"
"                privileges_add_perm('universe', 'all', perm, intermediate)\n"
"            end\n"
"            if priv.lua_eval then\n"
"                privileges_add_perm('lua_eval', 'all', perm, intermediate)\n"
"            end\n"
"            privileges_add_perm('space', priv.spaces, perm, intermediate)\n"
"            privileges_add_perm('function', priv.functions, perm, intermediate)\n"
"            privileges_add_perm('sequence', priv.sequences, perm, intermediate)\n"
"            privileges_add_perm('lua_call', priv.lua_call, perm, intermediate)\n"
"            privileges_add_perm('sql', priv.sql, perm, intermediate)\n"
"        end\n"
"    end\n"
"\n"
"    local roles = config_data.roles or {}\n"
"\n"
"    for _, role_name in ipairs(roles) do\n"
"        -- Unlike spaces, functions and sequences, role is allowed to be\n"
"        -- named 'all', so `privileges_add_perm()` isn't used.\n"
"        intermediate['role'][role_name] = intermediate['role'][role_name] or {}\n"
"        intermediate['role'][role_name]['execute'] = true\n"
"    end\n"
"\n"
"    return intermediate\n"
"end\n"
"\n"
"-- Intermediate representation is basically a set, so this function subtracts\n"
"-- `current` from `target`.\n"
"-- Return privileges that are present in `target` but not in `current` as\n"
"-- a list in the following format:\n"
"-- - obj_type: my_type\n"
"--   obj_name: my_name\n"
"--   privs:\n"
"--    - read\n"
"--    - write\n"
"--    - ...\n"
"--\n"
"local function privileges_subtract(target, current)\n"
"    local lacking = {}\n"
"\n"
"    for obj_type, privileges_group in pairs(target) do\n"
"        for obj_name, privileges in pairs(privileges_group) do\n"
"            local lacking_privs = {}\n"
"            for priv, target_val in pairs(privileges) do\n"
"                if target_val and (current[obj_type][obj_name] == nil or\n"
"                                   not current[obj_type][obj_name][priv]) then\n"
"                    table.insert(lacking_privs, priv)\n"
"                end\n"
"            end\n"
"            if next(lacking_privs) then\n"
"                table.insert(lacking, {\n"
"                    obj_type = obj_type,\n"
"                    obj_name = obj_name,\n"
"                    privs = lacking_privs,\n"
"                })\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    return lacking\n"
"end\n"
"\n"
"local function privileges_add_defaults(name, role_or_user, intermediate)\n"
"    local res = table.deepcopy(intermediate)\n"
"\n"
"    if role_or_user == 'user' then\n"
"        if name == 'guest' then\n"
"            privileges_add_perm('role', 'public', 'execute', res)\n"
"\n"
"            privileges_add_perm('universe', '', 'session', res)\n"
"            privileges_add_perm('universe', '', 'usage', res)\n"
"\n"
"        elseif name == 'admin' then\n"
"            privileges_add_perm('universe', '', 'read', res)\n"
"            privileges_add_perm('universe', '', 'write', res)\n"
"            privileges_add_perm('universe', '', 'execute', res)\n"
"            privileges_add_perm('universe', '', 'session', res)\n"
"            privileges_add_perm('universe', '', 'usage', res)\n"
"            privileges_add_perm('universe', '', 'create', res)\n"
"            privileges_add_perm('universe', '', 'drop', res)\n"
"            privileges_add_perm('universe', '', 'alter', res)\n"
"            privileges_add_perm('universe', '', 'reference', res)\n"
"            privileges_add_perm('universe', '', 'trigger', res)\n"
"            privileges_add_perm('universe', '', 'insert', res)\n"
"            privileges_add_perm('universe', '', 'update', res)\n"
"            privileges_add_perm('universe', '', 'delete', res)\n"
"\n"
"        else\n"
"            -- Newly created user:\n"
"            privileges_add_perm('role', 'public', 'execute', res)\n"
"\n"
"            privileges_add_perm('universe', '', 'session', res)\n"
"            privileges_add_perm('universe', '', 'usage', res)\n"
"\n"
"            privileges_add_perm('user', name, 'alter', res)\n"
"        end\n"
"\n"
"    elseif role_or_user == 'role' then\n"
"        -- luacheck: ignore 542 empty if branch\n"
"        if name == 'public' then\n"
"            privileges_add_perm('function', 'box.schema.user.info', 'execute',\n"
"                                res)\n"
"            privileges_add_perm('function', 'LUA', 'read', res)\n"
"\n"
"            privileges_add_perm('space', '_vcollation', 'read', res)\n"
"            privileges_add_perm('space', '_vspace', 'read', res)\n"
"            privileges_add_perm('space', '_vsequence', 'read', res)\n"
"            privileges_add_perm('space', '_vindex', 'read', res)\n"
"            privileges_add_perm('space', '_vfunc', 'read', res)\n"
"            privileges_add_perm('space', '_vuser', 'read', res)\n"
"            privileges_add_perm('space', '_vpriv', 'read', res)\n"
"            privileges_add_perm('space', '_vspace_sequence', 'read', res)\n"
"\n"
"            privileges_add_perm('space', '_truncate', 'write', res)\n"
"\n"
"            privileges_add_perm('space', '_session_settings', 'read', res)\n"
"            privileges_add_perm('space', '_session_settings', 'write', res)\n"
"\n"
"        elseif name == 'replication' then\n"
"            privileges_add_perm('space', '_cluster', 'write', res)\n"
"            privileges_add_perm('universe', '', 'read', res)\n"
"\n"
"        elseif name == 'super' then\n"
"            privileges_add_perm('universe', '', 'read', res)\n"
"            privileges_add_perm('universe', '', 'write', res)\n"
"            privileges_add_perm('universe', '', 'execute', res)\n"
"            privileges_add_perm('universe', '', 'session', res)\n"
"            privileges_add_perm('universe', '', 'usage', res)\n"
"            privileges_add_perm('universe', '', 'create', res)\n"
"            privileges_add_perm('universe', '', 'drop', res)\n"
"            privileges_add_perm('universe', '', 'alter', res)\n"
"            privileges_add_perm('universe', '', 'reference', res)\n"
"            privileges_add_perm('universe', '', 'trigger', res)\n"
"            privileges_add_perm('universe', '', 'insert', res)\n"
"            privileges_add_perm('universe', '', 'update', res)\n"
"            privileges_add_perm('universe', '', 'delete', res)\n"
"\n"
"        else\n"
"            -- Newly created role has NO permissions.\n"
"        end\n"
"    else\n"
"        assert(false, 'neither role nor user provided')\n"
"    end\n"
"\n"
"    return res\n"
"end\n"
"\n"
"--\n"
"-- Create a credential sharding role if it is assigned to any user or role and\n"
"-- has not been declared in 'credentials.roles'.\n"
"--\n"
"-- Replicasets with the sharding storage role must be assigned the credentials\n"
"-- sharding role. In these replicasets, the role will be assigned all necessary\n"
"-- privileges.\n"
"--\n"
"-- For replicasets that do not have the sharding storage role assigned, the\n"
"-- credentials sharding role will not have any privileges.\n"
"--\n"
"-- If the snapshot version does not match the Tarantool schema version, DDL\n"
"-- operations, including creating roles and granting privileges, are prohibited.\n"
"-- For user-created changes, this will throw an error as expected, but since the\n"
"-- \"sharding\" role is created by default, the error may still appear even if the\n"
"-- user does not make any changes. To partially avoid this problem, we do not\n"
"-- create the \"sharding\" role if it is not in use. This is a temporary solution.\n"
"local function sharding_role(configdata)\n"
"    local roles = configdata:get('credentials.roles')\n"
"    local users = configdata:get('credentials.users')\n"
"    local has_sharding_role = false\n"
"    for _, role in pairs(roles or {}) do\n"
"        for _, role_role in pairs(role.roles or {}) do\n"
"            has_sharding_role = has_sharding_role or role_role == 'sharding'\n"
"        end\n"
"    end\n"
"    for _, user in pairs(users or {}) do\n"
"        for _, user_role in pairs(user.roles or {}) do\n"
"            has_sharding_role = has_sharding_role or user_role == 'sharding'\n"
"        end\n"
"    end\n"
"    if not has_sharding_role then\n"
"        return\n"
"    end\n"
"\n"
"    -- Add necessary privileges if storage sharding role is enabled.\n"
"    local sharding_roles = configdata:get('sharding.roles')\n"
"    if sharding_roles == nil or #sharding_roles == 0 then\n"
"        return {}\n"
"    end\n"
"    local is_storage = false\n"
"    for _, role in pairs(sharding_roles) do\n"
"        is_storage = is_storage or role == 'storage'\n"
"    end\n"
"    if not is_storage then\n"
"        return {}\n"
"    end\n"
"\n"
"    local funcs = {}\n"
"    --\n"
"    -- The error will be thrown later, in sharding.lua. Here we are simply\n"
"    -- trying to avoid the \"module not found\" error.\n"
"    --\n"
"    local ok, vshard = pcall(require, 'vshard')\n"
"    if ok and expression.eval('v >= 0.1.25', {v = vshard.consts.VERSION}) then\n"
"        local vexports = require('vshard.storage.exports')\n"
"        local exports = vexports.compile(vexports.log[#vexports.log])\n"
"        for name in pairs(exports.funcs) do\n"
"            table.insert(funcs, name)\n"
"        end\n"
"    end\n"
"    return {\n"
"        privileges = {{\n"
"            permissions = {'execute'},\n"
"            functions = funcs,\n"
"        }},\n"
"        roles = {'replication'},\n"
"    }\n"
"end\n"
"\n"
"-- Get the latest credentials configuration from the config and add empty\n"
"-- default users and roles configuration, if they are missing.\n"
"local function get_credentials(config)\n"
"    local configdata = config._configdata\n"
"    local credentials = configdata:get('credentials')\n"
"\n"
"    -- If credentials section in config is empty, skip applier.\n"
"    if credentials == nil then\n"
"        return {}\n"
"    end\n"
"\n"
"    -- Tarantool has the following roles and users present by default on every\n"
"    -- instance:\n"
"    --\n"
"    -- Default roles:\n"
"    -- * super\n"
"    -- * public\n"
"    -- * replication\n"
"    --\n"
"    -- Default users:\n"
"    -- * guest\n"
"    -- * admin\n"
"    --\n"
"    -- These roles and users have according privileges pre-granted by design.\n"
"    -- Credentials applier adds such privileges with `priviliges_add_default()`\n"
"    -- when syncing. So, for the excessive (non-default) privs to be removed,\n"
"    -- these roles and users must be present inside configuration at least in a\n"
"    -- form of an empty table. Otherwise, the privileges will be left unchanged,\n"
"    -- similar to all used-defined roles and users.\n"
"\n"
"    credentials.roles = credentials.roles or {}\n"
"    credentials.roles['super'] = credentials.roles['super'] or {}\n"
"    credentials.roles['public'] = credentials.roles['public'] or {}\n"
"    credentials.roles['replication'] = credentials.roles['replication'] or {}\n"
"\n"
"    -- Add a semi-default role 'sharding'.\n"
"    credentials.roles['sharding'] = credentials.roles['sharding'] or\n"
"                                    sharding_role(configdata)\n"
"\n"
"    credentials.users = credentials.users or {}\n"
"    credentials.users['guest'] = credentials.users['guest'] or {}\n"
"    credentials.users['admin'] = credentials.users['admin'] or {}\n"
"\n"
"    return credentials\n"
"end\n"
"\n"
"-- }}} Sync helpers\n"
"\n"
"-- {{{ Triggers for grants after obj creation\n"
"\n"
"-- This is fiber channel that contains all scheduled tasks\n"
"-- in form of a table with the following format:\n"
"-- {type = 'BLOCKING_FULL_SYNC'/'BACKGROUND_FULL_SYNC'/obj_type,\n"
"--  name = obj_name}\n"
"local sync_tasks\n"
"\n"
"-- This is a map of all objects that are mentioned in the config.\n"
"-- It is always in sync with the latest reload and has the following format:\n"
"-- [obj_type][obj_name] = true/nil\n"
"local target_object_map = {\n"
"    ['space'] = {},\n"
"    ['function'] = {},\n"
"    ['sequence'] = {},\n"
"}\n"
"\n"
"-- Iterate through requests and sync when the according flag is set.\n"
"local function on_commit_trigger(iterator)\n"
"    for _, old_obj, new_obj, space_id in iterator() do\n"
"\n"
"        local obj_type\n"
"        if space_id == box.schema.SPACE_ID then\n"
"            obj_type = 'space'\n"
"        elseif space_id == box.schema.FUNC_ID then\n"
"            obj_type = 'function'\n"
"        elseif space_id == box.schema.SEQUENCE_ID then\n"
"            obj_type = 'sequence'\n"
"        else\n"
"            goto skip\n"
"        end\n"
"\n"
"        if new_obj == nil then\n"
"            goto skip\n"
"        end\n"
"\n"
"        -- Check if the request is for obj creation or obj rename.\n"
"        if old_obj == nil or old_obj.name ~= new_obj.name then\n"
"            local obj_name = new_obj.name\n"
"\n"
"            -- Check that the object is present inside config.\n"
"            if target_object_map[obj_type][obj_name] then\n"
"                if not sync_tasks:is_full() then\n"
"                    sync_tasks:put({name = obj_name, type = obj_type})\n"
"                else\n"
"                    -- In some rare cases the channel can fill up.\n"
"                    -- It means that a great number of objects requires\n"
"                    -- sync. Thus perform a full synchronization.\n"
"                    while not sync_tasks:is_empty() do\n"
"                        sync_tasks:get()\n"
"                    end\n"
"                    sync_tasks:put({type = 'BACKGROUND_FULL_SYNC'})\n"
"                end\n"
"            end\n"
"        end\n"
"\n"
"        ::skip::\n"
"    end\n"
"end\n"
"\n"
"-- Check that request is space/function/sequence creation or rename\n"
"-- and set on_commit_trigger for the transaction.\n"
"local function on_replace_trigger(old_obj, new_obj, obj_type, request_type)\n"
"    if box.session.type() == 'applier' then\n"
"        -- The request is caused by replication, so the instance is in\n"
"        -- RO mode. Thus, skip all applier actions.\n"
"        return\n"
"    end\n"
"\n"
"    if obj_type == '_space' then\n"
"        obj_type = 'space'\n"
"    elseif obj_type == '_func' then\n"
"        obj_type = 'function'\n"
"    elseif obj_type == '_sequence' then\n"
"        obj_type = 'sequence'\n"
"    else\n"
"        return\n"
"    end\n"
"\n"
"    -- Check if the request is object creation or rename\n"
"    if new_obj == nil then\n"
"        return\n"
"    end\n"
"\n"
"    if not ((request_type == 'INSERT' and old_obj == nil) or\n"
"            (request_type == 'UPDATE' and old_obj ~= nil and\n"
"                old_obj.name ~= new_obj.name)) then\n"
"        return\n"
"    end\n"
"\n"
"    local obj_name = new_obj.name\n"
"    -- Check that the object is present inside config.\n"
"    if not target_object_map[obj_type][obj_name] then\n"
"        return\n"
"    end\n"
"\n"
"    -- Set the on_commit trigger to apply the action. Action could not\n"
"    -- be applied at this point because the object may not be fully\n"
"    -- registered.\n"
"    box.on_commit(on_commit_trigger)\n"
"end\n"
"\n"
"-- }}} Triggers for grants after obj creation\n"
"\n"
"-- {{{ Main sync logic\n"
"\n"
"-- Grant or revoke some user/role privileges for an object.\n"
"local privileges_action_f = function(grant_or_revoke, role_or_user, name, privs,\n"
"                                     obj_type, obj_name)\n"
"    assert(grant_or_revoke == 'grant' or grant_or_revoke == 'revoke')\n"
"    privs = table.concat(privs, ',')\n"
"\n"
"    -- Try to apply the action immediately. If the object doesn't exist,\n"
"    -- the sync will be applied inside the trigger on object creation/rename.\n"
"    local ok, err = pcall(box.schema[role_or_user][grant_or_revoke],\n"
"                          name, privs, obj_type, obj_name)\n"
"\n"
"    if ok then\n"
"        log.verbose('credentials.apply: box.schema.%s.%s(%q, %q, %q, %q)',\n"
"                    role_or_user, grant_or_revoke, name, privs,\n"
"                    obj_type, obj_name)\n"
"        return\n"
"    end\n"
"    if err.code ~= box.error.NO_SUCH_SPACE and\n"
"            err.code ~= box.error.NO_SUCH_FUNCTION and\n"
"            err.code ~= box.error.NO_SUCH_SEQUENCE then\n"
"        err = ('credentials.apply: box.schema.%s.%s(%q, %q, %q, %q) failed: %s')\n"
"              :format(role_or_user, grant_or_revoke, name, privs, obj_type,\n"
"                      obj_name, err)\n"
"        config:_alert('credentials_grant_or_revoke_err',\n"
"                      {type = 'error', message = err})\n"
"    else\n"
"        local msg = \"credentials.apply: %s %q hasn't been created yet, \" ..\n"
"                    \"'box.schema.%s.%s(%q, %q, %q, %q)' will be applied later\"\n"
"        msg = msg:format(obj_type, obj_name, role_or_user, grant_or_revoke,\n"
"                         name, privs, obj_type, obj_name)\n"
"        config:_alert('credentials_grant_or_revoke_warn',\n"
"                      {type = 'warn', message = msg})\n"
"    end\n"
"end\n"
"\n"
"-- Perform a synchronization of privileges for all users/roles.\n"
"-- obj_to_sync is an optional parameter, if it is non-nil,\n"
"-- the sync is being performed only for the provided object.\n"
"local function sync_privileges(credentials, obj_to_sync)\n"
"    assert(obj_to_sync == nil or type(obj_to_sync) == 'table')\n"
"\n"
"    if obj_to_sync then\n"
"        log.verbose('credentials.apply: syncing privileges for %s %q',\n"
"                    obj_to_sync.type, obj_to_sync.name)\n"
"    end\n"
"\n"
"    -- The privileges synchronization between A and B is performed in 3 steps:\n"
"    -- 1. Grant all privileges that are present in B,\n"
"    --    but not present in A (`grant(B - A)`).\n"
"    -- 2. Add default privileges to B (`B = B + defaults`).\n"
"    -- 3. Revoke all privileges that are not present in B,\n"
"    --    but present in A (`revoke(A - B)).\n"
"    --\n"
"    -- Default privileges are not granted on step 1, so they stay revoked if\n"
"    -- revoked manually (e.g. with `box.schema.{user,role}.revoke()`).\n"
"    -- However, defaults should never be revoked, so target state B is enriched\n"
"    -- with them before step 3.\n"
"    local function sync(role_or_user, name, config_privileges)\n"
"        assert(role_or_user == 'user' or role_or_user == 'role')\n"
"        if not obj_to_sync then\n"
"            log.verbose('credentials.apply: syncing privileges for %s %q',\n"
"                        role_or_user, name)\n"
"        end\n"
"\n"
"        local box_privileges = box.schema[role_or_user].info(name)\n"
"\n"
"        config_privileges = privileges_from_config(config_privileges)\n"
"        box_privileges = privileges_from_box(box_privileges)\n"
"\n"
"        local grants = privileges_subtract(config_privileges, box_privileges)\n"
"\n"
"        for _, to_grant in ipairs(grants) do\n"
"            -- Note that grants are filtered per object, if required.\n"
"            if obj_to_sync == nil or (obj_to_sync.type == to_grant.obj_type and\n"
"                                     obj_to_sync.name == to_grant.obj_name) then\n"
"                privileges_action_f('grant', role_or_user, name, to_grant.privs,\n"
"                                    to_grant.obj_type, to_grant.obj_name)\n"
"            end\n"
"        end\n"
"\n"
"        config_privileges = privileges_add_defaults(name, role_or_user,\n"
"                                                    config_privileges)\n"
"\n"
"        local revokes = privileges_subtract(box_privileges, config_privileges)\n"
"\n"
"        for _, to_revoke in ipairs(revokes) do\n"
"            -- Note that revokes are filtered per object, if required.\n"
"            if obj_to_sync == nil or (obj_to_sync.type == to_revoke.obj_type and\n"
"                                    obj_to_sync.name == to_revoke.obj_name) then\n"
"                privileges_action_f('revoke', role_or_user, name,\n"
"                                    to_revoke.privs, to_revoke.obj_type,\n"
"                                    to_revoke.obj_name)\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    -- Note that the logic inside the sync() function, i.e. privileges dump\n"
"    -- from box, diff calculation and grants/revokes should happen inside one\n"
"    -- transaction.\n"
"\n"
"    for name, user_def in pairs(credentials.users or {}) do\n"
"        sync('user', name, user_def)\n"
"    end\n"
"\n"
"    for name, role_def in pairs(credentials.roles or {}) do\n"
"        sync('role', name, role_def)\n"
"    end\n"
"end\n"
"\n"
"-- }}} Main sync logic\n"
"\n"
"-- {{{ Create roles\n"
"\n"
"local function create_role(role_name)\n"
"    if box.schema.role.exists(role_name) then\n"
"        log.verbose('credentials.apply: role %q already exists', role_name)\n"
"    else\n"
"        log.verbose('credentials.apply: create role %q', role_name)\n"
"        box.schema.role.create(role_name)\n"
"    end\n"
"end\n"
"\n"
"local function create_roles(role_map)\n"
"    for role_name, role_def in pairs(role_map or {}) do\n"
"        if role_def ~= nil then\n"
"            create_role(role_name)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- }}} Create roles\n"
"\n"
"-- {{{ Create users\n"
"\n"
"local function create_user(user_name)\n"
"    if box.schema.user.exists(user_name) then\n"
"        log.verbose('credentials.apply: user %q already exists', user_name)\n"
"    else\n"
"        log.verbose('credentials.apply: create user %q', user_name)\n"
"        box.schema.user.create(user_name)\n"
"    end\n"
"end\n"
"\n"
"local function set_password(user_name, password)\n"
"    if password == nil then\n"
"        if user_name == 'guest' then\n"
"            -- Guest can't have a password, so this is valid and there\n"
"            -- is nothing to do.\n"
"            return\n"
"        end\n"
"\n"
"        local auth_def = box.space._user.index.name:get({user_name})[5]\n"
"        if next(auth_def) == nil then\n"
"            -- No password is currently set, there is nothing to do.\n"
"            log.verbose('credentials.apply: user %q already has no password',\n"
"                        user_name)\n"
"            return\n"
"        end\n"
"\n"
"        log.verbose('credentials.apply: remove password for user %q', user_name)\n"
"        -- No password is set, so remove it for the user.\n"
"        -- There is no handy function for it, so remove it directly in\n"
"        -- system space _user. Users are described in the following way:\n"
"        --\n"
"        -- - [32, 1, 'myusername', 'user', {'chap-sha1': '<password_hash>'},\n"
"        --          [], 1692279984]\n"
"        --\n"
"        -- The command below overrides with empty map the fifth tuple field,\n"
"        -- containing hash (with auth_type == 'chap-sha1' and hash with salt\n"
"        -- with auth_type == 'pap-sha256').\n"
"        box.space._user.index.name:update(user_name,\n"
"            {{'=', 5, setmetatable({}, {__serialize = 'map'})}})\n"
"\n"
"        return\n"
"    end\n"
"\n"
"    if user_name == 'guest' then\n"
"        local message = 'credentials.apply: setting a password for ' ..\n"
"                        'the guest user is not allowed'\n"
"        config:_alert('credentials_apply_password',\n"
"                      {type = 'error', message = message})\n"
"    end\n"
"\n"
"    local auth_def = box.space._user.index.name:get({user_name})[5]\n"
"    if next(auth_def) == nil then\n"
"        -- No password is currently set for the user, just set a new one.\n"
"        log.verbose('credentials.apply: set a password for user %q', user_name)\n"
"        box.schema.user.passwd(user_name, password)\n"
"        return\n"
"    end\n"
"\n"
"    local auth_type = auth_def['chap-sha1'] and 'chap-sha1' or 'pap-sha256'\n"
"\n"
"    local new_password = false\n"
"\n"
"    if auth_type == 'chap-sha1' then\n"
"        local current_hash = auth_def['chap-sha1']\n"
"\n"
"        local new_hash = box.schema.user.password(password)\n"
"        if new_hash ~= current_hash then\n"
"            new_password = true\n"
"        end\n"
"    else\n"
"        assert(auth_def['pap-sha256'])\n"
"        local current_salt = auth_def['pap-sha256'][1]\n"
"        local current_hash = auth_def['pap-sha256'][2]\n"
"\n"
"        local new_hash = digest.sha256(current_salt .. password)\n"
"        if new_hash == current_hash then\n"
"            -- Note: passwd() generated new random salt, it will be different\n"
"            -- from current_salt.\n"
"            new_password = true\n"
"        end\n"
"    end\n"
"\n"
"    if not new_password then\n"
"        -- Note that security.auth_type is applied by box_cfg applier.\n"
"        -- It is executed before credentials applier, so the current\n"
"        -- box.cfg.auth_type is already set.\n"
"        if box.cfg.auth_type == auth_type then\n"
"            log.verbose('credentials.apply: a password is already set ' ..\n"
"                        'for user %q', user_name)\n"
"        else\n"
"            log.verbose('credentials.apply: a password for user %q has ' ..\n"
"                        'different auth_type, resetting it', user_name)\n"
"            box.schema.user.passwd(user_name, password)\n"
"        end\n"
"    else\n"
"        log.verbose('credentials.apply: set a password for user %q',\n"
"                    user_name)\n"
"        box.schema.user.passwd(user_name, password)\n"
"    end\n"
"\n"
"end\n"
"\n"
"local function create_users(user_map)\n"
"    for user_name, user_def in pairs(user_map or {}) do\n"
"        if user_def ~= nil then\n"
"            create_user(user_name)\n"
"            set_password(user_name, user_def.password)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- }}} Create users\n"
"\n"
"-- {{{ Applier\n"
"\n"
"-- Objects that are inside the config must be registered in target_object_map,\n"
"-- so they will trigger synchronization every time an object with such name\n"
"-- is created or is being renamed to the config-provided name.\n"
"local function register_objects(users_or_roles_config)\n"
"    for _name, config in pairs(users_or_roles_config or {}) do\n"
"        local privileges = privileges_from_config(config)\n"
"        for obj_type in pairs(target_object_map) do\n"
"            for obj_name in pairs(privileges[obj_type] or {}) do\n"
"                target_object_map[obj_type][obj_name] = true\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Fiber channel for resulting message to block execution\n"
"-- of main fiber in case of 'BLOCKING_FULL_SYNC'.\n"
"local wait_sync\n"
"\n"
"-- The worker gets synchronization commands from sync_tasks.\n"
"-- If the instance is in RO mode, the worker will wait for it\n"
"-- to be switched to RW.\n"
"-- Possible commands:\n"
"-- {type = 'BLOCKING_FULL_SYNC'} - perform a full sync, write message\n"
"--                                 to `wait_sync` on return\n"
"-- {type = 'BACKGROUND_FULL_SYNC'} - perform a full sync, skip return message\n"
"-- {type = obj_type, name = obj_name} - perform a per-object sync, skip\n"
"--                                      return message\n"
"local function sync_credentials_worker()\n"
"    while true do\n"
"        local obj_to_sync = sync_tasks:get()\n"
"\n"
"        if obj_to_sync.type == 'BLOCKING_FULL_SYNC' or\n"
"                obj_to_sync.type == 'BACKGROUND_FULL_SYNC' then\n"
"\n"
"            if box.info.ro then\n"
"                log.verbose('credentials: the database is in the read-only ' ..\n"
"                    'mode. Waiting for the read-write mode to set up the ' ..\n"
"                    'credentials given in the configuration. If another ' ..\n"
"                    'instance in the replicaset will write these new ' ..\n"
"                    'credentials, this instance will receive them and will ' ..\n"
"                    'skip further actions.')\n"
"                box.ctl.wait_rw()\n"
"            end\n"
"\n"
"            -- On config reload, drop the old list of registered objects.\n"
"            target_object_map = {\n"
"                ['space'] = {},\n"
"                ['function'] = {},\n"
"                ['sequence'] = {},\n"
"            }\n"
"\n"
"            local credentials = get_credentials(config)\n"
"\n"
"            register_objects(credentials.roles)\n"
"            register_objects(credentials.users)\n"
"\n"
"            box.atomic(function()\n"
"                create_roles(credentials.roles)\n"
"                create_users(credentials.users)\n"
"\n"
"                sync_privileges(credentials)\n"
"            end)\n"
"\n"
"            if obj_to_sync.type == 'BLOCKING_FULL_SYNC' then\n"
"                wait_sync:put('Done')\n"
"            end\n"
"        else\n"
"            local credentials = get_credentials(config)\n"
"\n"
"            box.atomic(sync_privileges, credentials, obj_to_sync)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Credentials are being synced inside this dedicated fiber worker.\n"
"local sync_credentials_fiber\n"
"\n"
"-- One-shot flag. It indicates that on_replace triggers are\n"
"-- set for box.space._space/_func/_sequence.\n"
"local triggers_are_set\n"
"\n"
"local function apply(config_module)\n"
"    config = config_module\n"
"\n"
"    -- Create a fiber channel for scheduled tasks for sync worker.\n"
"    if not sync_tasks then\n"
"        -- There is no good reason to have the channel capacity limited,\n"
"        -- but tarantool fiber channels have an upper limit by design.\n"
"        -- Thus, in rare cases when the channel is full, the mandatory\n"
"        -- full sync is executed, dropping all other scheduled tasks.\n"
"        sync_tasks = fiber.channel(1024)\n"
"    else\n"
"        -- Clear scheduled syncs on reload. Full sync will be executed\n"
"        -- afterwards anyway.\n"
"        while not sync_tasks:is_empty() do\n"
"            sync_tasks:get()\n"
"        end\n"
"    end\n"
"\n"
"    if not wait_sync then\n"
"        wait_sync = fiber.channel()\n"
"    end\n"
"\n"
"    -- Set trigger on after space/function/sequence creation.\n"
"    if not triggers_are_set then\n"
"        triggers_are_set = true\n"
"        box.space._space:on_replace(on_replace_trigger)\n"
"        box.space._func:on_replace(on_replace_trigger)\n"
"        box.space._sequence:on_replace(on_replace_trigger)\n"
"    end\n"
"\n"
"    -- Fiber is required here to have the credentials synced in the background\n"
"    -- when Tarantool is in Read Only mode after it is switched to RW.\n"
"\n"
"    -- Note that only one fiber exists at a time.\n"
"    if sync_credentials_fiber == nil or\n"
"            sync_credentials_fiber:status() == 'dead' then\n"
"        sync_credentials_fiber = fiber.new(sync_credentials_worker)\n"
"    end\n"
"\n"
"    -- If Tarantool is already in Read Write mode, credentials are still\n"
"    -- applied in the fiber to avoid possible concurrency issues, but the\n"
"    -- main applier fiber is blocked by `wait_sync`.\n"
"    if not box.info.ro then\n"
"        -- Schedule a full sync with a result message on return.\n"
"        sync_tasks:put({type = 'BLOCKING_FULL_SYNC'})\n"
"\n"
"        wait_sync:get()\n"
"    else\n"
"        -- Schedule a full sync in the background. It will be executed\n"
"        -- when the instance switches to RW.\n"
"        sync_tasks:put({type = 'BACKGROUND_FULL_SYNC'})\n"
"    end\n"
"end\n"
"\n"
"-- }}} Applier\n"
"\n"
"return {\n"
"    name = 'credentials',\n"
"    apply = apply,\n"
"    -- Exported for testing purposes.\n"
"    _internal = {\n"
"        privileges_from_box = privileges_from_box,\n"
"        privileges_from_config = privileges_from_config,\n"
"        privileges_subtract = privileges_subtract,\n"
"        privileges_add_defaults = privileges_add_defaults,\n"
"        sync_privileges = sync_privileges,\n"
"        set_password = set_password,\n"
"    },\n"
"}\n"
""
;
