const char config_applier_mkdir_lua[] =
"local fio = require('fio')\n"
"local log = require('internal.config.utils.log')\n"
"\n"
"-- Create a directory if it doesn't exist.\n"
"local function safe_mkdir(prefix, dir, work_dir)\n"
"    -- All the directories are interpreted as relative to\n"
"    -- `process.work_dir`. The first box.cfg() call sets a\n"
"    -- current working directory to this path.\n"
"    --\n"
"    -- However, we should prepend paths manually before the first\n"
"    -- box.cfg() call.\n"
"    --\n"
"    -- The absolute path is checked explicitly due to gh-8816.\n"
"    local needs_prepending = type(box.cfg) == 'function' and\n"
"        work_dir ~= nil and not dir:startswith('/')\n"
"    if needs_prepending then\n"
"        dir = fio.pathjoin(work_dir, dir)\n"
"    end\n"
"\n"
"    local stat = fio.stat(dir)\n"
"\n"
"    if stat == nil then\n"
"        log.verbose('%s: create directory: %s', prefix, dir)\n"
"        local _, err = fio.mktree(dir)\n"
"        if err ~= nil then\n"
"            error(('%s: failed to create directory %s: %s'):format(prefix,\n"
"                dir, err))\n"
"        end\n"
"    else\n"
"        if fio.path.is_dir(dir) then\n"
"            log.verbose('%s: the directory already exists: %s', prefix, dir)\n"
"        else\n"
"            error(('%s: the file is supposed to be a directory, ' ..\n"
"                'but it is not a directory: %s'):format(prefix, dir))\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function apply(config)\n"
"    local configdata = config._configdata\n"
"    local work_dir = configdata:get('process.work_dir', {use_default = true})\n"
"\n"
"    -- Create process.work_dir directory if box.cfg() is not\n"
"    -- called yet.\n"
"    if type(box.cfg) == 'function' and work_dir ~= nil then\n"
"        local prefix = ('mkdir.apply[%s]'):format('process.work_dir')\n"
"        safe_mkdir(prefix, work_dir)\n"
"    end\n"
"\n"
"    configdata:filter(function(w)\n"
"        return w.schema.mkdir ~= nil\n"
"    end, {use_default = true}):each(function(w)\n"
"        if w.data == nil then\n"
"            return\n"
"        end\n"
"        local prefix = ('mkdir.apply[%s]'):format(table.concat(w.path, '.'))\n"
"        safe_mkdir(prefix, w.data, work_dir)\n"
"    end)\n"
"\n"
"    configdata:filter(function(w)\n"
"        return w.schema.mk_parent_dir ~= nil\n"
"    end, {use_default = true}):each(function(w)\n"
"        if w.data == nil then\n"
"            return\n"
"        end\n"
"        local prefix = ('mkdir.apply[%s]'):format(table.concat(w.path, '.'))\n"
"        safe_mkdir(prefix, fio.dirname(w.data), work_dir)\n"
"    end)\n"
"\n"
"    local console = configdata:get('console', {use_default = true})\n"
"    if console.enabled then\n"
"        local prefix = ('mkdir.apply[%s]'):format('console.socket')\n"
"        safe_mkdir(prefix, fio.dirname(console.socket), work_dir)\n"
"    end\n"
"\n"
"    local log = configdata:get('log', {use_default = true})\n"
"    if log.to == 'file' then\n"
"        local prefix = ('mkdir.apply[%s]'):format('log.file')\n"
"        safe_mkdir(prefix, fio.dirname(log.file), work_dir)\n"
"    end\n"
"\n"
"    local audit_log = configdata:get('audit_log', {use_default = true})\n"
"    if audit_log ~= nil and audit_log.to == 'file' then\n"
"        local prefix = ('mkdir.apply[%s]'):format('audit_log.file')\n"
"        safe_mkdir(prefix, fio.dirname(audit_log.file), work_dir)\n"
"    end\n"
"end\n"
"\n"
"return {\n"
"    name = 'mkdir',\n"
"    apply = apply,\n"
"}\n"
""
;
