const char config_applier_roles_lua[] =
"local log = require('internal.config.utils.log')\n"
"\n"
"local last_loaded = {}\n"
"local last_roles_ordered = {}\n"
"\n"
"local function stop_roles(roles_to_skip)\n"
"    local roles_to_stop = {}\n"
"    for id = #last_roles_ordered, 1, -1 do\n"
"        local role_name = last_roles_ordered[id]\n"
"        if roles_to_skip == nil or roles_to_skip[role_name] == nil then\n"
"            table.insert(roles_to_stop, role_name)\n"
"        end\n"
"    end\n"
"    if #roles_to_stop == 0 then\n"
"        return\n"
"    end\n"
"    local deps = {}\n"
"    for role_name in pairs(roles_to_skip or {}) do\n"
"        local role = last_loaded[role_name] or {}\n"
"        -- There is no need to check transitive dependencies for roles_to_skip,\n"
"        -- because they were already checked when roles were started, i.e. if\n"
"        -- role A depends on role B, which depends on role C, and we stop\n"
"        -- role C, then we will get the error that role B depends on role C.\n"
"        for _, dep in pairs(role.dependencies or {}) do\n"
"            deps[dep] = deps[dep] or {}\n"
"            table.insert(deps[dep], role_name)\n"
"        end\n"
"    end\n"
"    for _, role_name in ipairs(roles_to_stop) do\n"
"        if deps[role_name] ~= nil then\n"
"            local err\n"
"            if #deps[role_name] == 1 then\n"
"                err =('role %q depends on it'):format(deps[role_name][1])\n"
"            else\n"
"                local names = {}\n"
"                for _, v in ipairs(deps[role_name]) do\n"
"                    table.insert(names, (\"%q\"):format(v))\n"
"                end\n"
"                local names_str = table.concat(names, ', ')\n"
"                err = ('roles %s depend on it'):format(names_str)\n"
"            end\n"
"            error(('Role %q cannot be stopped because %s'):format(role_name,\n"
"                                                                  err), 0)\n"
"        end\n"
"    end\n"
"    for _, role_name in ipairs(roles_to_stop) do\n"
"        log.verbose('roles.apply: stop role ' .. role_name)\n"
"        local ok, err = pcall(last_loaded[role_name].stop)\n"
"        if not ok then\n"
"            error(('Error stopping role %s: %s'):format(role_name, err), 0)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function resort_roles(original_order, roles)\n"
"    local ordered = {}\n"
"\n"
"    -- Needed to detect circular dependencies.\n"
"    local to_add = {}\n"
"\n"
"    -- To skip already added roles.\n"
"    local added = {}\n"
"\n"
"    local function add_role(role_name)\n"
"        if added[role_name] then\n"
"            return\n"
"        end\n"
"\n"
"        to_add[role_name] = true\n"
"\n"
"        for _, dep in ipairs(roles[role_name].dependencies or {}) do\n"
"            -- Detect a role that is not in the list of instance's roles.\n"
"            if not roles[dep] then\n"
"                local err = 'Role %q requires role %q, but the latter is ' ..\n"
"                            'not in the list of roles of the instance'\n"
"                error(err:format(role_name, dep), 0)\n"
"            end\n"
"\n"
"            -- Detect a circular dependency.\n"
"            if to_add[dep] and role_name == dep then\n"
"                local err = 'Circular dependency: role %q depends on itself'\n"
"                error(err:format(role_name), 0)\n"
"            end\n"
"            if to_add[dep] and role_name ~= dep then\n"
"                local err = 'Circular dependency: roles %q and %q depend on ' ..\n"
"                            'each other'\n"
"                error(err:format(role_name, dep), 0)\n"
"            end\n"
"\n"
"            -- Go into the recursion: add the dependency.\n"
"            add_role(dep)\n"
"        end\n"
"\n"
"        to_add[role_name] = nil\n"
"        added[role_name] = true\n"
"        table.insert(ordered, role_name)\n"
"    end\n"
"\n"
"    -- Keep the order, where the dependency tree doesn't obligate\n"
"    -- us to change it.\n"
"    for _, role_name in ipairs(original_order) do\n"
"        assert(roles[role_name] ~= nil)\n"
"        add_role(role_name)\n"
"    end\n"
"\n"
"    return ordered\n"
"end\n"
"\n"
"local function apply(config)\n"
"    local configdata = config._configdata\n"
"    local role_names = configdata:get('roles', {use_default = true})\n"
"    if role_names == nil or next(role_names) == nil then\n"
"        stop_roles()\n"
"        return\n"
"    end\n"
"\n"
"    -- Remove duplicates.\n"
"    local roles = {}\n"
"    local roles_ordered = {}\n"
"    for _, role_name in pairs(role_names) do\n"
"        if roles[role_name] == nil then\n"
"            table.insert(roles_ordered, role_name)\n"
"        end\n"
"        roles[role_name] = true\n"
"    end\n"
"\n"
"    -- Stop removed roles.\n"
"    stop_roles(roles)\n"
"\n"
"    -- Load roles.\n"
"    local loaded = {}\n"
"    for _, role_name in ipairs(roles_ordered) do\n"
"        local role = last_loaded[role_name]\n"
"        if not role then\n"
"            log.verbose('roles.apply: load role ' .. role_name)\n"
"            role = require(role_name)\n"
"            local funcs = {'validate', 'apply', 'stop'}\n"
"            for _, func_name in pairs(funcs) do\n"
"                if type(role[func_name]) ~= 'function' then\n"
"                    local err = 'Role %s does not contain function %s'\n"
"                    error(err:format(role_name, func_name), 0)\n"
"                end\n"
"            end\n"
"        end\n"
"        loaded[role_name] = role\n"
"        if role.dependencies ~= nil and type(role.dependencies) ~= 'table' then\n"
"            local err = 'Role %q has field \"dependencies\" of type %s, '..\n"
"                        'array-like table or nil expected'\n"
"            error(err:format(role_name, type(role.dependencies)), 0)\n"
"        end\n"
"    end\n"
"\n"
"    -- Re-sorting of roles taking into account dependencies between them.\n"
"    roles_ordered = resort_roles(roles_ordered, loaded)\n"
"\n"
"    -- Validate configs for all roles.\n"
"    local roles_cfg = configdata:get('roles_cfg', {use_default = true}) or {}\n"
"    for _, role_name in ipairs(roles_ordered) do\n"
"        local ok, err = pcall(loaded[role_name].validate, roles_cfg[role_name])\n"
"        if not ok then\n"
"            error(('Wrong config for role %s: %s'):format(role_name, err), 0)\n"
"        end\n"
"    end\n"
"\n"
"    -- Apply configs for all roles.\n"
"    for _, role_name in ipairs(roles_ordered) do\n"
"        log.verbose('roles.apply: apply config for role ' .. role_name)\n"
"        local ok, err = pcall(loaded[role_name].apply, roles_cfg[role_name])\n"
"        if not ok then\n"
"            error(('Error applying role %s: %s'):format(role_name, err), 0)\n"
"        end\n"
"    end\n"
"\n"
"    last_loaded = loaded\n"
"    last_roles_ordered = roles_ordered\n"
"end\n"
"\n"
"return {\n"
"    name = 'roles',\n"
"    apply = apply,\n"
"}\n"
""
;
