const char config_applier_sharding_lua[] =
"local expression = require('internal.config.utils.expression')\n"
"local log = require('internal.config.utils.log')\n"
"_G.vshard = nil\n"
"\n"
"-- Watcher which will create all the necessary functions.\n"
"local watcher = nil\n"
"\n"
"local function apply(config)\n"
"    local configdata = config._configdata\n"
"    local roles = configdata:get('sharding.roles')\n"
"    if roles == nil then\n"
"        return\n"
"    end\n"
"    -- Make sure vshard is available and its version is not too old.\n"
"    local ok, vshard = pcall(require, 'vshard')\n"
"    if not ok then\n"
"        error('The vshard module is not available', 0)\n"
"    end\n"
"    if expression.eval('v < 0.1.25', {v = vshard.consts.VERSION}) then\n"
"        error('The vshard module is too old: the minimum supported version ' ..\n"
"              'is 0.1.25.', 0)\n"
"    end\n"
"\n"
"    _G.vshard = vshard\n"
"    local is_storage = false\n"
"    local is_router = false\n"
"    for _, role in pairs(roles) do\n"
"        if role == 'storage' then\n"
"            is_storage = true\n"
"        elseif role == 'router' then\n"
"            is_router = true\n"
"        end\n"
"    end\n"
"    local cfg = configdata:sharding()\n"
"    if is_storage then\n"
"        -- Start a watcher which will create all the necessary functions.\n"
"        if watcher == nil then\n"
"            local function deploy_funcs()\n"
"                local vexports = require('vshard.storage.exports')\n"
"                local exports = vexports.compile(vexports.log[#vexports.log])\n"
"                vexports.deploy_funcs(exports)\n"
"            end\n"
"            watcher = box.watch('box.status', function(_, status)\n"
"                -- It's ok, if deploy_funcs() will be triggered several times.\n"
"                if status.is_ro == false then\n"
"                    deploy_funcs()\n"
"                end\n"
"            end)\n"
"        end\n"
"        log.info('sharding: apply storage config')\n"
"        -- Name may be not set in box.info.name, e.g. during names applying.\n"
"        -- Configure vshard anyway, pass configuration name.\n"
"        _G.vshard.storage.cfg(cfg, configdata:names().instance_name)\n"
"    elseif watcher ~= nil then\n"
"        watcher:unregister()\n"
"        watcher = nil\n"
"    end\n"
"    if is_router then\n"
"        log.info('sharding: apply router config')\n"
"        _G.vshard.router.cfg(cfg)\n"
"    end\n"
"end\n"
"\n"
"return {\n"
"    name = 'sharding',\n"
"    apply = apply,\n"
"}\n"
""
;
