const char config_cluster_config_lua[] =
"local schema = require('internal.config.utils.schema')\n"
"local instance_config = require('internal.config.instance_config')\n"
"local expression = require('internal.config.utils.expression')\n"
"\n"
"-- Extract a field from a table.\n"
"--\n"
"-- f({a = 1, b = 2}, 'a') -> {b = 2}, 1\n"
"--\n"
"-- The original table is not modified, but copied if necessary.\n"
"local function table_extract_field(t, field_name)\n"
"    local field_value = t[field_name]\n"
"    if field_value == nil then\n"
"        return t\n"
"    end\n"
"    t = table.copy(t)\n"
"    t[field_name] = nil\n"
"    return t, field_value\n"
"end\n"
"\n"
"-- Cluster config methods.\n"
"local methods = {}\n"
"\n"
"function methods.find_instance(_self, data, instance_name)\n"
"    -- Find group, replicaset, instance configuration for the\n"
"    -- given instance.\n"
"    local groups = data.groups or {}\n"
"    for gn, g in pairs(groups) do\n"
"        local replicasets = g.replicasets or {}\n"
"        for rn, r in pairs(replicasets) do\n"
"            local instances = r.instances or {}\n"
"            if instances[instance_name] ~= nil then\n"
"                return {\n"
"                    group = g,\n"
"                    replicaset = r,\n"
"                    instance = instances[instance_name],\n"
"                    group_name = gn,\n"
"                    replicaset_name = rn,\n"
"                }\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    return nil\n"
"end\n"
"\n"
"function methods.instantiate(self, data, instance_name)\n"
"    -- No topology information provided.\n"
"    if data.groups == nil then\n"
"        return data\n"
"    end\n"
"\n"
"    local found = self:find_instance(data, instance_name)\n"
"\n"
"    if found == nil then\n"
"        local res = table.copy(data)\n"
"        res.groups = nil\n"
"        return res\n"
"    end\n"
"\n"
"    local res = {}\n"
"    res = instance_config:merge(res, data)\n"
"    res = instance_config:merge(res, found.group)\n"
"    res = instance_config:merge(res, found.replicaset)\n"
"    res = instance_config:merge(res, found.instance)\n"
"    return res\n"
"end\n"
"\n"
"-- Construct a record from other records and extra fields.\n"
"--\n"
"-- record_from_fields({\n"
"--     <record foo>\n"
"--     <record bar>\n"
"--     extra_field_baz = <...>,\n"
"--     extra_field_fiz = <...>,\n"
"-- })\n"
"--\n"
"-- It allows to write the cluster config schema in a more readable\n"
"-- way.\n"
"local function record_from_fields(fields)\n"
"    local res = {\n"
"        type = 'record',\n"
"        fields = {},\n"
"        -- <..annotations..>\n"
"    }\n"
"\n"
"    for k, v in pairs(fields) do\n"
"        if type(k) == 'number' then\n"
"            -- Assume that a numeric key contains a record to\n"
"            -- copy its fields and annotations into the resulting\n"
"            -- record.\n"
"            assert(type(v) == 'table')\n"
"            assert(v.type == 'record')\n"
"\n"
"            -- Copy fields.\n"
"            for kk, vv in pairs(v.fields) do\n"
"                assert(res.fields[kk] == nil, 'record_from_fields: duplicate '..\n"
"                                              'fields '..tostring(kk))\n"
"                res.fields[kk] = vv\n"
"            end\n"
"\n"
"            -- Copy annotations.\n"
"            for kk, vv in pairs(v) do\n"
"                if kk ~= 'fields' and kk ~= 'type' then\n"
"                    assert(res[kk] == nil, 'record_from_fields: duplicate '..\n"
"                                           'annotations '..tostring(kk))\n"
"                    res[kk] = vv\n"
"                end\n"
"            end\n"
"        else\n"
"            -- Assume that a string key represents a field name\n"
"            -- and the corresponding value contains a schema node\n"
"            -- for the field.\n"
"            assert(type(k) == 'string')\n"
"\n"
"            -- Copy the field.\n"
"            assert(res.fields[k] == nil, 'record_from_fields: duplicate '..\n"
"                                         'fields '..tostring(k))\n"
"            res.fields[k] = v\n"
"        end\n"
"    end\n"
"\n"
"    return res\n"
"end\n"
"\n"
"--\n"
"-- Return the instance configuration schema as a record with the given scope\n"
"-- annotation.\n"
"--\n"
"local function instance_config_with_scope(scope)\n"
"    assert(type(instance_config) == 'table')\n"
"    assert(type(instance_config.schema) == 'table')\n"
"    assert(instance_config.schema.type == 'record')\n"
"\n"
"    local schema = table.copy(instance_config.schema)\n"
"    schema.scope = scope\n"
"    return schema\n"
"end\n"
"\n"
"-- Validate instance_name, replicaset_name, group_name.\n"
"--\n"
"-- Please, keep it is-sync with src/box/node_name.[ch].\n"
"function methods.validate_name(_self, name)\n"
"    local NODE_NAME_LEN_MAX = 63\n"
"    if #name == 0 then\n"
"        return false, 'Zero length name is forbidden'\n"
"    end\n"
"    if #name > NODE_NAME_LEN_MAX then\n"
"        return false, ('A name must fit %d characters limit, got %q'):format(\n"
"            NODE_NAME_LEN_MAX, name)\n"
"    end\n"
"    if name:match('^[a-z0-9_-]+$') == nil then\n"
"        return false, ('A name must contain only lowercase letters, digits, ' ..\n"
"            'dash and underscore, got %q'):format(name)\n"
"    end\n"
"    if name:match('^[a-z]') == nil then\n"
"        return false, ('A name must start from a lowercase letter, ' ..\n"
"            'got %q'):format(name)\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"-- A schema node that represents an instance name, a replicaset\n"
"-- name, a group name.\n"
"local name = schema.scalar({\n"
"    type = 'string',\n"
"    validate = function(name, w)\n"
"        local ok, err = methods.validate_name(nil, name)\n"
"        if not ok then\n"
"            w.error(err)\n"
"        end\n"
"    end,\n"
"})\n"
"\n"
"-- Nested cluster config. It is an auxiliary schema.\n"
"--\n"
"-- The cluster config has the following structure.\n"
"--\n"
"-- {\n"
"--     conditional = array-of(*nested_cluster_config + if),\n"
"--     *nested_cluster_config,\n"
"-- }\n"
"--\n"
"-- Note 1: The asterisk denotes unpacking all the fields into the\n"
"-- parent schema node.\n"
"--\n"
"-- Note 2: The nested cluster config in the conditional section is\n"
"-- represented as an arbitrary map in the schema, but validatated\n"
"-- against the nested cluster config schema before merging into\n"
"-- the main config.\n"
"local schema_name = 'nested_cluster_config'\n"
"local nested_cluster_config = schema.new(schema_name, record_from_fields({\n"
"    instance_config_with_scope('global'),\n"
"    groups = schema.map({\n"
"        key = name,\n"
"        value = record_from_fields({\n"
"            instance_config_with_scope('group'),\n"
"            replicasets = schema.map({\n"
"                key = name,\n"
"                value = record_from_fields({\n"
"                    leader = schema.scalar({type = 'string'}),\n"
"                    bootstrap_leader = schema.scalar({type = 'string'}),\n"
"                    instance_config_with_scope('replicaset'),\n"
"                    instances = schema.map({\n"
"                        key = name,\n"
"                        value = instance_config_with_scope('instance'),\n"
"                    }),\n"
"                }),\n"
"            }),\n"
"        }),\n"
"    }),\n"
"}))\n"
"\n"
"-- {{{ Support conditional sections\n"
"\n"
"local conditional_vars = {\n"
"    tarantool_version = _TARANTOOL:match('^%d+%.%d+%.%d+'),\n"
"}\n"
"assert(conditional_vars.tarantool_version ~= nil)\n"
"\n"
"-- Merge conditional sections into the main config.\n"
"--\n"
"-- Accepts a cluster configuration data with conditional sections\n"
"-- as an input and returns the data with the sections removed and\n"
"-- applied to the main config.\n"
"--\n"
"-- Conditional sections that don't fit the `if` clause criteria\n"
"-- are skipped.\n"
"--\n"
"--  | conditional:\n"
"--  | - aaa: {bbb: 1}\n"
"--  |   if: tarantool_version >= 3.0.0\n"
"--  | - ccc: {ddd: 2}\n"
"--  |   if: tarantool_version < 1.0.0\n"
"--  | eee: {fff: 3}\n"
"--\n"
"-- ->\n"
"--\n"
"--  | aaa: {bbb: 1}\n"
"--  | # no ccc\n"
"--  | eee: {fff: 3}\n"
"function methods.apply_conditional(_self, data)\n"
"    if data == nil then\n"
"        return data\n"
"    end\n"
"    local data, conditional = table_extract_field(data, 'conditional')\n"
"    if conditional == nil then\n"
"        return data\n"
"    end\n"
"\n"
"    -- Look over the conditional sections and if the predicate in\n"
"    -- `if` field evaluates to `true`, merge the section in the\n"
"    -- data.\n"
"    local res = data\n"
"    for _, conditional_config in ipairs(conditional) do\n"
"        local config, expr = table_extract_field(conditional_config, 'if')\n"
"        if expression.eval(expr, conditional_vars) then\n"
"            -- NB: The nested configuration doesn't allow a\n"
"            -- presence of the 'conditional' field. Use the\n"
"            -- corresponding schema to perform the merge: nested\n"
"            -- cluster config.\n"
"            --\n"
"            -- It is also important that we use this schema\n"
"            -- instead of a map of 'any' values to perform a deep\n"
"            -- merge, not just replace fields at the top level.\n"
"            --\n"
"            -- NB: The section is already validated in\n"
"            -- validate_conditional().\n"
"            res = nested_cluster_config:merge(res, config)\n"
"        end\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- Verify conditional sections, whose `if` predicate evaluates to\n"
"-- `true`.\n"
"--\n"
"-- Also, verify that the predicate is present in each of such\n"
"-- sections and has correct expression.\n"
"local function validate_conditional(data, w)\n"
"    -- Each conditional section should have 'if' key.\n"
"    local data, expr = table_extract_field(data, 'if')\n"
"    if expr == nil then\n"
"        w.error('A conditional section should have field \"if\"')\n"
"    end\n"
"\n"
"    -- Fail the validation if the expression is incorrect.\n"
"    local ok, res = pcall(expression.eval, expr, conditional_vars)\n"
"    if not ok then\n"
"        w.error(res)\n"
"    end\n"
"\n"
"    -- Fail the validation if this conditional section is to be\n"
"    -- applied, but it doesn't fit the schema.\n"
"    --\n"
"    -- NB: This validation doesn't accept 'conditional' field, so\n"
"    -- the appropriate schema is used for the validation (nested\n"
"    -- cluster config).\n"
"    if res then\n"
"        nested_cluster_config:validate(data)\n"
"    end\n"
"end\n"
"\n"
"return schema.new('cluster_config', record_from_fields({\n"
"    conditional = schema.array({\n"
"        items = schema.map({\n"
"            -- This map represents a cluster config with\n"
"            -- additional 'if' key.\n"
"            --\n"
"            -- However, it is not necessarily represents a cluster\n"
"            -- config of the given tarantool version: it may be\n"
"            -- older or newer and so it may contain fields that\n"
"            -- are unknown from our point of view.\n"
"            --\n"
"            -- It is why the mapping type is used here: we allow\n"
"            -- arbitrary data and valitate only those items that\n"
"            -- are going to be merged into the main config\n"
"            -- (i.e. the validation is performed after the version\n"
"            -- comparison).\n"
"            key = schema.scalar({type = 'string'}),\n"
"            value = schema.scalar({type = 'any'}),\n"
"            validate = validate_conditional,\n"
"        })\n"
"    }),\n"
"    nested_cluster_config.schema,\n"
"}), {\n"
"    methods = methods,\n"
"})\n"
"\n"
"-- }}} Support conditional sections\n"
""
;
