const char config_configdata_lua[] =
"-- Accumulates configuration data of different kinds and provides\n"
"-- accessors.\n"
"--\n"
"-- Intended to be used as an immutable object.\n"
"\n"
"local fun = require('fun')\n"
"local urilib = require('uri')\n"
"local uuid = require('uuid')\n"
"local instance_config = require('internal.config.instance_config')\n"
"local cluster_config = require('internal.config.cluster_config')\n"
"local snapshot = require('internal.config.utils.snapshot')\n"
"\n"
"local function choose_iconfig(self, opts)\n"
"    if opts ~= nil and opts.peer ~= nil then\n"
"        local peers = self._peers\n"
"        local peer = peers[opts.peer]\n"
"        if peer == nil then\n"
"            error(('Unknown peer %q'):format(opts.peer), 0)\n"
"        end\n"
"        if opts ~= nil and opts.use_default then\n"
"            return peer.iconfig_def\n"
"        end\n"
"        return peer.iconfig\n"
"    end\n"
"\n"
"    if opts ~= nil and opts.use_default then\n"
"        return self._iconfig_def\n"
"    else\n"
"        return self._iconfig\n"
"    end\n"
"end\n"
"\n"
"local methods = {}\n"
"\n"
"-- Acquire a value from the instance config.\n"
"--\n"
"-- opts:\n"
"--     use_default: boolean\n"
"--     peer: string\n"
"function methods.get(self, path, opts)\n"
"    local data = choose_iconfig(self, opts)\n"
"    return instance_config:get(data, path)\n"
"end\n"
"\n"
"-- Filter data based on the instance schema annotations.\n"
"--\n"
"-- opts:\n"
"--     use_default: boolean\n"
"--     peer: string\n"
"function methods.filter(self, f, opts)\n"
"    local data = choose_iconfig(self, opts)\n"
"    return instance_config:filter(data, f)\n"
"end\n"
"\n"
"-- List of names of the instances in the same replicaset.\n"
"--\n"
"-- The names are useful to pass to other methods as opts.peer.\n"
"function methods.peers(self)\n"
"    return self._peer_names\n"
"end\n"
"\n"
"-- Group, replicaset and instance names.\n"
"function methods.names(self)\n"
"    return {\n"
"        group_name = self._group_name,\n"
"        replicaset_name = self._replicaset_name,\n"
"        instance_name = self._instance_name,\n"
"        replicaset_uuid = self._replicaset_uuid,\n"
"        instance_uuid = self._instance_uuid,\n"
"    }\n"
"end\n"
"\n"
"local function instance_sharding(iconfig, instance_name)\n"
"    local roles = instance_config:get(iconfig, 'sharding.roles')\n"
"    if roles == nil or #roles == 0 then\n"
"        return nil\n"
"    end\n"
"    assert(type(roles) == 'table')\n"
"    local is_storage = false\n"
"    for _, role in pairs(roles) do\n"
"        is_storage = is_storage or role == 'storage'\n"
"    end\n"
"    if not is_storage then\n"
"        return nil\n"
"    end\n"
"    local zone = instance_config:get(iconfig, 'sharding.zone')\n"
"    local uri = instance_config:instance_uri(iconfig, 'sharding')\n"
"    if uri == nil then\n"
"        local err = 'No suitable URI provided for instance %q'\n"
"        error(err:format(instance_name), 0)\n"
"    end\n"
"    --\n"
"    -- Currently, vshard does not accept URI without a username. So if we got a\n"
"    -- URI without a username, use \"guest\" as the username without a password.\n"
"    --\n"
"    local u, err = urilib.parse(uri)\n"
"    -- NB: The URI is validated, so the parsing can't fail.\n"
"    assert(u ~= nil, err)\n"
"    if u.login == nil then\n"
"        u.login = 'guest'\n"
"        uri = urilib.format(u, true)\n"
"    end\n"
"    local uuid = instance_config:get(iconfig, 'database.instance_uuid')\n"
"\n"
"    local user = instance_config:get(iconfig, {'credentials', 'users', u.login})\n"
"    --\n"
"    -- If the user is not described in the credentials, this may mean that the\n"
"    -- user already exists and may have all the necessary privileges. If not, an\n"
"    -- error will be thrown later.\n"
"    --\n"
"    if user ~= nil then\n"
"        -- Check that the vshard storage user has the credential sharding role.\n"
"        local function check_sharding_role(roles)\n"
"            if roles == nil or next(roles) == nil then\n"
"                return false\n"
"            end\n"
"            for _, role_name in pairs(roles) do\n"
"                if role_name == 'sharding' then\n"
"                    return true\n"
"                end\n"
"            end\n"
"            for _, role_name in pairs(roles) do\n"
"                local path = {'credentials', 'roles', role_name, 'roles'}\n"
"                if check_sharding_role(instance_config:get(iconfig, path)) then\n"
"                    return true\n"
"                end\n"
"            end\n"
"            return false\n"
"        end\n"
"\n"
"        if not check_sharding_role(user.roles) then\n"
"            local err = \"storage user %q should have %q role\"\n"
"            error(err:format(u.login, 'sharding'), 0)\n"
"        end\n"
"    end\n"
"\n"
"    return {\n"
"        uri = uri,\n"
"        uuid = uuid,\n"
"        zone = zone,\n"
"    }\n"
"end\n"
"\n"
"function methods.sharding(self)\n"
"    local sharding = {}\n"
"    local rebalancers = {}\n"
"    for group_name, group in pairs(self._cconfig.groups) do\n"
"        for replicaset_name, value in pairs(group.replicasets) do\n"
"            local lock\n"
"            local replicaset_uuid\n"
"            local replicaset_cfg = {}\n"
"            local is_rebalancer = nil\n"
"            for instance_name, _ in pairs(value.instances) do\n"
"                local vars = {\n"
"                    instance_name = instance_name,\n"
"                    replicaset_name = replicaset_name,\n"
"                    group_name = group_name,\n"
"                }\n"
"                local iconfig = cluster_config:instantiate(self._cconfig,\n"
"                                                           instance_name)\n"
"                iconfig = instance_config:apply_default(iconfig)\n"
"                iconfig = instance_config:apply_vars(iconfig, vars)\n"
"                if lock == nil then\n"
"                    lock = instance_config:get(iconfig, 'sharding.lock')\n"
"                end\n"
"                if is_rebalancer == nil then\n"
"                    local roles = instance_config:get(iconfig, 'sharding.roles')\n"
"                    for _, role in pairs(roles) do\n"
"                        is_rebalancer = is_rebalancer or role == 'rebalancer'\n"
"                    end\n"
"                    if is_rebalancer then\n"
"                        table.insert(rebalancers, replicaset_name)\n"
"                    end\n"
"                end\n"
"                local isharding = instance_sharding(iconfig, instance_name)\n"
"                if isharding ~= nil then\n"
"                    if replicaset_uuid == nil then\n"
"                        replicaset_uuid = instance_config:get(iconfig,\n"
"                            'database.replicaset_uuid')\n"
"                    end\n"
"                    replicaset_cfg[instance_name] = isharding\n"
"                end\n"
"            end\n"
"            if next(replicaset_cfg) ~= nil then\n"
"                sharding[replicaset_name] = {\n"
"                    rebalancer = is_rebalancer or nil,\n"
"                    replicas = replicaset_cfg,\n"
"                    uuid = replicaset_uuid,\n"
"                    master = 'auto',\n"
"                    lock = lock,\n"
"                }\n"
"            end\n"
"        end\n"
"    end\n"
"    if #rebalancers > 1 then\n"
"        local err = \"The rebalancer role must be present in no more than \" ..\n"
"                    \"one replicaset. Replicasets with the role: %s\"\n"
"        error(err:format(table.concat(rebalancers, \", \")), 0)\n"
"    end\n"
"    local cfg = {\n"
"        sharding = sharding,\n"
"        box_cfg_mode = 'manual',\n"
"        --\n"
"        -- We set this option to \"manual\" to be able to manage privileges using\n"
"        -- the credentials config section and to be able to create the necessary\n"
"        -- vshard functions in case all instances in a replicaset are running in\n"
"        -- read-only mode (which is possible, for example, in case of\n"
"        -- replication.failover == election).\n"
"        --\n"
"        schema_management_mode = 'manual_access',\n"
"        identification_mode = 'name_as_key',\n"
"    }\n"
"\n"
"    local vshard_global_options = {\n"
"        'shard_index',\n"
"        'bucket_count',\n"
"        'rebalancer_disbalance_threshold',\n"
"        'rebalancer_max_receiving',\n"
"        'rebalancer_max_sending',\n"
"        'sync_timeout',\n"
"        'connection_outdate_delay',\n"
"        'failover_ping_timeout',\n"
"        'discovery_mode',\n"
"        'sched_ref_quota',\n"
"        'sched_move_quota',\n"
"    }\n"
"    for _, v in pairs(vshard_global_options) do\n"
"        cfg[v] = instance_config:get(self._iconfig_def, 'sharding.'..v)\n"
"    end\n"
"    return cfg\n"
"end\n"
"\n"
"-- Should be called only if the 'manual' failover method is\n"
"-- configured.\n"
"function methods.leader(self)\n"
"    assert(self._failover == 'manual')\n"
"    return self._leader\n"
"end\n"
"\n"
"-- Should be called only if the 'manual' failover method is\n"
"-- configured.\n"
"function methods.is_leader(self)\n"
"    assert(self._failover == 'manual')\n"
"    return self._leader == self._instance_name\n"
"end\n"
"\n"
"function methods.bootstrap_leader(self)\n"
"    return self._bootstrap_leader\n"
"end\n"
"\n"
"-- Should be called only if the 'supervised' failover method is\n"
"-- configured.\n"
"function methods.bootstrap_leader_name(self)\n"
"    assert(self._failover == 'supervised')\n"
"    return self._bootstrap_leader_name\n"
"end\n"
"\n"
"-- Returns instance_uuid and replicaset_uuid, saved in config.\n"
"local function find_uuids_by_name(peers, instance_name)\n"
"    for name, peer in pairs(peers) do\n"
"        if name == instance_name then\n"
"            local iconfig = peer.iconfig_def\n"
"            return instance_config:get(iconfig, 'database.instance_uuid'),\n"
"                   instance_config:get(iconfig, 'database.replicaset_uuid')\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"local function find_peer_name_by_uuid(peers, instance_uuid)\n"
"    for name, peer in pairs(peers) do\n"
"        local uuid = instance_config:get(peer.iconfig_def,\n"
"                                         'database.instance_uuid')\n"
"        if uuid == instance_uuid then\n"
"            return name\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"function methods.peer_name_by_uuid(self, instance_uuid)\n"
"    return find_peer_name_by_uuid(self._peers, instance_uuid)\n"
"end\n"
"\n"
"local function find_saved_names(iconfig)\n"
"    if type(box.cfg) == 'function' then\n"
"        local snap_path = snapshot.get_path(iconfig)\n"
"        -- Bootstrap is going to be done, no names are saved.\n"
"        if snap_path == nil then\n"
"            return nil\n"
"        end\n"
"\n"
"        -- Read system spaces of snap file.\n"
"        return snapshot.get_names(snap_path)\n"
"    end\n"
"\n"
"    -- Box.cfg was already done. No sense in snapshot\n"
"    -- reading, we can get all data from memory.\n"
"    local peers = {}\n"
"    for _, row in ipairs(box.space._cluster:select(nil, {limit = 32})) do\n"
"        if row[3] ~= nil then\n"
"            peers[row[3]] = row[2]\n"
"        end\n"
"    end\n"
"\n"
"    return {\n"
"        replicaset_name = box.info.replicaset.name,\n"
"        replicaset_uuid = box.info.replicaset.uuid,\n"
"        instance_name = box.info.name,\n"
"        instance_uuid = box.info.uuid,\n"
"        peers = peers,\n"
"    }\n"
"end\n"
"\n"
"-- Return a map, which shows, which instances doesn't have a name\n"
"-- set, info about the current replicaset name is also included in map.\n"
"function methods.missing_names(self)\n"
"    local missing_names = {\n"
"        -- Note, that replicaset_name cannot start with underscore (_peers\n"
"        -- name is forbidden), so we won't overwrite it with list of peers.\n"
"        _peers = {},\n"
"    }\n"
"\n"
"    local saved_names = find_saved_names(self._iconfig_def)\n"
"    if saved_names == nil then\n"
"        -- All names will be set during replicaset bootstrap.\n"
"        return missing_names\n"
"    end\n"
"\n"
"    -- Missing name of the current replicaset.\n"
"    if saved_names.replicaset_name == nil then\n"
"        missing_names[self._replicaset_name] = saved_names.replicaset_uuid\n"
"    end\n"
"\n"
"    for name, peer in pairs(self._peers) do\n"
"        local iconfig = peer.iconfig_def\n"
"        -- We allow anonymous replica without instance_uuid. Anonymous replica\n"
"        -- cannot have name set, it's enough to validate replicaset_name/uuid.\n"
"        if instance_config:get(iconfig, 'replication.anon') then\n"
"            goto continue\n"
"        end\n"
"\n"
"        -- cfg_uuid may be box.NULL if instance_uuid is not passed to config.\n"
"        local cfg_uuid = instance_config:get(iconfig, 'database.instance_uuid')\n"
"        if cfg_uuid == box.NULL then\n"
"            cfg_uuid = 'unknown'\n"
"        end\n"
"\n"
"        if not saved_names.peers[name] then\n"
"            missing_names._peers[name] = cfg_uuid\n"
"        end\n"
"\n"
"        ::continue::\n"
"    end\n"
"\n"
"    return missing_names\n"
"end\n"
"\n"
"local mt = {\n"
"    __index = methods,\n"
"}\n"
"\n"
"-- Validate UUIDs and names passed to config against the data,\n"
"-- saved inside snapshot. Fail early if mismatch is found.\n"
"local function validate_names(saved_names, config_names, iconfig)\n"
"    -- Snapshot always has replicaset uuid and\n"
"    -- at least one peer in _cluster space.\n"
"    if saved_names.replicaset_uuid == nil then\n"
"        local snap_path = snapshot.get_path(iconfig)\n"
"        error(('Snapshot file %s has no \"replicaset_uuid\" key in _cluster ' ..\n"
"            'system space. The snapshot is likely corrupted.'):format(\n"
"            snap_path), 0)\n"
"    end\n"
"    if saved_names.instance_uuid == uuid.NULL then\n"
"        local snap_path = snapshot.get_path(iconfig)\n"
"        error(('Snapshot file %s has no \"Instance\" header with an instance ' ..\n"
"            'UUID. The snapshot is likely corrupted.'):format(\n"
"            snap_path), 0)\n"
"    end\n"
"    -- Config always has names set.\n"
"    assert(config_names.replicaset_name ~= nil)\n"
"    assert(config_names.instance_name ~= nil)\n"
"\n"
"    if config_names.replicaset_uuid ~= nil and\n"
"       config_names.replicaset_uuid ~= saved_names.replicaset_uuid then\n"
"        error(string.format('Replicaset UUID mismatch. Snapshot: %s, ' ..\n"
"                            'config: %s.', saved_names.replicaset_uuid,\n"
"                            config_names.replicaset_uuid), 0)\n"
"    end\n"
"\n"
"    if saved_names.replicaset_name ~= nil and\n"
"       saved_names.replicaset_name ~= config_names.replicaset_name then\n"
"        error(string.format('Replicaset name mismatch. Snapshot: %s, ' ..\n"
"                            'config: %s.', saved_names.replicaset_name,\n"
"                            config_names.replicaset_name), 0)\n"
"    end\n"
"\n"
"    if config_names.instance_uuid ~= nil and\n"
"       config_names.instance_uuid ~= saved_names.instance_uuid then\n"
"        error(string.format('Instance UUID mismatch. Snapshot: %s, ' ..\n"
"                            'config: %s.', saved_names.instance_uuid,\n"
"                            config_names.instance_uuid), 0)\n"
"    end\n"
"\n"
"    if saved_names.instance_name ~= nil and\n"
"       saved_names.instance_name ~= config_names.instance_name then\n"
"        error(string.format('Instance name mismatch. Snapshot: %s, ' ..\n"
"                            'config: %s.', saved_names.instance_name,\n"
"                            config_names.instance_name), 0)\n"
"    end\n"
"\n"
"    -- Fail early, if current UUID is not set, but no name is found\n"
"    -- inside the snapshot file. Ignore this failure, if replica is\n"
"    -- configured as anonymous, anon replicas cannot have names.\n"
"    local iconfig = config_names.peers[config_names.instance_name].iconfig_def\n"
"    if not instance_config:get(iconfig, 'replication.anon') then\n"
"        if saved_names.instance_name == nil and\n"
"           config_names.instance_uuid == nil then\n"
"            error(string.format('Instance name for %s is not set in snapshot' ..\n"
"                                ' and UUID is missing in the config. Found ' ..\n"
"                                '%s in snapshot.', config_names.instance_name,\n"
"                                saved_names.instance_uuid), 0)\n"
"        end\n"
"        if saved_names.replicaset_name == nil and\n"
"           config_names.replicaset_uuid == nil then\n"
"            error(string.format('Replicaset name for %s is not set in ' ..\n"
"                                'snapshot and  UUID is missing in the ' ..\n"
"                                'config. Found %s in snapshot.',\n"
"                                config_names.replicaset_name,\n"
"                                saved_names.replicaset_uuid), 0)\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- A couple of replication.failover specific checks.\n"
"local function validate_failover(found, peers, failover, leader)\n"
"    if failover ~= 'manual' then\n"
"        -- Verify that no leader is set in the \"off\", \"election\"\n"
"        -- or \"supervised\" failover mode.\n"
"        if leader ~= nil then\n"
"            error(('\"leader\" = %q option is set for replicaset %q of group ' ..\n"
"                '%q, but this option cannot be used together with ' ..\n"
"                'replication.failover = %q'):format(leader,\n"
"                found.replicaset_name, found.group_name, failover), 0)\n"
"        end\n"
"    end\n"
"\n"
"    if failover ~= 'off' then\n"
"        -- Verify that peers in the given replicaset have no direct\n"
"        -- database.mode option set if the replicaset is configured\n"
"        -- with the \"manual\", \"election\" or \"supervised\" failover\n"
"        -- mode.\n"
"        --\n"
"        -- This check doesn't verify the whole cluster config, only\n"
"        -- the given replicaset.\n"
"        for peer_name, peer in pairs(peers) do\n"
"            local mode = instance_config:get(peer.iconfig, 'database.mode')\n"
"            if mode ~= nil then\n"
"                error(('database.mode = %q is set for instance %q of ' ..\n"
"                    'replicaset %q of group %q, but this option cannot be ' ..\n"
"                    'used together with replication.failover = %q'):format(mode,\n"
"                    peer_name, found.replicaset_name, found.group_name,\n"
"                    failover), 0)\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    if failover == 'manual' then\n"
"        -- Verify that the 'leader' option is set to a name of an\n"
"        -- existing instance from the given replicaset (or unset).\n"
"        if leader ~= nil and peers[leader] == nil then\n"
"            error(('\"leader\" = %q option is set for replicaset %q of group ' ..\n"
"                '%q, but instance %q is not found in this replicaset'):format(\n"
"                leader, found.replicaset_name, found.group_name, leader), 0)\n"
"        end\n"
"    end\n"
"\n"
"    -- Verify that 'election_mode' option is set to a value other\n"
"    -- than 'off' only in the 'failover: election' mode.\n"
"    --\n"
"    -- The alternative would be silent ignoring the election\n"
"    -- mode if failover mode is not 'election'.\n"
"    --\n"
"    -- For a while, a simple and straightforward approach is\n"
"    -- chosen: let the user create an explicit consistent\n"
"    -- configuration manually.\n"
"    --\n"
"    -- We can relax it in a future, though. For example, if two\n"
"    -- conflicting options are set in different scopes, we can\n"
"    -- ignore one from the outer scope.\n"
"    if failover ~= 'election' then\n"
"        for peer_name, peer in pairs(peers) do\n"
"            local election_mode = instance_config:get(peer.iconfig_def,\n"
"                'replication.election_mode')\n"
"            if election_mode ~= nil and election_mode ~= 'off' then\n"
"                error(('replication.election_mode = %q is set for instance ' ..\n"
"                    '%q of replicaset %q of group %q, but this option is ' ..\n"
"                    'only applicable if replication.failover = \"election\"; ' ..\n"
"                    'the replicaset is configured with replication.failover ' ..\n"
"                    '= %q; if this particular instance requires its own ' ..\n"
"                    'election mode, for example, if it is an anonymous ' ..\n"
"                    'replica, consider configuring the election mode ' ..\n"
"                    'specifically for this particular instance'):format(\n"
"                    election_mode, peer_name, found.replicaset_name,\n"
"                    found.group_name, failover), 0)\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Verify replication.anon = true prerequisites.\n"
"--\n"
"-- First, it verifies that the given replicaset contains at least\n"
"-- one non-anonymous replica.\n"
"--\n"
"-- The key idea of the rest of the checks is that an anonymous\n"
"-- replica must be in the read-only mode.\n"
"--\n"
"-- Different failover modes control read-only/read-write mode in\n"
"-- different ways, so we need specific checks for each of them in\n"
"-- regard of an anonymous replica.\n"
"--\n"
"-- These checks don't verify the whole cluster config, only the\n"
"-- given replicaset.\n"
"local function validate_anon(found, peers, failover, leader)\n"
"    -- failover: <any>\n"
"    --\n"
"    -- A replicaset can't consist of only anonymous replicas.\n"
"    assert(next(peers) ~= nil)\n"
"    local found_non_anon = false\n"
"    for _, peer in pairs(peers) do\n"
"        local is_anon =\n"
"            instance_config:get(peer.iconfig_def, 'replication.anon')\n"
"        if not is_anon then\n"
"            found_non_anon = true\n"
"            break\n"
"        end\n"
"    end\n"
"    if not found_non_anon then\n"
"        error(('All the instances of replicaset %q of group %q are ' ..\n"
"            'configured as anonymous replicas; it effectively means that ' ..\n"
"            'the whole replicaset is read-only; moreover, it means that ' ..\n"
"            'default replication.peers construction logic will create ' ..\n"
"            'empty upstream list and each instance is de-facto isolated: ' ..\n"
"            'neither is connected to any other; this configuration is ' ..\n"
"            'forbidden, because it looks like there is no meaningful ' ..\n"
"            'use case'):format(found.replicaset_name, found.group_name), 0)\n"
"    end\n"
"\n"
"    -- failover: off\n"
"    --\n"
"    -- An anonymous replica shouldn't be set to RW.\n"
"    if failover == 'off' then\n"
"        for peer_name, peer in pairs(peers) do\n"
"            local is_anon =\n"
"                instance_config:get(peer.iconfig_def, 'replication.anon')\n"
"            local mode =\n"
"                instance_config:get(peer.iconfig_def, 'database.mode')\n"
"            if is_anon and mode == 'rw' then\n"
"                error(('database.mode = \"rw\" is set for instance %q of ' ..\n"
"                    'replicaset %q of group %q, but this option cannot be ' ..\n"
"                    'used together with replication.anon = true'):format(\n"
"                    peer_name, found.replicaset_name, found.group_name), 0)\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    -- failover: manual\n"
"    --\n"
"    -- An anonymous replica can't be a leader.\n"
"    if failover == 'manual' and leader ~= nil then\n"
"        assert(peers[leader] ~= nil)\n"
"        local iconfig_def = peers[leader].iconfig_def\n"
"        local is_anon = instance_config:get(iconfig_def, 'replication.anon')\n"
"        if is_anon then\n"
"            error(('replication.anon = true is set for instance %q of ' ..\n"
"                'replicaset %q of group %q that is configured as a ' ..\n"
"                'leader; a leader can not be an anonymous replica'):format(\n"
"                leader, found.replicaset_name, found.group_name), 0)\n"
"        end\n"
"    end\n"
"\n"
"    -- failover: election\n"
"    --\n"
"    -- An anonymous replica can be in `election_mode: off`, but\n"
"    -- not any other.\n"
"    --\n"
"    -- Let's look on illustrative examples below. The following\n"
"    -- one works.\n"
"    --\n"
"    -- replicasets:\n"
"    --   r-001:\n"
"    --     replication:\n"
"    --       failover: election\n"
"    --     instances:\n"
"    --       i-001: {}       # candidate\n"
"    --       i-002: {}       # candidate\n"
"    --       i-003: {}       # candidate\n"
"    --       i-004:          # off --------+\n"
"    --         replication:  #             +--> OK\n"
"    --           anon: true  # anonymous --+\n"
"    --\n"
"    -- All the non-anonymous instances have effective default\n"
"    -- 'replication.election_mode: candidate', while anonymous\n"
"    -- replicas default to 'off'.\n"
"    --\n"
"    -- However, the following example doesn't work.\n"
"    --\n"
"    -- replicasets:\n"
"    --   r-001:\n"
"    --     replication:\n"
"    --       failover: election\n"
"    --       election_mode: candidate # !!\n"
"    --     instances:\n"
"    --       i-001: {}       # candidate\n"
"    --       i-002: {}       # candidate\n"
"    --       i-003: {}       # candidate\n"
"    --       i-004:          # candidate --+\n"
"    --         replication:  #             +--> error\n"
"    --           anon: true  # anonymous --+\n"
"    --\n"
"    -- The default 'off' is not applied, because the explicit\n"
"    -- 'candidate' value is set in the replicaset scope. It can be\n"
"    -- fixed like so:\n"
"    --\n"
"    -- <...>\n"
"    --       i-004:\n"
"    --         replication:\n"
"    --           anon: true\n"
"    --           election_mode: off # !!\n"
"    if failover == 'election' then\n"
"        for peer_name, peer in pairs(peers) do\n"
"            local is_anon = instance_config:get(peer.iconfig_def,\n"
"                'replication.anon')\n"
"            local election_mode = instance_config:get(peer.iconfig_def,\n"
"                'replication.election_mode')\n"
"            if is_anon and election_mode ~= nil and election_mode ~= 'off' then\n"
"                error(('replication.election_mode = %q is set for instance ' ..\n"
"                    '%q of replicaset %q of group %q, but this option ' ..\n"
"                    'cannot be used together with replication.anon = true; ' ..\n"
"                    'consider setting replication.election_mode = \"off\" ' ..\n"
"                    'explicitly for this instance'):format(\n"
"                    election_mode, peer_name, found.replicaset_name,\n"
"                    found.group_name), 0)\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function new(iconfig, cconfig, instance_name)\n"
"    -- Find myself in a cluster config, determine peers in the same\n"
"    -- replicaset.\n"
"    local found = cluster_config:find_instance(cconfig, instance_name)\n"
"    assert(found ~= nil)\n"
"\n"
"    -- Precalculate configuration with applied defaults.\n"
"    local iconfig_def = instance_config:apply_default(iconfig)\n"
"\n"
"    -- Substitute {{ instance_name }} with actual instance name in\n"
"    -- the original config and in the config with defaults.\n"
"    --\n"
"    -- The same for {{ replicaset_name }} and {{ group_name }}.\n"
"    local vars = {\n"
"        instance_name = instance_name,\n"
"        replicaset_name = found.replicaset_name,\n"
"        group_name = found.group_name,\n"
"    }\n"
"    iconfig = instance_config:apply_vars(iconfig, vars)\n"
"    iconfig_def = instance_config:apply_vars(iconfig_def, vars)\n"
"\n"
"    local replicaset_uuid = instance_config:get(iconfig_def,\n"
"        'database.replicaset_uuid')\n"
"    local instance_uuid = instance_config:get(iconfig_def,\n"
"        'database.instance_uuid')\n"
"\n"
"    -- Save instance configs of the peers from the same replicaset.\n"
"    local peers = {}\n"
"    for peer_name, _ in pairs(found.replicaset.instances) do\n"
"        -- Build config for each peer from the cluster config.\n"
"        -- Build a config with applied defaults as well.\n"
"        local peer_iconfig = cluster_config:instantiate(cconfig, peer_name)\n"
"        local peer_iconfig_def = instance_config:apply_default(peer_iconfig)\n"
"\n"
"        -- Substitute variables according to the instance name\n"
"        -- of the peer.\n"
"        --\n"
"        -- The replicaset and group names are same as for the\n"
"        -- current instance.\n"
"        local peer_vars = {\n"
"            instance_name = peer_name,\n"
"            replicaset_name = found.replicaset_name,\n"
"            group_name = found.group_name,\n"
"        }\n"
"        peer_iconfig = instance_config:apply_vars(peer_iconfig, peer_vars)\n"
"        peer_iconfig_def = instance_config:apply_vars(peer_iconfig_def,\n"
"            peer_vars)\n"
"\n"
"        peers[peer_name] = {\n"
"            iconfig = peer_iconfig,\n"
"            iconfig_def = peer_iconfig_def,\n"
"        }\n"
"    end\n"
"\n"
"    -- Make the order of the peers predictable and the same on all\n"
"    -- instances in the replicaset.\n"
"    local peer_names = fun.iter(peers):totable()\n"
"    table.sort(peer_names)\n"
"\n"
"    -- The replication.failover option is forbidden for the\n"
"    -- instance scope of the cluster config, so it is common for\n"
"    -- the whole replicaset. We can extract it from the\n"
"    -- configuration of the given instance.\n"
"    --\n"
"    -- There is a nuance: the option still can be set using an\n"
"    -- environment variable. We can't detect incorrect usage in\n"
"    -- this case (say, different failover modes for different\n"
"    -- instances in the same replicaset), because we have no\n"
"    -- access to environment of other instances.\n"
"    local failover = instance_config:get(iconfig_def, 'replication.failover')\n"
"    local leader = found.replicaset.leader\n"
"    validate_failover(found, peers, failover, leader)\n"
"\n"
"    local bootstrap_strategy = instance_config:get(iconfig_def,\n"
"        'replication.bootstrap_strategy')\n"
"    local bootstrap_leader = found.replicaset.bootstrap_leader\n"
"    if bootstrap_strategy ~= 'config' then\n"
"        if bootstrap_leader ~= nil then\n"
"            error(('The \"bootstrap_leader\" option cannot be set for '..\n"
"                   'replicaset %q because \"bootstrap_strategy\" for instance '..\n"
"                   '%q is not \"config\"'):format(found.replicaset_name,\n"
"                                                instance_name), 0)\n"
"        end\n"
"    elseif bootstrap_leader == nil then\n"
"        error(('The \"bootstrap_leader\" option cannot be empty for replicaset '..\n"
"               '%q because \"bootstrap_strategy\" for instance %q is '..\n"
"               '\"config\"'):format(found.replicaset_name, instance_name), 0)\n"
"    else\n"
"        if peers[bootstrap_leader] == nil then\n"
"            error(('\"bootstrap_leader\" = %q option is set for replicaset %q '..\n"
"                   'of group %q, but instance %q is not found in this '..\n"
"                   'replicaset'):format(bootstrap_leader, found.replicaset_name,\n"
"                                        found.group_name, bootstrap_leader), 0)\n"
"        end\n"
"    end\n"
"\n"
"    -- Verify that there is at least one non-anonymous replica in\n"
"    -- the given replicaset.\n"
"    --\n"
"    -- Verify that `replication.anon: true` (if any) doesn't\n"
"    -- conflict with any other option (say, database.mode,\n"
"    -- <replicaset>.leader or replication.election_mode).\n"
"    validate_anon(found, peers, failover, leader)\n"
"\n"
"\n"
"    -- Verify \"replication.failover\" = \"supervised\" strategy\n"
"    -- prerequisites.\n"
"    local bootstrap_leader_name\n"
"    if failover == 'supervised' then\n"
"        -- An instance that is potentially a bootstrap leader\n"
"        -- starts in RW in assumption that the bootstrap strategy\n"
"        -- will choose it as the bootstrap leader.\n"
"        --\n"
"        -- It doesn't work in at least 'config' and 'supervised'\n"
"        -- bootstrap strategies. It is possible to support them,\n"
"        -- but an extra logic that is not implemented yet is\n"
"        -- required.\n"
"        --\n"
"        -- See applier/box_cfg.lua for the details.\n"
"        if bootstrap_strategy ~= 'auto' then\n"
"            error(('\"bootstrap_strategy\" = %q is set for replicaset %q, but ' ..\n"
"                'it is not supported with \"replication.failover\" = ' ..\n"
"                '\"supervised\"'):format(bootstrap_strategy,\n"
"                found.replicaset_name), 0)\n"
"        end\n"
"        assert(bootstrap_leader == nil)\n"
"\n"
"        -- Choose first non-anonymous instance.\n"
"        for _, peer_name in ipairs(peer_names) do\n"
"            assert(peers[peer_name] ~= nil)\n"
"            local iconfig_def = peers[peer_name].iconfig_def\n"
"            local is_anon = instance_config:get(iconfig_def, 'replication.anon')\n"
"            if not is_anon then\n"
"                bootstrap_leader_name = peer_name\n"
"                break\n"
"            end\n"
"        end\n"
"        assert(bootstrap_leader_name ~= nil)\n"
"    end\n"
"\n"
"    -- Names and UUIDs are always validated: during instance start\n"
"    -- and during config reload.\n"
"    local saved_names = find_saved_names(iconfig_def)\n"
"    if saved_names ~= nil then\n"
"        local config_instance_uuid, config_replicaset_uuid =\n"
"            find_uuids_by_name(peers, instance_name)\n"
"        validate_names(saved_names, {\n"
"            replicaset_name = found.replicaset_name,\n"
"            instance_name = instance_name,\n"
"            -- UUIDs from config, generated one should not be used here.\n"
"            replicaset_uuid = config_replicaset_uuid,\n"
"            instance_uuid = config_instance_uuid,\n"
"            peers = peers,\n"
"        }, iconfig_def)\n"
"    end\n"
"\n"
"    return setmetatable({\n"
"        _iconfig = iconfig,\n"
"        _iconfig_def = iconfig_def,\n"
"        _cconfig = cconfig,\n"
"        _peer_names = peer_names,\n"
"        _replicaset_uuid = replicaset_uuid,\n"
"        _instance_uuid = instance_uuid,\n"
"        _peers = peers,\n"
"        _group_name = found.group_name,\n"
"        _replicaset_name = found.replicaset_name,\n"
"        _instance_name = instance_name,\n"
"        _failover = failover,\n"
"        _leader = leader,\n"
"        _bootstrap_leader = bootstrap_leader,\n"
"        _bootstrap_leader_name = bootstrap_leader_name,\n"
"    }, mt)\n"
"end\n"
"\n"
"return {\n"
"    new = new,\n"
"}\n"
""
;
