const char config_init_lua[] =
"local fio = require('fio')\n"
"local instance_config = require('internal.config.instance_config')\n"
"local cluster_config = require('internal.config.cluster_config')\n"
"local configdata = require('internal.config.configdata')\n"
"local log = require('internal.config.utils.log')\n"
"local tarantool = require('tarantool')\n"
"local datetime = require('datetime')\n"
"\n"
"-- Tarantool Enterprise Edition has its own additions\n"
"-- for config module.\n"
"--\n"
"-- Tarantool Community Edition may be extended using the same\n"
"-- mechanism.\n"
"local function load_extras()\n"
"    local has_extras, extras = pcall(require, 'internal.config.extras')\n"
"\n"
"    -- The module is built into Tarantool EE executable. If it\n"
"    -- can't be loaded, something is definitely going wrong.\n"
"    if not has_extras and tarantool.package == 'Tarantool Enterprise' then\n"
"        error('Tarantool Enterprise Edition build seems broken: built-in ' ..\n"
"            'module internal.config.extras is not found')\n"
"    end\n"
"\n"
"    -- The module may be provided by a user for Tarantool CE, but\n"
"    -- it is optional.\n"
"    if not has_extras then\n"
"        return\n"
"    end\n"
"\n"
"    -- Verify the contract.\n"
"    assert(type(extras) == 'table')\n"
"    assert(type(extras.initialize) == 'function')\n"
"    assert(type(extras.post_apply) == 'function')\n"
"\n"
"    return extras\n"
"end\n"
"\n"
"local extras = load_extras()\n"
"\n"
"-- {{{ Helpers\n"
"\n"
"-- Remove indent from a text.\n"
"--\n"
"-- Similar to Python's textwrap.dedent().\n"
"--\n"
"-- It strips all newlines from beginning and all whitespace\n"
"-- characters from the end for convenience use with multiline\n"
"-- string literals ([[ <...> ]]).\n"
"local function dedent(s)\n"
"    local lines = s:lstrip('\\n'):rstrip():split('\\n')\n"
"\n"
"    local indent = math.huge\n"
"    for _, line in ipairs(lines) do\n"
"        if #line ~= 0 then\n"
"            indent = math.min(indent, #line:match('^ *'))\n"
"        end\n"
"    end\n"
"\n"
"    local res = {}\n"
"    for _, line in ipairs(lines) do\n"
"        table.insert(res, line:sub(indent + 1))\n"
"    end\n"
"    return table.concat(res, '\\n')\n"
"end\n"
"\n"
"-- Extract all fields from a table except ones that start from\n"
"-- the underscore.\n"
"--\n"
"-- Useful for __serialize.\n"
"local function filter_out_private_fields(t)\n"
"    local res = {}\n"
"    for k, v in pairs(t) do\n"
"        if not k:startswith('_') then\n"
"            res[k] = v\n"
"        end\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"-- }}} Helpers\n"
"\n"
"local methods = {}\n"
"local mt = {\n"
"    __index = methods,\n"
"    __serialize = filter_out_private_fields,\n"
"}\n"
"\n"
"local function selfcheck(self, method_name)\n"
"    if type(self) ~= 'table' or getmetatable(self) ~= mt then\n"
"        local fmt_str = 'Use config:%s(<...>) instead of config.%s(<...>)'\n"
"        error(fmt_str:format(method_name, method_name), 0)\n"
"    end\n"
"end\n"
"\n"
"local function broadcast(self)\n"
"    box.broadcast('config.info', self:info())\n"
"end\n"
"\n"
"function methods._alert(self, key, alert)\n"
"    assert(alert.type == 'error' or alert.type == 'warn')\n"
"    if alert.type == 'error' then\n"
"        log.error(alert.message)\n"
"    else\n"
"        log.warn(alert.message)\n"
"    end\n"
"    alert.timestamp = datetime.now()\n"
"    self._alerts[key] = alert\n"
"end\n"
"\n"
"function methods._alert_drop(self, key)\n"
"    self._alerts[key] = nil\n"
"    if self._status == 'check_warnings' and table.equals(self._alerts, {}) then\n"
"        self._status = 'ready'\n"
"    end\n"
"end\n"
"\n"
"function methods._meta(self, source_name, key, value)\n"
"    local data = self._metadata[source_name] or {}\n"
"    data[key] = value\n"
"    self._metadata[source_name] = data\n"
"end\n"
"\n"
"function methods._register_source(self, source)\n"
"    assert(type(source) == 'table')\n"
"    assert(source.name ~= nil)\n"
"    assert(source.type == 'instance' or source.type == 'cluster')\n"
"    assert(source.sync ~= nil)\n"
"    assert(source.get ~= nil)\n"
"    table.insert(self._sources, source)\n"
"end\n"
"\n"
"function methods._register_applier(self, applier)\n"
"    assert(type(applier) == 'table')\n"
"    assert(applier.name ~= nil)\n"
"    assert(applier.apply ~= nil)\n"
"    table.insert(self._appliers, applier)\n"
"end\n"
"\n"
"function methods._initialize(self)\n"
"    -- The sources are synchronized in the order of registration:\n"
"    -- env, file, etcd (present in Tarantool EE), env for\n"
"    -- defaults.\n"
"    --\n"
"    -- The configuration values from the first source has highest\n"
"    -- priority. The menthal rule here is the following: values\n"
"    -- closer to the process are preferred: env first, then file,\n"
"    -- then etcd (if available). And only then the env source with\n"
"    -- defaults.\n"
"    self:_register_source(require('internal.config.source.env').new())\n"
"\n"
"    if self._config_file ~= nil then\n"
"        self:_register_source(require('internal.config.source.file').new())\n"
"    end\n"
"\n"
"    self:_register_applier(require('internal.config.applier.compat'))\n"
"    self:_register_applier(require('internal.config.applier.mkdir'))\n"
"    self:_register_applier(require('internal.config.applier.console'))\n"
"    self:_register_applier(require('internal.config.applier.box_cfg'))\n"
"    self:_register_applier(require('internal.config.applier.credentials'))\n"
"    self:_register_applier(require('internal.config.applier.fiber'))\n"
"    self:_register_applier(require('internal.config.applier.sharding'))\n"
"    self:_register_applier(require('internal.config.applier.roles'))\n"
"    self:_register_applier(require('internal.config.applier.app'))\n"
"\n"
"    if extras ~= nil then\n"
"        extras.initialize(self)\n"
"    end\n"
"\n"
"    self:_register_source(require('internal.config.source.env').new({\n"
"        env_var_suffix = 'default',\n"
"    }))\n"
"end\n"
"\n"
"-- Collect data from configuration sources.\n"
"--\n"
"-- Return instance config, cluster config and sources information.\n"
"--\n"
"-- This method doesn't store anything in the config object.\n"
"--\n"
"-- It can be used with self._instance_name == nil to obtain a\n"
"-- cluster configuration. All the instance_name related checks are\n"
"-- in the self:_store() method.\n"
"function methods._collect(self, opts)\n"
"    assert(type(opts) == 'table')\n"
"    local sync_source = opts.sync_source\n"
"\n"
"    local iconfig = {}\n"
"    local cconfig = {}\n"
"\n"
"    -- For error reporting.\n"
"    local source_info = {}\n"
"\n"
"    for _, source in ipairs(self._sources) do\n"
"        -- Gather config values.\n"
"        --\n"
"        -- The configdata object is not constructed yet, so we\n"
"        -- pass currently collected instance config as the second\n"
"        -- argument. The 'config' section of the config may\n"
"        -- contain a configuration needed for a source.\n"
"        if sync_source == source.name or sync_source == 'all' then\n"
"            source:sync(self, iconfig)\n"
"        end\n"
"\n"
"        -- Validate configurations gathered from the sources.\n"
"        if source.type == 'instance' then\n"
"            instance_config:validate(source:get())\n"
"        elseif source.type == 'cluster' then\n"
"            cluster_config:validate(source:get())\n"
"        else\n"
"            assert(false)\n"
"        end\n"
"\n"
"        -- Merge configurations from the sources.\n"
"        --\n"
"        -- Instantiate a cluster config to an instance config for\n"
"        -- cluster config sources.\n"
"        --\n"
"        -- The configuration values from a first source has highest\n"
"        -- priority. We should keep already gathered values in the\n"
"        -- accumulator and fill only missed ones from next sources.\n"
"        --\n"
"        -- :merge() prefers values from the second argument, so the\n"
"        -- accumulator is passed as the second.\n"
"        local source_iconfig\n"
"        if source.type == 'cluster' then\n"
"            local source_cconfig = source:get()\n"
"\n"
"            -- Extract and merge conditional sections into the\n"
"            -- data from the source.\n"
"            --\n"
"            -- It is important to call :apply_conditional() before\n"
"            -- :merge() for each cluster config.\n"
"            --\n"
"            -- The 'conditional' field is an array and if several\n"
"            -- sources have the field, the last one replaces all\n"
"            -- the previous ones. If we merge all the configs and\n"
"            -- then call :apply_conditional(), we loss all the\n"
"            -- conditional sections except the last one.\n"
"            --\n"
"            -- The same is applicable for config sources that\n"
"            -- construct a config from several separately stored\n"
"            -- parts using :merge(). They should call\n"
"            -- :apply_conditional() for each of the parts before\n"
"            -- :merge().\n"
"            source_cconfig = cluster_config:apply_conditional(source_cconfig)\n"
"\n"
"            cconfig = cluster_config:merge(source_cconfig, cconfig)\n"
"            source_iconfig = cluster_config:instantiate(cconfig,\n"
"                self._instance_name)\n"
"        elseif source.type == 'instance' then\n"
"            source_iconfig = source:get()\n"
"        else\n"
"            assert(false)\n"
"        end\n"
"        iconfig = instance_config:merge(source_iconfig, iconfig)\n"
"\n"
"        -- If a source returns an empty table, mark it as ones\n"
"        -- that provide no data.\n"
"        local has_data = next(source:get()) ~= nil\n"
"        table.insert(source_info, ('* %q [type: %s]%s'):format(source.name,\n"
"            source.type, has_data and '' or ' (no data)'))\n"
"    end\n"
"\n"
"    return iconfig, cconfig, source_info\n"
"end\n"
"\n"
"-- Store the given configuration in the config object.\n"
"--\n"
"-- The method accepts collected instance config, cluster config\n"
"-- and configuration source information and based on the given\n"
"-- values performs several checks and stores the configuration\n"
"-- within the config object as a configdata object.\n"
"function methods._store(self, iconfig, cconfig, source_info)\n"
"    if next(cconfig) == nil then\n"
"        local is_reload = self._status == 'reload_in_progress'\n"
"        local action = is_reload and 'Reload' or 'Startup'\n"
"        local source_info_str = table.concat(source_info, '\\n')\n"
"        error(dedent([[\n"
"            %s failure.\n"
"\n"
"            No cluster config received from the given configuration sources.\n"
"\n"
"            %s\n"
"\n"
"            The %q instance cannot find itself in the group/replicaset/instance\n"
"            topology and it is unknown, whether it should join a replicaset or\n"
"            create its own database.\n"
"\n"
"            Recipes:\n"
"\n"
"            * Use --config <file> command line option.\n"
"            * Use TT_CONFIG_ETCD_* environment variables (available on Tarantool\n"
"              Enterprise Edition).\n"
"        ]]):format(action, source_info_str, self._instance_name), 0)\n"
"    end\n"
"\n"
"    if cluster_config:find_instance(cconfig, self._instance_name) == nil then\n"
"        local is_reload = self._status == 'reload_in_progress'\n"
"        local action = is_reload and 'Reload' or 'Startup'\n"
"        local source_info_str = table.concat(source_info, '\\n')\n"
"        error(dedent([[\n"
"            %s failure.\n"
"\n"
"            Unable to find instance %q in the group/replicaset/instance\n"
"            topology provided by the given cluster configuration sources.\n"
"\n"
"            %s\n"
"\n"
"            It is unknown, whether the instance should join a replicaset or\n"
"            create its own database.\n"
"\n"
"            Minimal cluster config:\n"
"\n"
"            groups:\n"
"              group-001:\n"
"                replicasets:\n"
"                  replicaset-001:\n"
"                    instances:\n"
"                      instance-001: {}\n"
"        ]]):format(action, self._instance_name, source_info_str), 0)\n"
"    end\n"
"\n"
"    self._configdata = configdata.new(iconfig, cconfig, self._instance_name)\n"
"end\n"
"\n"
"-- Invoke compat, mkdir, console and box_cfg appliers at the first\n"
"-- phase. Invoke all the other ones at the second phase.\n"
"function methods._apply_on_startup(self, opts)\n"
"    local first_phase_appliers = {\n"
"        compat = true,\n"
"        mkdir = true,\n"
"        console = true,\n"
"        box_cfg = true,\n"
"    }\n"
"\n"
"    assert(type(opts) == 'table')\n"
"    local phase = opts.phase\n"
"\n"
"    if phase == 1 then\n"
"        local needs_retry = false\n"
"        for _, applier in ipairs(self._appliers) do\n"
"            if first_phase_appliers[applier.name] then\n"
"                local res = applier.apply(self)\n"
"                if res ~= nil and res.needs_retry then\n"
"                    needs_retry = true\n"
"                end\n"
"            end\n"
"        end\n"
"        return needs_retry\n"
"    elseif phase == 2 then\n"
"        for _, applier in ipairs(self._appliers) do\n"
"            if not first_phase_appliers[applier.name] then\n"
"                applier.apply(self)\n"
"            end\n"
"        end\n"
"\n"
"        self._configdata_applied = self._configdata\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"function methods._apply(self)\n"
"    for _, applier in ipairs(self._appliers) do\n"
"        applier.apply(self)\n"
"    end\n"
"\n"
"    self._configdata_applied = self._configdata\n"
"end\n"
"\n"
"function methods._post_apply(self)\n"
"    for _, applier in ipairs(self._appliers) do\n"
"        if applier.post_apply ~= nil then\n"
"            applier.post_apply(self)\n"
"        end\n"
"    end\n"
"\n"
"    if extras ~= nil then\n"
"        extras.post_apply(self)\n"
"    end\n"
"end\n"
"\n"
"-- Set proper status depending on received alerts.\n"
"function methods._set_status_based_on_alerts(self)\n"
"    local status = 'ready'\n"
"    for _, alert in pairs(self._alerts) do\n"
"        assert(alert.type == 'error' or alert.type == 'warn')\n"
"        if alert.type == 'error' then\n"
"            status = 'check_errors'\n"
"            break\n"
"        end\n"
"        status = 'check_warnings'\n"
"    end\n"
"    self._status = status\n"
"    broadcast(self)\n"
"end\n"
"\n"
"function methods._startup(self, instance_name, config_file)\n"
"    assert(self._status == 'uninitialized')\n"
"\n"
"    local ok, err = cluster_config:validate_name(instance_name)\n"
"    if not ok then\n"
"        error(('[--name] %s'):format(err), 0)\n"
"    end\n"
"\n"
"    self._status = 'startup_in_progress'\n"
"    self._instance_name = instance_name\n"
"    -- box.cfg() changes a current working directory that\n"
"    -- invalidates relative paths.\n"
"    --\n"
"    -- Let's calculate an absolute path to access the file later\n"
"    -- on :reload().\n"
"    if config_file ~= nil then\n"
"        self._config_file = fio.abspath(config_file)\n"
"    end\n"
"    broadcast(self)\n"
"\n"
"    self:_initialize()\n"
"\n"
"    -- Startup phase 1/2.\n"
"    --\n"
"    -- Start compat, mkdir, console and box_cfg appliers. The\n"
"    -- latter may force the read-only mode on this phase.\n"
"    --\n"
"    -- This phase may take a long time.\n"
"    self:_store(self:_collect({sync_source = 'all'}))\n"
"    local needs_retry = self:_apply_on_startup({phase = 1})\n"
"\n"
"    -- Startup phase 2/2.\n"
"    --\n"
"    -- If the previous phase is considered as potentially taking a\n"
"    -- long time, re-read the configuration and apply the fresh\n"
"    -- one.\n"
"    --\n"
"    -- Otherwise finish applying the existing configuration.\n"
"    if needs_retry then\n"
"        self:_store(self:_collect({sync_source = 'all'}))\n"
"        self:_apply()\n"
"    else\n"
"        self:_apply_on_startup({phase = 2})\n"
"    end\n"
"    self:_set_status_based_on_alerts()\n"
"\n"
"    self:_post_apply()\n"
"    self:_set_status_based_on_alerts()\n"
"end\n"
"\n"
"function methods._print_env_list(self)\n"
"    local env_source = require('internal.config.source.env').new()\n"
"    io.stdout:write(env_source:_env_list())\n"
"end\n"
"\n"
"function methods.get(self, path)\n"
"    selfcheck(self, 'get')\n"
"    if self._status == 'uninitialized' then\n"
"        assert(self._configdata_applied == nil)\n"
"        error('config:get(): no instance config available yet', 0)\n"
"    end\n"
"    return self._configdata_applied:get(path, {use_default = true})\n"
"end\n"
"\n"
"function methods._reload_noexc(self, opts)\n"
"    assert(type(opts) == 'table')\n"
"    if self._status == 'uninitialized' then\n"
"        return false, 'config:reload(): no instance config available yet'\n"
"    end\n"
"    if self._status == 'startup_in_progress' or\n"
"       self._status == 'reload_in_progress' then\n"
"        return false, 'config:reload(): instance configuration is already in '..\n"
"                      'progress'\n"
"    end\n"
"    self._status = 'reload_in_progress'\n"
"    broadcast(self)\n"
"\n"
"    self._alerts = {}\n"
"    self._metadata = {}\n"
"\n"
"    local ok, err = pcall(function(opts)\n"
"        self:_store(self:_collect(opts))\n"
"        self:_apply()\n"
"        self:_set_status_based_on_alerts()\n"
"        self:_post_apply()\n"
"    end, opts)\n"
"\n"
"    assert(not ok or err == nil)\n"
"    if not ok then\n"
"        self:_alert('reload_error', {type = 'error', message = err})\n"
"    end\n"
"\n"
"    self:_set_status_based_on_alerts()\n"
"\n"
"    return ok, err\n"
"end\n"
"\n"
"function methods.reload(self)\n"
"    selfcheck(self, 'reload')\n"
"    local ok, err = self:_reload_noexc({sync_source = 'all'})\n"
"    if not ok then\n"
"        error(err, 0)\n"
"    end\n"
"end\n"
"\n"
"function methods.info(self)\n"
"    selfcheck(self, 'info')\n"
"    -- Don't return alert keys from config:info().\n"
"    local alerts = {}\n"
"    for _, alert in pairs(self._alerts) do\n"
"        table.insert(alerts, alert)\n"
"    end\n"
"    table.sort(alerts, function(a, b)\n"
"        return a.timestamp < b.timestamp\n"
"    end)\n"
"\n"
"    return {\n"
"        alerts = alerts,\n"
"        meta = self._metadata,\n"
"        status = self._status,\n"
"    }\n"
"end\n"
"\n"
"-- The object is a singleton. The constructor should be called\n"
"-- only once.\n"
"local function new()\n"
"    return setmetatable({\n"
"        _sources = {},\n"
"        _appliers = {},\n"
"        -- There are values the module need to hold, which are not\n"
"        -- part of the configuration. They're stored here.\n"
"        _instance_name = nil,\n"
"        _config_file = nil,\n"
"        -- Collected config values.\n"
"        _configdata = nil,\n"
"        -- Track applied config values as well.\n"
"        _configdata_applied = nil,\n"
"        -- Track situations when something is going wrong.\n"
"        _alerts = {},\n"
"        -- Metadata from sources.\n"
"        _metadata = {},\n"
"        -- Current status.\n"
"        _status = 'uninitialized',\n"
"    }, mt)\n"
"end\n"
"\n"
"return new()\n"
""
;
