const char config_instance_config_lua[] =
"local schema = require('internal.config.utils.schema')\n"
"local tarantool = require('tarantool')\n"
"local compat = require('compat')\n"
"local uuid = require('uuid')\n"
"local urilib = require('uri')\n"
"local fio = require('fio')\n"
"local file = require('internal.config.utils.file')\n"
"\n"
"-- List of annotations:\n"
"--\n"
"-- * enterprise_edition (boolean)\n"
"--\n"
"--   Available only in Tarantool Enterprise Edition.\n"
"--\n"
"-- * default (any)\n"
"--\n"
"--   Default value.\n"
"--\n"
"-- * scope ('global', 'group', 'replicaset', 'instance')\n"
"--\n"
"--   A place of an instance config option in the cluster config\n"
"--   hierarchy.\n"
"--\n"
"-- * box_cfg (string)\n"
"--\n"
"--   A name of the corresponding box.cfg() option.\n"
"--\n"
"-- * box_cfg_nondynamic (boolean)\n"
"--\n"
"--   `true` if the option can only be set at first box.cfg() call\n"
"--   and cannot be changed by a subsequent box.cfg() call.\n"
"--\n"
"-- * allowed_values\n"
"--\n"
"--   A list of allowed values.\n"
"--\n"
"-- * mkdir (boolean)\n"
"--\n"
"--   Create the given directory before box.cfg().\n"
"--\n"
"-- * mk_parent_dir (boolean)\n"
"--\n"
"--   Create a parent directory for the given file before box.cfg().\n"
"\n"
"-- Verify that replication.failover option is not present in the\n"
"-- instance scope of the cluster config.\n"
"local function validate_replication_failover(data, w)\n"
"    -- scope == nil means that it is the instance config.\n"
"    local scope = w.schema.scope\n"
"\n"
"    -- There is no much sense to set the failover option for a\n"
"    -- particular instance, not the whole replicaset. So, the\n"
"    -- option is forbidden for the instance scope of the cluster\n"
"    -- config.\n"
"    --\n"
"    -- However, it is allowed for the instance config to accept an\n"
"    -- instantiated cluster config as valid.\n"
"    if scope ~= 'instance' then\n"
"        return\n"
"    end\n"
"\n"
"    if data.replication ~= nil and data.replication.failover ~= nil then\n"
"        w.error('replication.failover must not be present in the %s scope',\n"
"            scope)\n"
"    end\n"
"end\n"
"\n"
"local function validate_outmost_record(data, w)\n"
"    validate_replication_failover(data, w)\n"
"end\n"
"\n"
"local function enterprise_edition_validate(data, w)\n"
"    -- OK if we're on Tarantool EE.\n"
"    if tarantool.package == 'Tarantool Enterprise' then\n"
"        return\n"
"    end\n"
"\n"
"    assert(tarantool.package == 'Tarantool')\n"
"\n"
"    -- OK, if the value is nil or box.NULL.\n"
"    if data == nil then\n"
"        return\n"
"    end\n"
"\n"
"    -- NB: Let's fail the validation for an empty table, because\n"
"    -- otherwise we will get a less descriptive error from a lower\n"
"    -- level API. For example, box.cfg({wal_ext = {}}) on Tarantool\n"
"    -- Community Edition says the following:\n"
"    --\n"
"    -- > Incorrect value for option 'wal_ext': unexpected option\n"
"\n"
"    w.error('This configuration parameter is available only in Tarantool ' ..\n"
"        'Enterprise Edition')\n"
"end\n"
"\n"
"local function enterprise_edition_apply_default_if(_data, _w)\n"
"    return tarantool.package == 'Tarantool Enterprise'\n"
"end\n"
"\n"
"-- Generate a function that calls two given functions in a row.\n"
"--\n"
"-- If one of the arguments is nil, return the other one.\n"
"--\n"
"-- If both arguments are nil, returns nil.\n"
"local function chain2(f1, f2)\n"
"    if f1 == nil then\n"
"        return f2\n"
"    end\n"
"    if f2 == nil then\n"
"        return f1\n"
"    end\n"
"    return function(...)\n"
"        f1(...)\n"
"        f2(...)\n"
"    end\n"
"end\n"
"\n"
"-- Available only in Tarantool Enterprise Edition.\n"
"local function enterprise_edition(schema_node)\n"
"    schema_node.enterprise_edition = true\n"
"\n"
"    -- Perform a domain specific validation first and only then\n"
"    -- check, whether the option is to be used with Tarantool\n"
"    -- Enterprise Edition.\n"
"    --\n"
"    -- This order is consistent with data type validation, which\n"
"    -- is also performed before the EE check.\n"
"    schema_node.validate = chain2(\n"
"        schema_node.validate,\n"
"        enterprise_edition_validate)\n"
"    schema_node.apply_default_if = chain2(\n"
"        schema_node.apply_default_if,\n"
"        enterprise_edition_apply_default_if)\n"
"\n"
"    return schema_node\n"
"end\n"
"\n"
"local function validate_uuid_str(data, w)\n"
"    if uuid.fromstr(data) == nil then\n"
"        w.error('Unable to parse the value as a UUID: %q', data)\n"
"    end\n"
"    if data == uuid.NULL:str() then\n"
"        w.error('nil UUID is reserved')\n"
"    end\n"
"end\n"
"\n"
"-- Accepts an uri object (one produced by urilib.parse()).\n"
"--\n"
"-- Performs several checks regarding ability to use the URI to\n"
"-- create a client socket. IOW, to call connect() on it.\n"
"--\n"
"-- The function returns `true` if the URI is OK to connect and\n"
"-- `false, err` otherwise.\n"
"--\n"
"-- If the URI doesn't fit the given criteria an error is raised.\n"
"-- The function returns nothing otherwise.\n"
"--\n"
"-- The following checks are performed:\n"
"--\n"
"-- * INADDR_ANY IPv4 address (0.0.0.0) or in6addr_any IPv6 address\n"
"--   (::) in the host part of the URI.\n"
"--\n"
"--   It means 'bind to all interfaces' for the bind() call, but it\n"
"--   has no meaning at the connect() call on a client.\n"
"-- * Zero TCP port (service part of the URI).\n"
"--\n"
"--   It means 'bind to a random free port' for the bind() call,\n"
"--   but it has no meaning at the connect() call on a client.\n"
"local function uri_is_suitable_to_connect(uri)\n"
"    assert(uri ~= nil)\n"
"\n"
"    if uri.ipv4 == '0.0.0.0' then\n"
"        return false, 'INADDR_ANY (0.0.0.0) cannot be used to create ' ..\n"
"            'a client socket'\n"
"    end\n"
"    if uri.ipv6 == '::' then\n"
"        return false, 'in6addr_any (::) cannot be used to create a client ' ..\n"
"            'socket'\n"
"    end\n"
"    if uri.service == '0' then\n"
"        return false, 'An URI with zero port cannot be used to create ' ..\n"
"            'a client socket'\n"
"    end\n"
"\n"
"    return true\n"
"end\n"
"\n"
"-- Verify 'uri' field in iproto.listen and iproto.advertise.* options.\n"
"local function validate_uri_field(has_login_field, used_to_connect)\n"
"    return function(data, w)\n"
"        -- Substitute variables with placeholders to don't confuse\n"
"        -- the URI parser with the curly brackets.\n"
"        data = data:gsub('{{ *.- *}}', 'placeholder')\n"
"        local uri, err = urilib.parse(data)\n"
"        if uri == nil then\n"
"            if data:find(',') then\n"
"                w.error('A single URI is expected, not a list of URIs')\n"
"            end\n"
"            w.error('Unable to parse an URI: %s', err)\n"
"        end\n"
"        if uri.login ~= nil and has_login_field then\n"
"            w.error(\"Login must be set via the 'login' option\")\n"
"        end\n"
"        if uri.login ~= nil and not has_login_field then\n"
"            w.error('Login cannot be set for as part of the URI')\n"
"        end\n"
"        assert(uri.password == nil)\n"
"        if uri.params ~= nil then\n"
"            w.error(\"URI parameters should be described in the 'params' \" ..\n"
"                \"field, not as the part of URI\")\n"
"        end\n"
"        if used_to_connect then\n"
"            local ok, err = uri_is_suitable_to_connect(uri)\n"
"            if not ok then\n"
"                w.error(err)\n"
"            end\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"-- Accept a value of 'iproto.listen' option and return the first URI\n"
"-- that is suitable to create a client socket (not just to listen\n"
"-- on the server as, say, 0.0.0.0:3301 or localhost:0).\n"
"--\n"
"-- See the uri_is_suitable_to_connect() method in the instance\n"
"-- schema object for details.\n"
"local function find_suitable_uri_to_connect(listen)\n"
"    for _, u in ipairs(listen) do\n"
"        assert(u.uri ~= nil)\n"
"        local uri = urilib.parse(u.uri)\n"
"        if uri ~= nil and uri_is_suitable_to_connect(uri) then\n"
"            -- The urilib.format() call has the second optional\n"
"            -- argument `write_password`. Let's assume that the\n"
"            -- given URIs are to listen on them and so have no\n"
"            -- user/password.\n"
"            --\n"
"            -- NB: We need to format the URI back to construct the\n"
"            -- 'uri' field. urilib.parse() creates separate 'host'\n"
"            -- and 'service'.\n"
"            return urilib.format(uri), u.params\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"local function find_password(self, iconfig, username)\n"
"    -- The guest user can't have a password.\n"
"    if username == 'guest' then\n"
"        return nil\n"
"    end\n"
"\n"
"    -- Find a user definition in the config.\n"
"    local user_def = self:get(iconfig, 'credentials.users.' .. username)\n"
"    if user_def == nil then\n"
"        error(('Cannot find user %s in the config to use its password in a '..\n"
"               'replication peer URI'):format(username), 0)\n"
"    end\n"
"\n"
"    -- There is a user definition without a password. Let's assume\n"
"    -- that the user has no password.\n"
"    if user_def.password ~= nil then\n"
"        return user_def.password\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"local function instance_uri(self, iconfig, advertise_type)\n"
"    assert(advertise_type == 'peer' or advertise_type == 'sharding')\n"
"\n"
"    -- An effective value of iproto.advertise.sharding defaults to\n"
"    -- iproto.advertise.peer.\n"
"    local uri\n"
"    if advertise_type == 'sharding' then\n"
"        uri = self:get(iconfig, 'iproto.advertise.sharding')\n"
"        if uri == nil then\n"
"            advertise_type = 'peer'\n"
"        end\n"
"    end\n"
"    if advertise_type == 'peer' then\n"
"        uri = self:get(iconfig, 'iproto.advertise.peer')\n"
"    end\n"
"    -- If there is no uri field, look it in iproto.listen.\n"
"    uri = uri ~= nil and uri or {}\n"
"    if uri.uri == nil then\n"
"        assert(uri.params == nil)\n"
"        local listen = self:get(iconfig, 'iproto.listen')\n"
"        if listen == nil then\n"
"            return nil\n"
"        end\n"
"        uri = table.copy(uri)\n"
"        uri.uri, uri.params = find_suitable_uri_to_connect(listen)\n"
"    end\n"
"    -- No URI found for the given instance.\n"
"    if uri.uri == nil then\n"
"        return nil\n"
"    end\n"
"\n"
"    -- If there is a login. but there are no password: lookup the\n"
"    -- credentials section.\n"
"    if uri.password == nil and uri.login ~= nil then\n"
"        uri = table.copy(uri)\n"
"        uri.password = find_password(self, iconfig, uri.login)\n"
"    end\n"
"    return uri\n"
"end\n"
"\n"
"local function apply_vars_f(data, w, vars)\n"
"    if w.schema.type == 'string' and data ~= nil then\n"
"        assert(type(data) == 'string')\n"
"        return (data:gsub('{{ *(.-) *}}', function(var_name)\n"
"            if vars[var_name] ~= nil then\n"
"                return vars[var_name]\n"
"            end\n"
"            w.error(('Unknown variable %q'):format(var_name))\n"
"        end))\n"
"    end\n"
"    return data\n"
"end\n"
"\n"
"-- The file path is interpreted as relative to `process.work_dir`.\n"
"-- The first box.cfg() call sets a current working directory to this path.\n"
"--\n"
"-- However, we should prepend paths manually before the first\n"
"-- box.cfg() call.\n"
"--\n"
"-- This function returns the prefix to add into config's paths.\n"
"--\n"
"-- TODO: Glue all such code from applier/mkdir.lua and\n"
"-- utils/snapshot.lua.\n"
"local function base_dir(self, iconfig)\n"
"    local work_dir = self:get(iconfig, 'process.work_dir')\n"
"    return type(box.cfg) == 'function' and work_dir or nil\n"
"end\n"
"\n"
"-- Interpret the given path as relative to the given base\n"
"-- directory.\n"
"local function rebase_file_path(base_dir, path)\n"
"    -- fio.pathjoin('/foo', '/bar') gives '/foo/bar/', see\n"
"    -- gh-8816.\n"
"    --\n"
"    -- Let's check whether the second path is absolute.\n"
"    local needs_prepending = base_dir ~= nil and not path:startswith('/')\n"
"    if needs_prepending then\n"
"        path = fio.pathjoin(base_dir, path)\n"
"    end\n"
"\n"
"    -- Let's consider the following example.\n"
"    --\n"
"    -- config:\n"
"    --   context:\n"
"    --     foo:\n"
"    --       from: file\n"
"    --       file: ../foo.txt\n"
"    -- process:\n"
"    --   work_dir: x\n"
"    --\n"
"    -- In such a case we get correct \"x/../foo.txt\" on\n"
"    -- startup, but fio.open() complains if there is no \"x\"\n"
"    -- directory.\n"
"    --\n"
"    -- The working directory is created at startup if needed,\n"
"    -- so a user is not supposed to create it manually before\n"
"    -- first tarantool startup.\n"
"    --\n"
"    -- Let's strip all these \"..\" path components using\n"
"    -- fio.abspath().\n"
"    path = fio.abspath(path)\n"
"\n"
"    -- However, if possible, use a relative path. Sometimes an\n"
"    -- absolute path overruns the Unix socket path length limit\n"
"    -- (107 on Linux and 103 on Mac OS -- in our case).\n"
"    --\n"
"    -- Using a relative path allows to listen on the given Unix\n"
"    -- socket path in such cases.\n"
"    local cwd = fio.cwd()\n"
"    if cwd == nil then\n"
"        return path\n"
"    end\n"
"    if not cwd:endswith('/') then\n"
"        cwd = cwd .. '/'\n"
"    end\n"
"    if path:startswith(cwd) then\n"
"        path = path:sub(#cwd + 1, #path)\n"
"    end\n"
"    return path\n"
"end\n"
"\n"
"-- Interpret the given path as relative to the given working\n"
"-- directory from the config.\n"
"--\n"
"-- This function takes care to the chdir performed by the first\n"
"-- box.cfg() call if the working directory is set.\n"
"local function prepare_file_path(self, iconfig, path)\n"
"    local base_dir = self:base_dir(iconfig)\n"
"    return rebase_file_path(base_dir, path)\n"
"end\n"
"\n"
"-- Read a config.context[name] variable depending of its \"from\"\n"
"-- type.\n"
"local function read_context_var_noexc(base_dir, def)\n"
"    if def.from == 'env' then\n"
"        local value = os.getenv(def.env)\n"
"        if value == nil then\n"
"            return false, ('no %q environment variable'):format(def.env)\n"
"        end\n"
"        return true, value\n"
"    elseif def.from == 'file' then\n"
"        local path = rebase_file_path(base_dir, def.file)\n"
"        return pcall(file.universal_read, path, 'file')\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"local function apply_vars(self, iconfig, vars)\n"
"    vars = table.copy(vars)\n"
"\n"
"    local base_dir = self:base_dir(iconfig)\n"
"\n"
"    -- Read config.context.* variables and add them into the\n"
"    -- variables list.\n"
"    local context_vars = self:get(iconfig, 'config.context')\n"
"    for name, def in pairs(context_vars or {}) do\n"
"        local ok, res = read_context_var_noexc(base_dir, def)\n"
"        if not ok then\n"
"            error(('Unable to read config.context.%s variable value: ' ..\n"
"                '%s'):format(name, res), 0)\n"
"        end\n"
"        if def.rstrip then\n"
"            res = res:rstrip()\n"
"        end\n"
"        vars['context.' .. name] = res\n"
"    end\n"
"\n"
"    return self:map(iconfig, apply_vars_f, vars)\n"
"end\n"
"\n"
"local function feedback_apply_default_if(_data, _w)\n"
"    return box.internal.feedback_daemon ~= nil\n"
"end\n"
"\n"
"local function feedback_validate(data, w)\n"
"    if data == nil or box.internal.feedback_daemon ~= nil then\n"
"        return\n"
"    end\n"
"    w.error('Tarantool is built without feedback reports sending support')\n"
"end\n"
"\n"
"return schema.new('instance_config', schema.record({\n"
"    config = schema.record({\n"
"        reload = schema.enum({\n"
"            'auto',\n"
"            'manual',\n"
"        }, {\n"
"            default = 'auto',\n"
"        }),\n"
"        -- Defaults can't be set there, because the `validate`\n"
"        -- annotation expects either no data or data with existing\n"
"        -- prefix field. The prefix field has no default. So,\n"
"        -- applying defaults to an empty data would make the data\n"
"        -- invalid.\n"
"        etcd = enterprise_edition(schema.record({\n"
"            prefix = schema.scalar({\n"
"                type = 'string',\n"
"                validate = function(data, w)\n"
"                    if not data:startswith('/') then\n"
"                        w.error(('config.etcd.prefix should be a path alike ' ..\n"
"                            'value, got %q'):format(data))\n"
"                    end\n"
"                end,\n"
"            }),\n"
"            endpoints = schema.array({\n"
"                items = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"            }),\n"
"            username = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            password = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            http = schema.record({\n"
"                request = schema.record({\n"
"                    timeout = schema.scalar({\n"
"                        type = 'number',\n"
"                        -- default = 0.3 is applied right in the\n"
"                        -- etcd source. See a comment above\n"
"                        -- regarding defaults in config.etcd.\n"
"                    }),\n"
"                    unix_socket = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                }),\n"
"            }),\n"
"            ssl = schema.record({\n"
"                ssl_key = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                ca_path = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                ca_file = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                verify_peer = schema.scalar({\n"
"                    type = 'boolean',\n"
"                }),\n"
"                verify_host = schema.scalar({\n"
"                    type = 'boolean',\n"
"                }),\n"
"            }),\n"
"        }, {\n"
"            validate = function(data, w)\n"
"                -- No config.etcd section at all -- OK.\n"
"                if data == nil or next(data) == nil then\n"
"                    return\n"
"                end\n"
"                -- There is some data -- the prefix should be there.\n"
"                if data.prefix == nil then\n"
"                    w.error('No config.etcd.prefix provided')\n"
"                end\n"
"            end,\n"
"        })),\n"
"        storage = enterprise_edition(schema.record({\n"
"            prefix = schema.scalar({\n"
"                type = 'string',\n"
"                validate = function(data, w)\n"
"                    if not data:startswith('/') then\n"
"                        w.error(('config.storage.prefix should be ' ..\n"
"                            'a path alike value, got %q'):format(data))\n"
"                    end\n"
"                end,\n"
"            }),\n"
"            endpoints = schema.array({\n"
"                items = schema.record({\n"
"                    uri = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                    login = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                    password = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                    params = schema.record({\n"
"                        transport = schema.enum({\n"
"                            'plain',\n"
"                            'ssl',\n"
"                        }),\n"
"                        ssl_key_file = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                        ssl_cert_file = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                        ssl_ca_file = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                        ssl_ciphers = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                        ssl_password = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                        ssl_password_file = enterprise_edition(schema.scalar({\n"
"                            type = 'string',\n"
"                        })),\n"
"                    }),\n"
"                }),\n"
"                validate = function(data, w)\n"
"                    if #data == 0 then\n"
"                        w.error('At least one endpoint must be' ..\n"
"                            'specified in config.storage.endpoints')\n"
"                    end\n"
"                end,\n"
"            }),\n"
"            timeout = schema.scalar({\n"
"                type = 'number',\n"
"                default = 3,\n"
"            }),\n"
"            reconnect_after = schema.scalar({\n"
"                type = 'number',\n"
"                default = 3,\n"
"            })\n"
"        }, {\n"
"            validate = function(data, w)\n"
"                if data == nil or next(data) == nil then\n"
"                    return\n"
"                end\n"
"                if data.prefix == nil and data.endpoints == nil then\n"
"                    return\n"
"                end\n"
"                if data.prefix == nil then\n"
"                    w.error('No config.storage.prefix provided')\n"
"                end\n"
"                if data.endpoints == nil then\n"
"                    w.error('No config.storage.endpoints provided')\n"
"                end\n"
"            end\n"
"        })),\n"
"        context = schema.map({\n"
"            key = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            value = schema.record({\n"
"                from = schema.enum({\n"
"                    'env',\n"
"                    'file',\n"
"                }),\n"
"                env = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                file = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                rstrip = schema.scalar({\n"
"                    type = 'boolean',\n"
"                }, {\n"
"                    default = false,\n"
"                }),\n"
"            }, {\n"
"                validate = function(var, w)\n"
"                    if var.from == nil then\n"
"                        w.error('\"from\" field must be defined in a context ' ..\n"
"                            'variable definition')\n"
"                    end\n"
"                    if var.from == 'env' and var.env == nil then\n"
"                        w.error('\"env\" field must define an environment ' ..\n"
"                            'variable name if \"from\" field is set to \"env\"')\n"
"                    end\n"
"                    if var.from == 'file' and var.file == nil then\n"
"                        w.error('\"file\" field must define a file name if ' ..\n"
"                            '\"from\" field is set to \"file\"')\n"
"                    end\n"
"                end,\n"
"            }),\n"
"        }),\n"
"    }),\n"
"    process = schema.record({\n"
"        strip_core = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'strip_core',\n"
"            box_cfg_nondynamic = true,\n"
"            default = true,\n"
"        }),\n"
"        coredump = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'coredump',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        }),\n"
"        background = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'background',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        }),\n"
"        title = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'custom_proc_title',\n"
"            default = 'tarantool - {{ instance_name }}',\n"
"        }),\n"
"        username = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'username',\n"
"            box_cfg_nondynamic = true,\n"
"            default = box.NULL,\n"
"        }),\n"
"        work_dir = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'work_dir',\n"
"            box_cfg_nondynamic = true,\n"
"            -- The mkdir annotation is not present here, because\n"
"            -- otherwise the directory would be created\n"
"            -- unconditionally. Instead, mkdir applier creates it\n"
"            -- only before the first box.cfg() call.\n"
"            default = box.NULL,\n"
"        }),\n"
"        pid_file = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'pid_file',\n"
"            box_cfg_nondynamic = true,\n"
"            mk_parent_dir = true,\n"
"            default = 'var/run/{{ instance_name }}/tarantool.pid',\n"
"        }),\n"
"    }),\n"
"    console = schema.record({\n"
"        enabled = schema.scalar({\n"
"            type = 'boolean',\n"
"            default = true,\n"
"        }),\n"
"        socket = schema.scalar({\n"
"            type = 'string',\n"
"            -- The mk_parent_dir annotation is not present here,\n"
"            -- because otherwise the directory would be created\n"
"            -- unconditionally. Instead, mkdir applier creates it\n"
"            -- if console.enabled is true.\n"
"            default = 'var/run/{{ instance_name }}/tarantool.control',\n"
"        }),\n"
"    }),\n"
"    fiber = schema.record({\n"
"        io_collect_interval = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'io_collect_interval',\n"
"            default = box.NULL,\n"
"        }),\n"
"        too_long_threshold = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'too_long_threshold',\n"
"            default = 0.5,\n"
"        }),\n"
"        worker_pool_threads = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'worker_pool_threads',\n"
"            default = 4,\n"
"        }),\n"
"        slice = schema.record({\n"
"            warn = schema.scalar({\n"
"                type = 'number',\n"
"                default = 0.5,\n"
"            }),\n"
"            err = schema.scalar({\n"
"                type = 'number',\n"
"                default = 1,\n"
"            }),\n"
"        }),\n"
"        top = schema.record({\n"
"            enabled = schema.scalar({\n"
"                type = 'boolean',\n"
"                default = false,\n"
"            }),\n"
"        }),\n"
"    }),\n"
"    log = schema.record({\n"
"        -- The logger destination is handled separately in the\n"
"        -- box_cfg applier, so there are no explicit box_cfg and\n"
"        -- box_cfg_nondynamic annotations.\n"
"        --\n"
"        -- The reason is that there is no direct-no-transform\n"
"        -- mapping from, say, `log.file` to `box_cfg.log`.\n"
"        -- The applier should add the `file:` prefix.\n"
"        to = schema.enum({\n"
"            'stderr',\n"
"            'file',\n"
"            'pipe',\n"
"            'syslog',\n"
"        }, {\n"
"            default = 'stderr',\n"
"        }),\n"
"        file = schema.scalar({\n"
"            type = 'string',\n"
"            -- The mk_parent_dir annotation is not present here,\n"
"            -- because otherwise the directory would be created\n"
"            -- unconditionally. Instead, mkdir applier creates it\n"
"            -- if log.to is 'file'.\n"
"            default = 'var/log/{{ instance_name }}/tarantool.log',\n"
"        }),\n"
"        pipe = schema.scalar({\n"
"            type = 'string',\n"
"            default = box.NULL,\n"
"        }),\n"
"        syslog = schema.record({\n"
"            identity = schema.scalar({\n"
"                type = 'string',\n"
"                default = 'tarantool',\n"
"            }),\n"
"            facility = schema.scalar({\n"
"                type = 'string',\n"
"                default = 'local7',\n"
"            }),\n"
"            server = schema.scalar({\n"
"                type = 'string',\n"
"                -- The logger tries /dev/log and then\n"
"                -- /var/run/syslog if no server is provided.\n"
"                default = box.NULL,\n"
"            }),\n"
"        }),\n"
"        nonblock = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'log_nonblock',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        }),\n"
"        level = schema.scalar({\n"
"            type = 'number, string',\n"
"            box_cfg = 'log_level',\n"
"            default = 5,\n"
"            allowed_values = {\n"
"                0, 'fatal',\n"
"                1, 'syserror',\n"
"                2, 'error',\n"
"                3, 'crit',\n"
"                4, 'warn',\n"
"                5, 'info',\n"
"                6, 'verbose',\n"
"                7, 'debug',\n"
"            },\n"
"        }),\n"
"        format = schema.enum({\n"
"            'plain',\n"
"            'json',\n"
"        }, {\n"
"            box_cfg = 'log_format',\n"
"            default = 'plain',\n"
"        }),\n"
"        -- box.cfg({log_modules = <...>}) replaces the previous\n"
"        -- value without any merging.\n"
"        --\n"
"        -- If a key in this map is removed in the provided\n"
"        -- configuration, then it will be removed in the actually\n"
"        -- applied configuration.\n"
"        --\n"
"        -- It is exactly what we need there to make the\n"
"        -- configuration independent of previously applied values.\n"
"        modules = schema.map({\n"
"            key = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            value = schema.scalar({\n"
"                type = 'number, string',\n"
"            }),\n"
"            box_cfg = 'log_modules',\n"
"            -- TODO: This default doesn't work now. It needs\n"
"            -- support of non-scalar schema nodes in\n"
"            -- <schema object>:map().\n"
"            default = box.NULL,\n"
"        }),\n"
"    }, {\n"
"        validate = function(log, w)\n"
"            if log.to == 'pipe' and log.pipe == nil then\n"
"                w.error('The pipe logger is set by the log.to parameter but ' ..\n"
"                    'the command is not set (log.pipe parameter)')\n"
"            end\n"
"        end,\n"
"    }),\n"
"    iproto = schema.record({\n"
"        listen = schema.array({\n"
"            items = schema.record({\n"
"                uri = schema.scalar({\n"
"                    type = 'string',\n"
"                    validate = validate_uri_field(false, false),\n"
"                }),\n"
"                params = schema.record({\n"
"                    transport = schema.enum({\n"
"                        'plain',\n"
"                        'ssl',\n"
"                    }),\n"
"                    -- Mandatory server options for TLS.\n"
"                    ssl_key_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_cert_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    -- Optional server options for TLS.\n"
"                    ssl_ca_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_ciphers = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                }),\n"
"            }, {\n"
"                validate = function(data, w)\n"
"                    -- If data is not nil then the URI should be there.\n"
"                    if data.uri == nil then\n"
"                        w.error('The URI is required for iproto.listen')\n"
"                    end\n"
"                end,\n"
"            }),\n"
"            box_cfg = 'listen',\n"
"        }),\n"
"        -- URIs for clients to let them know where to connect.\n"
"        --\n"
"        -- There are several possibly different URIs:\n"
"        --\n"
"        -- * client\n"
"        --\n"
"        --   The informational value for clients. It is not used\n"
"        --   in tarantool anyhow (only validated). Contains only\n"
"        --   host:port: no user, no password.\n"
"        --\n"
"        --   Must be suitable to connect (no INADDR_ANY, no\n"
"        --   in6addr_any, no zero port).\n"
"        --\n"
"        --   Note: the host:port part may represent a Unix domain\n"
"        --   socket: host = 'unix/', port = '/path/to/socket'.\n"
"        --\n"
"        -- * peer\n"
"        --\n"
"        --   The general purpose peer URI, used for connections\n"
"        --   within the cluster (replica -> master, router ->\n"
"        --   storage, rebalancer -> storage).\n"
"        --\n"
"        --   Usually points to a user with the 'replication' role.\n"
"        --\n"
"        -- * sharding\n"
"        --\n"
"        --   The URI for router and rebalancer.\n"
"        --\n"
"        --   If unset, the general peer URI should be used.\n"
"        --\n"
"        -- The iproto.advertise.{peer,sharding} options have the\n"
"        -- following syntax variants:\n"
"        --\n"
"        -- With URI:\n"
"        -- {uri = ...}\n"
"        -- {uri = ..., params = ...}\n"
"        -- {login = ..., uri = ...}\n"
"        -- {login = ..., password = ..., uri = ...}\n"
"        -- {login = ..., uri = ..., params = ...}\n"
"        -- {login = ..., password = ..., uri = ..., params = ...}\n"
"        --\n"
"        -- Without URI:\n"
"        --\n"
"        -- {login = ...}\n"
"        -- {login = ..., password = ...}\n"
"        --\n"
"        -- where uri is host:port.\n"
"        --\n"
"        -- Note: the host:port part may represent a Unix domain\n"
"        -- socket: host = 'unix/', port = '/path/to/socket'.\n"
"        --\n"
"        -- If there is no uri field, it is to be looked in\n"
"        -- iproto.listen.\n"
"        --\n"
"        -- If there is a user, but no password, the password is to\n"
"        -- be looked in the `credentials` section of the\n"
"        -- configuration (except user 'guest', which can't  have a\n"
"        -- password).\n"
"        advertise = schema.record({\n"
"            client = schema.scalar({\n"
"                type = 'string',\n"
"                default = box.NULL,\n"
"                validate = validate_uri_field(false, true),\n"
"            }),\n"
"            peer = schema.record({\n"
"                login = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                password = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                uri = schema.scalar({\n"
"                    type = 'string',\n"
"                    validate = validate_uri_field(true, true),\n"
"                }),\n"
"                params = schema.record({\n"
"                    transport = schema.enum({\n"
"                        'plain',\n"
"                        'ssl',\n"
"                    }),\n"
"                    ssl_ca_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_cert_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_ciphers = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_key_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                }),\n"
"            }, {\n"
"                validate = function(data, w)\n"
"                    if next(data) == nil then\n"
"                        w.error('An URI should have at least one field')\n"
"                    end\n"
"                    -- If a password is set, a login must also be specified.\n"
"                    if data.password ~= nil and data.login == nil then\n"
"                        w.error('Password cannot be set without setting login')\n"
"                    end\n"
"                    -- If a params is set, an uri must also be specified.\n"
"                    if data.params ~= nil and data.uri == nil then\n"
"                        w.error('Params cannot be set without setting uri')\n"
"                    end\n"
"                end,\n"
"            }),\n"
"            sharding = schema.record({\n"
"                login = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                password = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                uri = schema.scalar({\n"
"                    type = 'string',\n"
"                    validate = validate_uri_field(true, true),\n"
"                }),\n"
"                params = schema.record({\n"
"                    transport = schema.enum({\n"
"                        'plain',\n"
"                        'ssl',\n"
"                    }),\n"
"                    ssl_ca_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_cert_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_ciphers = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_key_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                    ssl_password_file = enterprise_edition(schema.scalar({\n"
"                        type = 'string',\n"
"                    })),\n"
"                }),\n"
"            }, {\n"
"                validate = function(data, w)\n"
"                    if next(data) == nil then\n"
"                        w.error('An URI should have at least one field')\n"
"                    end\n"
"                    -- If a password is set, a login must also be specified.\n"
"                    if data.password ~= nil and data.login == nil then\n"
"                        w.error('Password cannot be set without setting login')\n"
"                    end\n"
"                    -- If a params is set, an uri must also be specified.\n"
"                    if data.params ~= nil and data.uri == nil then\n"
"                        w.error('Params cannot be set without setting uri')\n"
"                    end\n"
"                end,\n"
"            }),\n"
"        }),\n"
"        threads = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'iproto_threads',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 1,\n"
"        }),\n"
"        net_msg_max = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'net_msg_max',\n"
"            default = 768,\n"
"        }),\n"
"        readahead = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'readahead',\n"
"            default = 16320,\n"
"        }),\n"
"    }),\n"
"    database = schema.record({\n"
"        instance_uuid = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'instance_uuid',\n"
"            default = box.NULL,\n"
"            validate = validate_uuid_str,\n"
"        }),\n"
"        replicaset_uuid = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'replicaset_uuid',\n"
"            default = box.NULL,\n"
"            validate = validate_uuid_str,\n"
"        }),\n"
"        hot_standby = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'hot_standby',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        }),\n"
"        -- Applied to box_cfg.read_only.\n"
"        --\n"
"        -- The effective default depends on amount of instances in\n"
"        -- a replicaset.\n"
"        --\n"
"        -- A singleton instance (the only instance in the\n"
"        -- replicaset) is in the 'rw' mode by default.\n"
"        --\n"
"        -- If the replicaset contains several (more than one)\n"
"        -- instances, the default is 'ro'.\n"
"        mode = schema.enum({\n"
"            'ro',\n"
"            'rw',\n"
"        }, {\n"
"            default = box.NULL,\n"
"        }),\n"
"        txn_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'txn_timeout',\n"
"            default = 365 * 100 * 86400,\n"
"        }),\n"
"        txn_isolation = schema.enum({\n"
"            'read-committed',\n"
"            'read-confirmed',\n"
"            'best-effort',\n"
"        }, {\n"
"            box_cfg = 'txn_isolation',\n"
"            default = 'best-effort',\n"
"        }),\n"
"        use_mvcc_engine = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'memtx_use_mvcc_engine',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        }),\n"
"    }),\n"
"    sql = schema.record({\n"
"        cache_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'sql_cache_size',\n"
"            default = 5 * 1024 * 1024,\n"
"        }),\n"
"    }),\n"
"    memtx = schema.record({\n"
"        memory = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'memtx_memory',\n"
"            default = 256 * 1024 * 1024,\n"
"        }),\n"
"        allocator = schema.enum({\n"
"            'small',\n"
"            'system',\n"
"        }, {\n"
"            box_cfg = 'memtx_allocator',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 'small',\n"
"        }),\n"
"        slab_alloc_granularity = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'slab_alloc_granularity',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 8,\n"
"        }),\n"
"        slab_alloc_factor = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'slab_alloc_factor',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 1.05,\n"
"        }),\n"
"        min_tuple_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'memtx_min_tuple_size',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 16,\n"
"        }),\n"
"        max_tuple_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'memtx_max_tuple_size',\n"
"            default = 1024 * 1024,\n"
"        }),\n"
"        sort_threads = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'memtx_sort_threads',\n"
"            box_cfg_nondynamic = true,\n"
"            default = box.NULL,\n"
"        }),\n"
"    }),\n"
"    vinyl = schema.record({\n"
"        bloom_fpr = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'vinyl_bloom_fpr',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 0.05,\n"
"        }),\n"
"        cache = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_cache',\n"
"            default = 128 * 1024 * 1024,\n"
"        }),\n"
"        defer_deletes = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'vinyl_defer_deletes',\n"
"            default = false,\n"
"        }),\n"
"        dir = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'vinyl_dir',\n"
"            box_cfg_nondynamic = true,\n"
"            mkdir = true,\n"
"            default = 'var/lib/{{ instance_name }}',\n"
"        }),\n"
"        max_tuple_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_max_tuple_size',\n"
"            default = 1024 * 1024,\n"
"        }),\n"
"        memory = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_memory',\n"
"            default = 128 * 1024 * 1024,\n"
"        }),\n"
"        page_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_page_size',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 8 * 1024,\n"
"        }),\n"
"        range_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_range_size',\n"
"            box_cfg_nondynamic = true,\n"
"            default = box.NULL,\n"
"        }),\n"
"        read_threads = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_read_threads',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 1,\n"
"        }),\n"
"        run_count_per_level = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_run_count_per_level',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 2,\n"
"        }),\n"
"        run_size_ratio = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'vinyl_run_size_ratio',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 3.5,\n"
"        }),\n"
"        timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'vinyl_timeout',\n"
"            default = 60,\n"
"        }),\n"
"        write_threads = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'vinyl_write_threads',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 4,\n"
"        }),\n"
"    }),\n"
"    wal = schema.record({\n"
"        dir = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'wal_dir',\n"
"            box_cfg_nondynamic = true,\n"
"            mkdir = true,\n"
"            default = 'var/lib/{{ instance_name }}',\n"
"        }),\n"
"        mode = schema.enum({\n"
"            'none',\n"
"            'write',\n"
"            'fsync',\n"
"        }, {\n"
"            box_cfg = 'wal_mode',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 'write',\n"
"        }),\n"
"        max_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'wal_max_size',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 256 * 1024 * 1024,\n"
"        }),\n"
"        dir_rescan_delay = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'wal_dir_rescan_delay',\n"
"            default = 2,\n"
"        }),\n"
"        queue_max_size = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'wal_queue_max_size',\n"
"            default = 16 * 1024 * 1024,\n"
"        }),\n"
"        cleanup_delay = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'wal_cleanup_delay',\n"
"            default = 4 * 3600,\n"
"        }),\n"
"        -- box.cfg({wal_ext = <...>}) replaces the previous\n"
"        -- value without any merging. See explanation why it is\n"
"        -- important in the log.modules description.\n"
"        ext = enterprise_edition(schema.record({\n"
"            old = schema.scalar({\n"
"                type = 'boolean',\n"
"                -- TODO: This default is applied despite the outer\n"
"                -- apply_default_if, because the annotation has no\n"
"                -- effect on child schema nodes.\n"
"                --\n"
"                -- This default is purely informational: lack of the\n"
"                -- value doesn't break configuration applying\n"
"                -- idempotence.\n"
"                -- default = false,\n"
"            }),\n"
"            new = schema.scalar({\n"
"                type = 'boolean',\n"
"                -- TODO: See wal.ext.old.\n"
"                -- default = false,\n"
"            }),\n"
"            spaces = schema.map({\n"
"                key = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                value = schema.record({\n"
"                    old = schema.scalar({\n"
"                        type = 'boolean',\n"
"                        default = false,\n"
"                    }),\n"
"                    new = schema.scalar({\n"
"                        type = 'boolean',\n"
"                        default = false,\n"
"                    }),\n"
"                }),\n"
"            }),\n"
"        }, {\n"
"            box_cfg = 'wal_ext',\n"
"            -- TODO: This default doesn't work now. It needs\n"
"            -- support of non-scalar schema nodes in\n"
"            -- <schema object>:map().\n"
"            default = box.NULL,\n"
"        })),\n"
"    }),\n"
"    snapshot = schema.record({\n"
"        dir = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'memtx_dir',\n"
"            box_cfg_nondynamic = true,\n"
"            mkdir = true,\n"
"            default = 'var/lib/{{ instance_name }}',\n"
"        }),\n"
"        by = schema.record({\n"
"            interval = schema.scalar({\n"
"                type = 'number',\n"
"                box_cfg = 'checkpoint_interval',\n"
"                default = 3600,\n"
"            }),\n"
"            wal_size = schema.scalar({\n"
"                type = 'integer',\n"
"                box_cfg = 'checkpoint_wal_threshold',\n"
"                default = 1e18,\n"
"            }),\n"
"        }),\n"
"        count = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'checkpoint_count',\n"
"            default = 2,\n"
"        }),\n"
"        snap_io_rate_limit = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'snap_io_rate_limit',\n"
"            default = box.NULL,\n"
"        }),\n"
"    }),\n"
"    replication = schema.record({\n"
"        failover = schema.enum({\n"
"            -- No failover ('off').\n"
"            --\n"
"            -- The leadership in replicasets is controlled using\n"
"            -- the database.mode options. It is allowed to have\n"
"            -- several writable instances in a replicaset.\n"
"            --\n"
"            -- The default database.mode is determined as follows.\n"
"            --\n"
"            -- * 1 instance in a replicaset: 'rw'.\n"
"            -- * >1 instances in a replicaset: 'ro'.\n"
"            'off',\n"
"            -- Manual failover ('manual').\n"
"            --\n"
"            -- The leadership is controlled using the 'leader'\n"
"            -- option of a replicaset. Master-master configuration\n"
"            -- is forbidden.\n"
"            --\n"
"            -- The database.mode option can't be set directly in\n"
"            -- the manual failover mode. The leader is configured\n"
"            -- in the read-write mode, all the other instances are\n"
"            -- read-only.\n"
"            'manual',\n"
"            -- Automatic leader election ('election').\n"
"            --\n"
"            -- Uses a RAFT based algorithm for the leader election.\n"
"            --\n"
"            -- No database.mode or 'leader' options should be set.\n"
"            'election',\n"
"            -- Automatic leader appointment by an external\n"
"            -- failover agent ('supervised').\n"
"            --\n"
"            -- The default database.mode is 'ro' (but there is an\n"
"            -- exception during a replicaset bootstrapping, see\n"
"            -- applier/box_cfg.lua).\n"
"            --\n"
"            -- The failover agent assigns the 'rw' mode to one\n"
"            -- instance in the replicaset.\n"
"            --\n"
"            -- No database.mode or 'leader' options should be set.\n"
"            --\n"
"            -- TODO: Raise an error if this value is set on\n"
"            -- Tarantool Community Edition. The instance side code\n"
"            -- that handles failover agent commands is part of\n"
"            -- Tarantool Enterprise Edition, so there is no much\n"
"            -- sense to enable this mode on the Community Edition.\n"
"            'supervised',\n"
"        }, {\n"
"            default = 'off',\n"
"        }),\n"
"        -- XXX: needs more validation\n"
"        peers = schema.array({\n"
"            items = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            box_cfg = 'replication',\n"
"            default = box.NULL,\n"
"        }),\n"
"        anon = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'replication_anon',\n"
"            default = false,\n"
"        }),\n"
"        threads = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'replication_threads',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 1,\n"
"        }),\n"
"        timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'replication_timeout',\n"
"            default = 1,\n"
"        }),\n"
"        synchro_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'replication_synchro_timeout',\n"
"            default = 5,\n"
"        }),\n"
"        connect_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'replication_connect_timeout',\n"
"            default = 30,\n"
"        }),\n"
"        sync_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'replication_sync_timeout',\n"
"            default = compat.box_cfg_replication_sync_timeout:is_old()\n"
"                and 300 or 0,\n"
"        }),\n"
"        sync_lag = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'replication_sync_lag',\n"
"            default = 10,\n"
"        }),\n"
"        synchro_quorum = schema.scalar({\n"
"            type = 'string, number',\n"
"            box_cfg = 'replication_synchro_quorum',\n"
"            default = 'N / 2 + 1',\n"
"        }),\n"
"        skip_conflict = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'replication_skip_conflict',\n"
"            default = false,\n"
"        }),\n"
"        election_mode = schema.enum({\n"
"            'off',\n"
"            'voter',\n"
"            'manual',\n"
"            'candidate',\n"
"        }, {\n"
"            box_cfg = 'election_mode',\n"
"            -- The effective default is determined depending of\n"
"            -- the replication.failover option.\n"
"            default = box.NULL,\n"
"        }),\n"
"        election_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'election_timeout',\n"
"            default = 5,\n"
"        }),\n"
"        election_fencing_mode = schema.enum({\n"
"            'off',\n"
"            'soft',\n"
"            'strict',\n"
"        }, {\n"
"            box_cfg = 'election_fencing_mode',\n"
"            default = 'soft',\n"
"        }),\n"
"        bootstrap_strategy = schema.enum({\n"
"            'auto',\n"
"            'config',\n"
"            'supervised',\n"
"            'legacy',\n"
"        }, {\n"
"            box_cfg = 'bootstrap_strategy',\n"
"            default = 'auto',\n"
"        }),\n"
"    }),\n"
"    -- Unlike other sections, credentials contains the append-only\n"
"    -- parameters. It means that deletion of a value from the\n"
"    -- config doesn't delete the corresponding user/role/privilege.\n"
"    credentials = schema.record({\n"
"        roles = schema.map({\n"
"            -- Name of the role.\n"
"            key = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            value = schema.record({\n"
"                -- Privileges granted to the role.\n"
"                privileges = schema.array({\n"
"                    items = schema.record({\n"
"                        permissions = schema.set({\n"
"                            'read',\n"
"                            'write',\n"
"                            'execute',\n"
"                            'create',\n"
"                            'alter',\n"
"                            'drop',\n"
"                            'usage',\n"
"                            'session',\n"
"                        }),\n"
"                        universe = schema.scalar({\n"
"                            type = 'boolean',\n"
"                        }),\n"
"                        spaces = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        functions = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        sequences = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        lua_eval = schema.scalar({\n"
"                            type = 'boolean',\n"
"                        }),\n"
"                        lua_call = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                                allowed_values = {'all'},\n"
"                            }),\n"
"                        }),\n"
"                        sql = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                                allowed_values = {'all'},\n"
"                            }),\n"
"                        }),\n"
"                    }),\n"
"                }),\n"
"                -- The given role has all the privileges from\n"
"                -- these underlying roles.\n"
"                roles = schema.array({\n"
"                    items = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                }),\n"
"            }),\n"
"        }),\n"
"        users = schema.map({\n"
"            -- Name of the user.\n"
"            key = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            -- Parameters of the user.\n"
"            value = schema.record({\n"
"                password = schema.scalar({\n"
"                    type = 'string',\n"
"                }),\n"
"                privileges = schema.array({\n"
"                    items = schema.record({\n"
"                        permissions = schema.set({\n"
"                            'read',\n"
"                            'write',\n"
"                            'execute',\n"
"                            'create',\n"
"                            'alter',\n"
"                            'drop',\n"
"                            'usage',\n"
"                            'session',\n"
"                        }),\n"
"                        universe = schema.scalar({\n"
"                            type = 'boolean',\n"
"                        }),\n"
"                        spaces = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        functions = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        sequences = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                            }),\n"
"                        }),\n"
"                        lua_eval = schema.scalar({\n"
"                            type = 'boolean',\n"
"                        }),\n"
"                        lua_call = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                                allowed_values = {'all'},\n"
"                            }),\n"
"                        }),\n"
"                        sql = schema.array({\n"
"                            items = schema.scalar({\n"
"                                type = 'string',\n"
"                                allowed_values = {'all'},\n"
"                            }),\n"
"                        }),\n"
"                    }),\n"
"                }),\n"
"                -- The given user has all the privileges from\n"
"                -- these underlying roles.\n"
"                roles = schema.array({\n"
"                    items = schema.scalar({\n"
"                        type = 'string',\n"
"                    }),\n"
"                }),\n"
"            }),\n"
"        }),\n"
"    }),\n"
"    app = schema.record({\n"
"        file = schema.scalar({\n"
"            type = 'string',\n"
"        }),\n"
"        module = schema.scalar({\n"
"            type = 'string',\n"
"        }),\n"
"        cfg = schema.map({\n"
"            key = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            value = schema.scalar({\n"
"                type = 'any',\n"
"            }),\n"
"        }),\n"
"    }, {\n"
"        validate = function(app, w)\n"
"            if app.file ~= nil and app.module ~= nil then\n"
"                w.error('Fields file and module cannot appear at the same time')\n"
"            end\n"
"        end,\n"
"    }),\n"
"    feedback = schema.record({\n"
"        enabled = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'feedback_enabled',\n"
"            default = true,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        crashinfo = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'feedback_crashinfo',\n"
"            default = true,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        host = schema.scalar({\n"
"            type = 'string',\n"
"            box_cfg = 'feedback_host',\n"
"            default = 'https://feedback.tarantool.io',\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        metrics_collect_interval = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'feedback_metrics_collect_interval',\n"
"            default = 60,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        send_metrics = schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'feedback_send_metrics',\n"
"            default = true,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        interval = schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'feedback_interval',\n"
"            default = 3600,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"        metrics_limit = schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'feedback_metrics_limit',\n"
"            default = 1024 * 1024,\n"
"            apply_default_if = feedback_apply_default_if,\n"
"            validate = feedback_validate,\n"
"        }),\n"
"    }),\n"
"    flightrec = schema.record({\n"
"        enabled = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'flightrec_enabled',\n"
"            default = false,\n"
"        })),\n"
"        logs_size = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_logs_size',\n"
"            default = 10485760,\n"
"        })),\n"
"        logs_max_msg_size = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_logs_max_msg_size',\n"
"            default = 4096,\n"
"        })),\n"
"        logs_log_level = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_logs_log_level',\n"
"            default = 6,\n"
"            allowed_values = {0, 1, 2, 3, 4, 5, 6, 7},\n"
"        })),\n"
"        metrics_interval = enterprise_edition(schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'flightrec_metrics_interval',\n"
"            default = 1.0,\n"
"        })),\n"
"        metrics_period = enterprise_edition(schema.scalar({\n"
"            type = 'number',\n"
"            box_cfg = 'flightrec_metrics_period',\n"
"            default = 60 * 3,\n"
"        })),\n"
"        requests_size = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_requests_size',\n"
"            default = 10485760,\n"
"        })),\n"
"        requests_max_req_size = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_requests_max_req_size',\n"
"            default = 16384,\n"
"        })),\n"
"        requests_max_res_size = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            box_cfg = 'flightrec_requests_max_res_size',\n"
"            default = 16384,\n"
"        })),\n"
"    }),\n"
"    security = schema.record({\n"
"        auth_type = schema.enum({\n"
"            'chap-sha1',\n"
"            'pap-sha256',\n"
"        }, {\n"
"            box_cfg = 'auth_type',\n"
"            default = 'chap-sha1',\n"
"            validate = function(auth_type, w)\n"
"                if auth_type ~= 'chap-sha1' and\n"
"                        tarantool.package ~= 'Tarantool Enterprise' then\n"
"                    w.error('\"chap-sha1\" is the only authentication method ' ..\n"
"                            '(auth_type) available in Tarantool Community ' ..\n"
"                            'Edition (%q requested)', auth_type)\n"
"                end\n"
"            end,\n"
"        }),\n"
"        auth_delay = enterprise_edition(schema.scalar({\n"
"            type = 'number',\n"
"            default = 0,\n"
"            box_cfg = 'auth_delay',\n"
"        })),\n"
"        auth_retries = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            default = 0,\n"
"            box_cfg = 'auth_retries',\n"
"        })),\n"
"        disable_guest = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'disable_guest',\n"
"        })),\n"
"        secure_erasing = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'secure_erasing',\n"
"        })),\n"
"        password_lifetime_days = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            default = 0,\n"
"            box_cfg = 'password_lifetime_days',\n"
"        })),\n"
"        password_min_length = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            default = 0,\n"
"            box_cfg = 'password_min_length',\n"
"        })),\n"
"        password_enforce_uppercase = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'password_enforce_uppercase',\n"
"        })),\n"
"        password_enforce_lowercase = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'password_enforce_lowercase',\n"
"        })),\n"
"        password_enforce_digits = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'password_enforce_digits',\n"
"        })),\n"
"        password_enforce_specialchars = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            default = false,\n"
"            box_cfg = 'password_enforce_specialchars',\n"
"        })),\n"
"        password_history_length = enterprise_edition(schema.scalar({\n"
"            type = 'integer',\n"
"            default = 0,\n"
"            box_cfg = 'password_history_length',\n"
"        })),\n"
"    }),\n"
"    metrics = schema.record({\n"
"        -- Metrics doesn't have box_cfg annotation, because currently nested\n"
"        -- options and maps/arrays defaults are not supported.\n"
"        include = schema.set({\n"
"            'all',\n"
"            'network',\n"
"            'operations',\n"
"            'system',\n"
"            'replicas',\n"
"            'info',\n"
"            'slab',\n"
"            'runtime',\n"
"            'memory',\n"
"            'spaces',\n"
"            'fibers',\n"
"            'cpu',\n"
"            'vinyl',\n"
"            'memtx',\n"
"            'luajit',\n"
"            'clock',\n"
"            'event_loop',\n"
"        }),\n"
"        exclude = schema.set({\n"
"            'all',\n"
"            'network',\n"
"            'operations',\n"
"            'system',\n"
"            'replicas',\n"
"            'info',\n"
"            'slab',\n"
"            'runtime',\n"
"            'memory',\n"
"            'spaces',\n"
"            'fibers',\n"
"            'cpu',\n"
"            'vinyl',\n"
"            'memtx',\n"
"            'luajit',\n"
"            'clock',\n"
"            'event_loop',\n"
"        }),\n"
"        labels = schema.map({\n"
"            key = schema.scalar({type = 'string'}),\n"
"            value = schema.scalar({type = 'string'}),\n"
"        }),\n"
"    }),\n"
"    sharding = schema.record({\n"
"        -- Instance vshard options.\n"
"        zone = schema.scalar({type = 'integer'}),\n"
"        -- Replicaset vshard options.\n"
"        lock = schema.scalar({\n"
"            type = 'boolean',\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope == 'instance' then\n"
"                    w.error('sharding.lock cannot be defined in the instance '..\n"
"                            'scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        -- TODO: Add validate.\n"
"        roles = schema.set({\n"
"            'router',\n"
"            'storage',\n"
"            'rebalancer',\n"
"        }),\n"
"        -- Global vshard options.\n"
"        shard_index = schema.scalar({\n"
"            type = 'string',\n"
"            default = 'bucket_id',\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.shard_index should be a defined in '..\n"
"                            'global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        bucket_count = schema.scalar({\n"
"            type = 'integer',\n"
"            default = 3000,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.bucket_count should be a defined in '..\n"
"                            'global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        rebalancer_disbalance_threshold = schema.scalar({\n"
"            type = 'number',\n"
"            default = 1,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.rebalancer_disbalance_threshold should '..\n"
"                            'be a defined in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        rebalancer_max_receiving = schema.scalar({\n"
"            type = 'integer',\n"
"            default = 100,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.rebalancer_max_receiving should '..\n"
"                            'be a defined in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        rebalancer_max_sending = schema.scalar({\n"
"            type = 'integer',\n"
"            default = 1,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.rebalancer_max_sending should '..\n"
"                            'be a defined in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        sync_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            default = 1,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.sync_timeout should be a defined in '..\n"
"                            'global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        connection_outdate_delay = schema.scalar({\n"
"            type = 'number',\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.connection_outdate_delay should be a '..\n"
"                            'defined in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        failover_ping_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            default = 5,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.failover_ping_timeout should be a '..\n"
"                            'defined in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        discovery_mode = schema.enum({\n"
"            'on',\n"
"            'off',\n"
"            'once',\n"
"        }, {\n"
"            default = 'on',\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.discovery_mode should be a defined in '..\n"
"                            'global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        sched_ref_quota = schema.scalar({\n"
"            type = 'number',\n"
"            default = 300,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.sched_ref_quota should be a defined ' ..\n"
"                            'in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"        sched_move_quota = schema.scalar({\n"
"            type = 'number',\n"
"            default = 1,\n"
"            validate = function(data, w)\n"
"                local scope = w.schema.computed.annotations.scope\n"
"                if data == nil or scope == nil then\n"
"                    return\n"
"                end\n"
"                if scope ~= 'global' then\n"
"                    w.error('sharding.sched_move_quota should be a defined ' ..\n"
"                            'in global scope')\n"
"                end\n"
"            end,\n"
"        }),\n"
"    }, {\n"
"        validate = function(data, w)\n"
"            -- Forbid sharding.roles in instance scope.\n"
"            local scope = w.schema.computed.annotations.scope\n"
"            if data.roles ~= nil and scope == 'instance' then\n"
"                w.error('sharding.roles cannot be defined in the instance ' ..\n"
"                        'scope')\n"
"            end\n"
"            -- Make sure that if the rebalancer role is present, the storage\n"
"            -- role is also present.\n"
"            if data.roles ~= nil then\n"
"                local has_storage = false\n"
"                local has_rebalancer = false\n"
"                for _, role in pairs(data.roles) do\n"
"                    has_storage = has_storage or role == 'storage'\n"
"                    has_rebalancer = has_rebalancer or role == 'rebalancer'\n"
"                end\n"
"                if has_rebalancer and not has_storage then\n"
"                    w.error('The rebalancer role cannot be present without ' ..\n"
"                            'the storage role')\n"
"                end\n"
"            end\n"
"        end,\n"
"    }),\n"
"    audit_log = enterprise_edition(schema.record({\n"
"        -- The same as the destination for the logger, audit logger destination\n"
"        -- is handled separately in the box_cfg applier, so there are no\n"
"        -- explicit box_cfg and box_cfg_nondynamic annotations.\n"
"        --\n"
"        -- The reason is that there is no direct-no-transform\n"
"        -- mapping from, say, `audit_log.file` to `box_cfg.audit_log`.\n"
"        -- The applier should add the `file:` prefix.\n"
"        to = enterprise_edition(schema.enum({\n"
"            'devnull',\n"
"            'file',\n"
"            'pipe',\n"
"            'syslog',\n"
"        }, {\n"
"            default = 'devnull',\n"
"        })),\n"
"        file = enterprise_edition(schema.scalar({\n"
"            type = 'string',\n"
"            -- The mk_parent_dir annotation is not present here,\n"
"            -- because otherwise the directory would be created\n"
"            -- unconditionally. Instead, mkdir applier creates it\n"
"            -- if audit_log.to is 'file'.\n"
"            default = 'var/log/{{ instance_name }}/audit.log',\n"
"        })),\n"
"        pipe = enterprise_edition(schema.scalar({\n"
"            type = 'string',\n"
"            default = box.NULL,\n"
"        })),\n"
"        syslog = schema.record({\n"
"            identity = enterprise_edition(schema.scalar({\n"
"                type = 'string',\n"
"                default = 'tarantool',\n"
"            })),\n"
"            facility = enterprise_edition(schema.scalar({\n"
"                type = 'string',\n"
"                default = 'local7',\n"
"            })),\n"
"            server = enterprise_edition(schema.scalar({\n"
"                type = 'string',\n"
"                -- The logger tries /dev/log and then\n"
"                -- /var/run/syslog if no server is provided.\n"
"                default = box.NULL,\n"
"            })),\n"
"        }),\n"
"        nonblock = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'audit_nonblock',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        })),\n"
"        format = enterprise_edition(schema.enum({\n"
"            'plain',\n"
"            'json',\n"
"            'csv',\n"
"        }, {\n"
"            box_cfg = 'audit_format',\n"
"            box_cfg_nondynamic = true,\n"
"            default = 'json',\n"
"        })),\n"
"        -- The reason for the absence of the box_cfg and box_cfg_nondynamic\n"
"        -- annotations is that this setting needs to be converted to a string\n"
"        -- before being set to 'audit_filter'. This will be done in box_cfg\n"
"        -- applier.\n"
"        --\n"
"        -- TODO: Add a validation and a default value. Currently, the audit_log\n"
"        -- validation can catch the setting of the option in the CE, but adding\n"
"        -- its own validation seems more appropriate.\n"
"        filter = schema.set({\n"
"            -- Events.\n"
"            \"audit_enable\",\n"
"            \"custom\",\n"
"            \"auth_ok\",\n"
"            \"auth_fail\",\n"
"            \"disconnect\",\n"
"            \"user_create\",\n"
"            \"user_drop\",\n"
"            \"role_create\",\n"
"            \"role_drop\",\n"
"            \"user_enable\",\n"
"            \"user_disable\",\n"
"            \"user_grant_rights\",\n"
"            \"user_revoke_rights\",\n"
"            \"role_grant_rights\",\n"
"            \"role_revoke_rights\",\n"
"            \"password_change\",\n"
"            \"access_denied\",\n"
"            \"eval\",\n"
"            \"call\",\n"
"            \"space_select\",\n"
"            \"space_create\",\n"
"            \"space_alter\",\n"
"            \"space_drop\",\n"
"            \"space_insert\",\n"
"            \"space_replace\",\n"
"            \"space_delete\",\n"
"            -- Groups of events.\n"
"            \"none\",\n"
"            \"all\",\n"
"            \"audit\",\n"
"            \"auth\",\n"
"            \"priv\",\n"
"            \"ddl\",\n"
"            \"dml\",\n"
"            \"data_operations\",\n"
"            \"compatibility\",\n"
"        }),\n"
"        spaces = enterprise_edition(schema.array({\n"
"            items = schema.scalar({\n"
"                type = 'string',\n"
"            }),\n"
"            box_cfg = 'audit_spaces',\n"
"            box_cfg_nondynamic = true,\n"
"            default = box.NULL,\n"
"        })),\n"
"        extract_key = enterprise_edition(schema.scalar({\n"
"            type = 'boolean',\n"
"            box_cfg = 'audit_extract_key',\n"
"            box_cfg_nondynamic = true,\n"
"            default = false,\n"
"        })),\n"
"    })),\n"
"    roles_cfg = schema.map({\n"
"        key = schema.scalar({type = 'string'}),\n"
"        value = schema.scalar({type = 'any'}),\n"
"    }),\n"
"    roles = schema.array({\n"
"        items = schema.scalar({type = 'string'})\n"
"    }),\n"
"    -- Options of the failover coordinator service.\n"
"    --\n"
"    -- TODO: Allow only in the global scope.\n"
"    failover = schema.record({\n"
"        probe_interval = schema.scalar({\n"
"            type = 'number',\n"
"            default = 10,\n"
"        }),\n"
"        connect_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            default = 1,\n"
"        }),\n"
"        call_timeout = schema.scalar({\n"
"            type = 'number',\n"
"            default = 1,\n"
"        }),\n"
"        lease_interval = schema.scalar({\n"
"            type = 'number',\n"
"            default = 30,\n"
"        }),\n"
"        renew_interval = schema.scalar({\n"
"            type = 'number',\n"
"            default = 10,\n"
"        }),\n"
"    }),\n"
"    -- Compatibility options.\n"
"    compat = schema.record({\n"
"        json_escape_forward_slash = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        yaml_pretty_multiline = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        fiber_channel_close_mode = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_cfg_replication_sync_timeout = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        sql_seq_scan_default = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        fiber_slice_default = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_info_cluster_meaning = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        binary_data_decoding = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_tuple_new_vararg = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_session_push_deprecation = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'old',\n"
"        }),\n"
"        sql_priv = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        c_func_iproto_multireturn = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_space_execute_priv = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_tuple_extension = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"        box_space_max = schema.enum({\n"
"            'old',\n"
"            'new',\n"
"        }, {\n"
"            default = 'new',\n"
"        }),\n"
"    }),\n"
"}, {\n"
"    -- This kind of validation cannot be implemented as the\n"
"    -- 'validate' annotation of a particular schema node. There\n"
"    -- are two reasons:\n"
"    --\n"
"    -- * Missed fields are not validated.\n"
"    -- * The outmost instance config record is marked with the\n"
"    --   'scope' annotation (when the instance config is part of\n"
"    --   the cluster config), but this annotation is not easy to\n"
"    --   reach from the 'validate' function of a nested schema\n"
"    --   node.\n"
"    validate = validate_outmost_record,\n"
"}), {\n"
"    methods = {\n"
"        instance_uri = instance_uri,\n"
"        apply_vars = apply_vars,\n"
"        base_dir = base_dir,\n"
"        prepare_file_path = prepare_file_path,\n"
"    },\n"
"})\n"
""
;
