const char config_source_env_lua[] =
"local uri = require('uri')\n"
"local fun = require('fun')\n"
"local schema = require('internal.config.utils.schema')\n"
"local tabulate = require('internal.config.utils.tabulate')\n"
"local instance_config = require('internal.config.instance_config')\n"
"\n"
"local methods = {}\n"
"local mt = {\n"
"    __index = methods,\n"
"}\n"
"\n"
"-- Adjust the given URI list to the form that is accepted by\n"
"-- instance config validators.\n"
"--\n"
"-- The target format is the following.\n"
"--\n"
"-- {\n"
"--     login = <...>,\n"
"--     password = <...>,\n"
"--     uri = <...>,\n"
"--     params = {\n"
"--         transport = 'plain',\n"
"--         <...>,\n"
"--     },\n"
"-- }\n"
"local function normalize_uri_list(src)\n"
"    return fun.iter(uri.parse_many(src)):map(function(u)\n"
"        local res = {\n"
"            login = u.login,\n"
"            password = u.password,\n"
"        }\n"
"        if u.params ~= nil then\n"
"            -- If there is only one value with the given key,\n"
"            -- transform {[k] = {v}} to {[k] = v}.\n"
"            res.params = fun.iter(u.params):map(function(k, v)\n"
"                if type(v) == 'table' and #v == 1 then\n"
"                    return k, v[1]\n"
"                end\n"
"                return k, v\n"
"            end):tomap()\n"
"        end\n"
"        u = table.copy(u)\n"
"        u.login = nil\n"
"        u.password = nil\n"
"        u.params = nil\n"
"        res.uri = uri.format(u)\n"
"        return res\n"
"    end):totable()\n"
"end\n"
"\n"
"local function box_cfg_env_var(box_cfg_option_name)\n"
"    local res = box.internal.cfg.env[box_cfg_option_name]\n"
"\n"
"    if res == nil then\n"
"        return nil\n"
"    end\n"
"\n"
"    -- TT_LISTEN needs a special handling, because\n"
"    -- config's schema is more strict regarding a\n"
"    -- form of an URI than box.cfg's schema.\n"
"    if box_cfg_option_name == 'listen' then\n"
"        return normalize_uri_list(res)\n"
"    end\n"
"\n"
"    return res\n"
"end\n"
"\n"
"function methods._env_var_name(self, path_in_schema)\n"
"    local env_var_name = 'TT_' .. table.concat(path_in_schema, '_'):upper()\n"
"    if self._env_var_suffix ~= nil then\n"
"        return env_var_name .. self._env_var_suffix\n"
"    end\n"
"    return env_var_name\n"
"end\n"
"\n"
"function methods._env_list(self)\n"
"    local rows = {}\n"
"\n"
"    local ce = 'Community Edition'\n"
"    local ee = 'Enterprise Edition'\n"
"\n"
"    -- A header of the table.\n"
"    table.insert(rows, {\n"
"        'ENVIRONMENT VARIABLE',\n"
"        'TYPE',\n"
"        'DEFAULT',\n"
"        'AVAILABILITY',\n"
"    })\n"
"    table.insert(rows, tabulate.SPACER)\n"
"\n"
"    -- Environment variables that duplicate CLI options.\n"
"    table.insert(rows, {\n"
"        'TT_INSTANCE_NAME',\n"
"        'string',\n"
"        'N/A',\n"
"        ce,\n"
"    })\n"
"    table.insert(rows, {\n"
"        'TT_CONFIG',\n"
"        'string',\n"
"        'nil',\n"
"        ce,\n"
"    })\n"
"    table.insert(rows, tabulate.SPACER)\n"
"\n"
"    -- Collect the options from the schema and sort them\n"
"    -- lexicographically.\n"
"    local options = instance_config:pairs():totable()\n"
"    table.sort(options, function(a, b)\n"
"        return table.concat(a.path, '_') < table.concat(b.path, '_')\n"
"    end)\n"
"\n"
"    -- Transform the schema nodes (options) into the environment\n"
"    -- variables description.\n"
"    for _, w in ipairs(options) do\n"
"        local default = w.schema.default\n"
"        if default == nil and type(default) == 'cdata' then\n"
"            default = 'box.NULL'\n"
"        end\n"
"\n"
"        table.insert(rows, {\n"
"            self._env_var_name(self, w.path),\n"
"            w.schema.type,\n"
"            tostring(default),\n"
"            w.schema.computed.annotations.enterprise_edition and ee or ce,\n"
"        })\n"
"    end\n"
"\n"
"    return tabulate.encode(rows)\n"
"end\n"
"\n"
"-- Gather most actual config values.\n"
"function methods.sync(self, _config_module, _iconfig)\n"
"    local values = {}\n"
"\n"
"    -- Handle old TT_* box.cfg()'s environment variables such as\n"
"    -- IPROTO_LISTEN.\n"
"    --\n"
"    -- Here we miss all the options without 'box_cfg' annotation:\n"
"    -- at least 'log', 'replication', 'audit_log' and 'metrics'.\n"
"    --\n"
"    -- They can be handled separately if there is a demand.\n"
"    if self.name == 'env (default)' then\n"
"        for _, w in instance_config:pairs() do\n"
"            if w.schema.box_cfg ~= nil then\n"
"                local value = box_cfg_env_var(w.schema.box_cfg)\n"
"                if value ~= nil then\n"
"                    instance_config:set(values, w.path, value)\n"
"                end\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    for _, w in instance_config:pairs() do\n"
"        local env_var_name = self:_env_var_name(w.path)\n"
"        local raw_value = os.getenv(env_var_name)\n"
"        local value = schema.fromenv(env_var_name, raw_value, w.schema)\n"
"        if value ~= nil then\n"
"            instance_config:set(values, w.path, value)\n"
"        end\n"
"    end\n"
"\n"
"    self._values = values\n"
"end\n"
"\n"
"-- Access the configuration after source:sync().\n"
"function methods.get(self)\n"
"    return self._values\n"
"end\n"
"\n"
"local function new(opts)\n"
"    local opts = opts or {}\n"
"    local env_var_suffix = opts.env_var_suffix\n"
"\n"
"    local name = 'env'\n"
"\n"
"    if env_var_suffix ~= nil then\n"
"        name = ('%s (%s)'):format(name, env_var_suffix)\n"
"        env_var_suffix = '_' .. env_var_suffix:upper()\n"
"    end\n"
"\n"
"    return setmetatable({\n"
"        name = name,\n"
"        type = 'instance',\n"
"\n"
"        _values = {},\n"
"        _env_var_suffix = env_var_suffix,\n"
"    }, mt)\n"
"end\n"
"\n"
"return {\n"
"    new = new,\n"
"}\n"
""
;
