const char config_utils_expression_lua[] =
"-- Evaluate C style expression.\n"
"--\n"
"-- The module implements Pratt's expression parsing algorithm.\n"
"--\n"
"-- The functionality is quite limited, because the module is\n"
"-- internal and is used for one specific task: process tarantool\n"
"-- version constraints in the declarative configuration module.\n"
"-- (`conditional[N].if`).\n"
"--\n"
"-- The module support one data type: version. A value may be\n"
"-- referenced in two ways:\n"
"--\n"
"-- 1. Version literal: 1.2.3 (always three components).\n"
"-- 2. Variable: foo, x23, _foo_bar and so.\n"
"--\n"
"-- The operations are the following.\n"
"--\n"
"-- 1. Logical OR: ||\n"
"-- 2. Logical AND: &&\n"
"-- 3. Compare: >, <, >=, <=, !=, ==\n"
"-- 4. Parentheses: (, )\n"
"\n"
"local expression_lexer = require('internal.config.utils.expression_lexer')\n"
"\n"
"-- NB: bpower is the abbreviation a for binding power.\n"
"local left_bpower_map = {\n"
"    ['||'] = 1,\n"
"    ['&&'] = 3,\n"
"    ['<'] = 5,\n"
"    ['>'] = 5,\n"
"    ['<='] = 5,\n"
"    ['>='] = 5,\n"
"    ['!='] = 5,\n"
"    ['=='] = 5,\n"
"}\n"
"\n"
"local right_bpower_map = {\n"
"    ['||'] = 2,\n"
"    ['&&'] = 4,\n"
"    ['<'] = 6,\n"
"    ['>'] = 6,\n"
"    ['<='] = 6,\n"
"    ['>='] = 6,\n"
"    ['!='] = 6,\n"
"    ['=='] = 6,\n"
"}\n"
"\n"
"local function parse_expr(lexer, min_bpower)\n"
"    local left\n"
"\n"
"    local token = lexer:next()\n"
"    if token == nil then\n"
"        error('Unexpected end of an expression', 0)\n"
"    elseif token.type == 'version_literal' then\n"
"        left = token\n"
"    elseif token.type == 'variable' then\n"
"        left = token\n"
"    elseif token.value == '(' then\n"
"        left = parse_expr(lexer, 0)\n"
"        token = lexer:next()\n"
"        if token == nil then\n"
"            error('Expected \")\", got end of an expression', 0)\n"
"        end\n"
"        -- All the exits from the recursive parse_expr() call with\n"
"        -- zero min_bpower are either with ')' or if there are no\n"
"        -- more tokens.\n"
"        assert(token.value == ')')\n"
"    else\n"
"        error(('Unexpected token %q'):format(token.value), 0)\n"
"    end\n"
"\n"
"    while lexer:peek() ~= nil do\n"
"        local op = lexer:peek()\n"
"        if op.value == ')' then\n"
"            break\n"
"        end\n"
"        if op.type ~= 'operation' then\n"
"            error(('Expected an operation, got %q'):format(op.value), 0)\n"
"        end\n"
"        local left_bpower = left_bpower_map[op.value]\n"
"        if left_bpower < min_bpower then\n"
"            break\n"
"        end\n"
"\n"
"        lexer:next()\n"
"        local right_bpower = right_bpower_map[op.value]\n"
"        local right = parse_expr(lexer, right_bpower)\n"
"\n"
"        left = {\n"
"            type = 'operation',\n"
"            value = op.value,\n"
"            left = left,\n"
"            right = right,\n"
"        }\n"
"    end\n"
"\n"
"    return left\n"
"end\n"
"\n"
"local function parse(s)\n"
"    if type(s) ~= 'string' then\n"
"        error(('Expected a string as an expression, got %s'):format(type(s)), 0)\n"
"    end\n"
"\n"
"    -- A convenience wrapper for the lexer.\n"
"    local lexer = setmetatable({\n"
"        tokens = expression_lexer.split(s),\n"
"        pos = 1,\n"
"    }, {\n"
"        __index = {\n"
"            peek = function(self)\n"
"                return self.tokens[self.pos]\n"
"            end,\n"
"            next = function(self)\n"
"                self.pos = self.pos + 1\n"
"                return self.tokens[self.pos - 1]\n"
"            end,\n"
"        },\n"
"    })\n"
"\n"
"    local ast = parse_expr(lexer, 0)\n"
"    if lexer:peek() ~= nil then\n"
"        error(('Expected end of an expression, got %q'):format(\n"
"            lexer:peek().value), 0)\n"
"    end\n"
"    return ast\n"
"end\n"
"\n"
"local function validate_expr(node, vars)\n"
"    -- luacheck: ignore 542 empty if branch\n"
"    if node.type == 'version_literal' then\n"
"        -- Nothing to validate.\n"
"    elseif node.type == 'variable' then\n"
"        local var = vars[node.value]\n"
"        if type(var) == 'nil' then\n"
"            error(('Unknown variable: %q'):format(node.value), 0)\n"
"        end\n"
"        if type(var) ~= 'string' then\n"
"            error(('Variable %q has type %s, expected string'):format(\n"
"                node.value, type(var)), 0)\n"
"        end\n"
"        if var == '' then\n"
"            error(('Expected a version in variable %q, got an empty ' ..\n"
"                'string'):format(node.value), 0)\n"
"        end\n"
"        local components = var:split('.')\n"
"        for _, v in ipairs(components) do\n"
"            if not v:match('^[0-9]+$') then\n"
"                error(('Variable %q is not a version string'):format(\n"
"                    node.value), 0)\n"
"            end\n"
"        end\n"
"        if #components ~= 3 then\n"
"            error(('Expected a three component version in variable %q, got ' ..\n"
"                '%d components'):format(node.value, #components), 0)\n"
"        end\n"
"    elseif node.type == 'operation' then\n"
"        local lbool = node.left.type == 'operation'\n"
"        local rbool = node.right.type == 'operation'\n"
"\n"
"        local requires_boolean = {\n"
"            ['||'] = true,\n"
"            ['&&'] = true,\n"
"        }\n"
"        if requires_boolean[node.value] and not (lbool and rbool) then\n"
"            error('A logical operator (&& or ||) accepts only boolean ' ..\n"
"                'expressions as arguments', 0)\n"
"        end\n"
"\n"
"        local requires_version_values = {\n"
"            ['<'] = true,\n"
"            ['>'] = true,\n"
"            ['<='] = true,\n"
"            ['>='] = true,\n"
"            ['!='] = true,\n"
"            ['=='] = true,\n"
"        }\n"
"        if requires_version_values[node.value] and (lbool or rbool) then\n"
"            error('A comparison operator (<, >, <=, >=, !=, ==) requires ' ..\n"
"                'version literals or variables as arguments', 0)\n"
"        end\n"
"\n"
"        validate_expr(node.left, vars)\n"
"        validate_expr(node.right, vars)\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"local function validate(node, vars)\n"
"    if node.type ~= 'operation' then\n"
"        error(('An expression should be a predicate, got %s'):format(\n"
"            node.type), 0)\n"
"    end\n"
"    validate_expr(node, vars)\n"
"end\n"
"\n"
"local function compare_version(a, b)\n"
"    local av = a:split('.')\n"
"    local bv = b:split('.')\n"
"    assert(#av == 3)\n"
"    assert(#bv == 3)\n"
"    for i = 1, 3 do\n"
"        local ac = tonumber(av[i])\n"
"        local bc = tonumber(bv[i])\n"
"        if ac ~= bc then\n"
"            return ac - bc\n"
"        end\n"
"    end\n"
"    return 0\n"
"end\n"
"\n"
"local operations = {\n"
"    ['||'] = function(a, b)\n"
"        return a or b\n"
"    end,\n"
"    ['&&'] = function(a, b)\n"
"        return a and b\n"
"    end,\n"
"    ['<'] = function(a, b)\n"
"        return compare_version(a, b) < 0\n"
"    end,\n"
"    ['>'] = function(a, b)\n"
"        return compare_version(a, b) > 0\n"
"    end,\n"
"    ['<='] = function(a, b)\n"
"        return compare_version(a, b) <= 0\n"
"    end,\n"
"    ['>='] = function(a, b)\n"
"        return compare_version(a, b) >= 0\n"
"    end,\n"
"    ['=='] = function(a, b)\n"
"        return compare_version(a, b) == 0\n"
"    end,\n"
"    ['!='] = function(a, b)\n"
"        return compare_version(a, b) ~= 0\n"
"    end,\n"
"}\n"
"\n"
"local function evaluate(node, vars)\n"
"    if node.type == 'version_literal' then\n"
"        return node.value\n"
"    elseif node.type == 'variable' then\n"
"        return vars[node.value]\n"
"    elseif node.type == 'operation' then\n"
"        local left = evaluate(node.left, vars)\n"
"        local op = operations[node.value]\n"
"        local right = evaluate(node.right, vars)\n"
"        return op(left, right)\n"
"    else\n"
"        assert(false)\n"
"    end\n"
"end\n"
"\n"
"local function eval(s, vars)\n"
"    local ast = parse(s)\n"
"    validate(ast, vars)\n"
"    return evaluate(ast, vars)\n"
"end\n"
"\n"
"return {\n"
"    parse = parse,\n"
"    validate = validate,\n"
"    evaluate = evaluate,\n"
"    eval = eval,\n"
"}\n"
""
;
