const char config_utils_snapshot_lua[] =
"local fio = require('fio')\n"
"local xlog = require('xlog')\n"
"local instance_config = require('internal.config.instance_config')\n"
"\n"
"local snapshot_path = nil\n"
"\n"
"-- Determine where snapshot should reside based on the given\n"
"-- configuration.\n"
"--\n"
"-- To be called before first box.cfg().\n"
"local function effective_snapshot_dir(iconfig)\n"
"    assert(iconfig ~= nil)\n"
"    -- The snapshot directory has a default value in the schema\n"
"    -- (it is a string). So, it can't be nil or box.NULL.\n"
"    local snap_dir = instance_config:get(iconfig, 'snapshot.dir')\n"
"    assert(snap_dir ~= nil)\n"
"\n"
"    -- If the path is absolute, just return it.\n"
"    --\n"
"    -- This check is necessary due to fio.pathjoin() peculiars,\n"
"    -- see gh-8816.\n"
"    if snap_dir:startswith('/') then\n"
"        return snap_dir\n"
"    end\n"
"\n"
"    -- We assume that the startup working directory is the current\n"
"    -- working directory. IOW, that this function is called before\n"
"    -- first box.cfg() call. Let's verify it.\n"
"    assert(type(box.cfg) == 'function')\n"
"\n"
"    -- If the snapshot directory is not absolute, it is relative\n"
"    -- to the working directory.\n"
"    --\n"
"    -- Determine an absolute path to the configured working\n"
"    -- directory considering that it may be relative to the\n"
"    -- working directory at the startup moment.\n"
"    local work_dir = instance_config:get(iconfig, 'process.work_dir')\n"
"    if work_dir == nil then\n"
"        work_dir = '.'\n"
"    end\n"
"    work_dir = fio.abspath(work_dir)\n"
"\n"
"    -- Now we know the absolute path to the configured working\n"
"    -- directory. Let's determine the snapshot directory path.\n"
"    return fio.abspath(fio.pathjoin(work_dir, snap_dir))\n"
"end\n"
"\n"
"-- Determine whether the instance will be recovered from an existing\n"
"-- snapshot and return its path. Should be called before box.cfg.\n"
"--\n"
"-- To be called before first box.cfg().\n"
"local function get_snapshot_path(iconfig)\n"
"    assert(type(box.cfg) == 'function')\n"
"    if snapshot_path == nil then\n"
"        local snap_dir = effective_snapshot_dir(iconfig)\n"
"        local glob = fio.glob(fio.pathjoin(snap_dir, '*.snap'))\n"
"        if #glob > 0 then\n"
"            table.sort(glob)\n"
"            snapshot_path = glob[#glob]\n"
"        end\n"
"    end\n"
"\n"
"    return snapshot_path\n"
"end\n"
"\n"
"-- Read snap file and return a map of saved UUIDs and names for\n"
"-- all instances and for the current replicaset.\n"
"local function get_snapshot_names(snap_path)\n"
"    local peers = {}\n"
"    local instance_uuid = xlog.meta(snap_path).instance_uuid\n"
"    local instance_name, replicaset_name, replicaset_uuid\n"
"    for _, row in xlog.pairs(snap_path) do\n"
"        local body = row.BODY\n"
"        if not body.space_id then\n"
"            goto continue\n"
"        end\n"
"\n"
"        if body.space_id > box.schema.CLUSTER_ID then\n"
"            -- No sense in scanning after _cluster.\n"
"            break\n"
"        end\n"
"\n"
"        if body.space_id == box.schema.SCHEMA_ID then\n"
"            if body.tuple[1] == 'replicaset_uuid' or\n"
"               body.tuple[1] == 'cluster' then\n"
"                replicaset_uuid = body.tuple[2]\n"
"            elseif body.tuple[1] == 'replicaset_name' then\n"
"                replicaset_name = body.tuple[2]\n"
"            end\n"
"        elseif body.space_id == box.schema.CLUSTER_ID then\n"
"            if body.tuple[2] == instance_uuid then\n"
"                instance_name = body.tuple[3]\n"
"            end\n"
"\n"
"            if body.tuple[3] ~= nil then\n"
"                peers[body.tuple[3]] = body.tuple[2]\n"
"            end\n"
"        end\n"
"        ::continue::\n"
"    end\n"
"\n"
"    return {\n"
"        replicaset_name = replicaset_name,\n"
"        replicaset_uuid = replicaset_uuid,\n"
"        instance_name = instance_name,\n"
"        instance_uuid = instance_uuid,\n"
"        peers = peers,\n"
"    }\n"
"end\n"
"\n"
"local function get_snapshot_schema_version(snap_path)\n"
"    for _, row in xlog.pairs(snap_path) do\n"
"        local body = row.BODY\n"
"        if not body.space_id then\n"
"            goto continue\n"
"        end\n"
"\n"
"        if body.space_id > box.schema.SCHEMA_ID then\n"
"            break\n"
"        end\n"
"\n"
"        if body.space_id == box.schema.SCHEMA_ID then\n"
"            if body.tuple[1] == 'version' then\n"
"                return body.tuple\n"
"            end\n"
"        end\n"
"        ::continue::\n"
"    end\n"
"\n"
"    assert(false)\n"
"end\n"
"\n"
"return {\n"
"    get_path = get_snapshot_path,\n"
"    get_names = get_snapshot_names,\n"
"    get_schema_version = get_snapshot_schema_version,\n"
"}\n"
""
;
