const char config_utils_tabulate_lua[] =
"-- Very simple pretty-printer of tabular data.\n"
"--\n"
"-- Example:\n"
"--\n"
"-- tabulate.encode({\n"
"--     {'a', 'b', 'c'},\n"
"--     tabulate.SPACER,\n"
"--     {'d', 'e', 'f'},\n"
"--     {'g', 'h', 'i'},\n"
"-- })\n"
"--\n"
"-- ->\n"
"--\n"
"-- | a | b | c |\n"
"-- | - | - | - |\n"
"-- | d | e | f |\n"
"-- | g | h | i |\n"
"\n"
"local SPACER = {}\n"
"\n"
"-- Format data as a table.\n"
"--\n"
"-- Accepts an array of rows. Each row in an array of values. Each\n"
"-- value is a string.\n"
"local function encode(rows)\n"
"    -- Calculate column widths and columns amount.\n"
"    local column_widths = {}\n"
"    for _i, row in ipairs(rows) do\n"
"        for j, v in ipairs(row) do\n"
"            assert(type(v) == 'string')\n"
"            column_widths[j] = math.max(column_widths[j] or 0, #v)\n"
"        end\n"
"    end\n"
"    local column_count = #column_widths\n"
"\n"
"    -- Use a table as a string buffer.\n"
"    local acc = {}\n"
"\n"
"    -- Add all the values into the accumulator with proper spacing\n"
"    -- around and appropriate separators.\n"
"    for _i, row in ipairs(rows) do\n"
"        if row == SPACER then\n"
"            for j = 1, column_count do\n"
"                local width = column_widths[j]\n"
"                table.insert(acc, '| ')\n"
"                table.insert(acc, ('-'):rep(width))\n"
"                table.insert(acc, ' ')\n"
"            end\n"
"            table.insert(acc, '|\\n')\n"
"        else\n"
"            for j = 1, column_count do\n"
"                assert(row[j] ~= nil)\n"
"                local width = column_widths[j]\n"
"                table.insert(acc, '| ')\n"
"                table.insert(acc, row[j]:ljust(width))\n"
"                table.insert(acc, ' ')\n"
"            end\n"
"            table.insert(acc, '|\\n')\n"
"        end\n"
"    end\n"
"\n"
"    return table.concat(acc)\n"
"end\n"
"\n"
"return {\n"
"    SPACER = SPACER,\n"
"    encode = encode,\n"
"}\n"
""
;
