const char loaders_lua[] =
"local minifio = require('internal.minifio')\n"
"\n"
"local soext = (jit.os == \"OSX\" and \"dylib\" or \"so\")\n"
"\n"
"local ROCKS_LIB_PATH = '.rocks/lib/tarantool'\n"
"local ROCKS_LUA_PATH = '.rocks/share/tarantool'\n"
"local LIB_TEMPLATES = { '\?.'..soext }\n"
"local LUA_TEMPLATES = { '\?.lua', '\?/init.lua' }\n"
"local ROCKS_LIB_TEMPLATES = {\n"
"    ROCKS_LIB_PATH .. '/\?.'..soext,\n"
"}\n"
"local ROCKS_LUA_TEMPLATES = {\n"
"    ROCKS_LUA_PATH .. '/\?.lua',\n"
"    ROCKS_LUA_PATH .. '/\?/init.lua',\n"
"}\n"
"\n"
"local package_searchroot\n"
"\n"
"local function searchroot()\n"
"    return package_searchroot or minifio.cwd()\n"
"end\n"
"\n"
"local function setsearchroot(path)\n"
"    if not path then\n"
"        -- Here we need to get this function caller's sourcedir.\n"
"        path = debug.sourcedir(3)\n"
"    elseif path == box.NULL then\n"
"        path = nil\n"
"    else\n"
"        assert(type(path) == 'string', 'Search root must be a string')\n"
"    end\n"
"    package_searchroot = path and minifio.abspath(path)\n"
"end\n"
"\n"
"local function mksymname(name)\n"
"    local mark = string.find(name, \"-\")\n"
"    if mark then name = string.sub(name, mark + 1) end\n"
"    return \"luaopen_\" .. string.gsub(name, \"%.\", \"_\")\n"
"end\n"
"\n"
"local function load_lib(file, name)\n"
"    return package.loadlib(file, mksymname(name))\n"
"end\n"
"\n"
"local function load_lua(file)\n"
"    return loadfile(file)\n"
"end\n"
"\n"
"local function traverse_path(path)\n"
"    path = minifio.abspath(path)\n"
"    local paths = { path }\n"
"\n"
"    while path ~= '/' do\n"
"        path = minifio.dirname(path)\n"
"        table.insert(paths, path)\n"
"    end\n"
"\n"
"    return paths\n"
"end\n"
"\n"
"-- Generate a path builder function, which yields the set of the\n"
"-- paths to be searched through for the desired module.\n"
"--\n"
"-- @param basepath_fn function which returns a base path for the\n"
"--     resulting path set\n"
"-- @param templates table with lua search templates\n"
"-- @param need_traverse bool flag which tells search function to\n"
"--     build multiple paths by expanding base path up to the\n"
"--     root ('/')\n"
"-- @return a path builder function which builds and yields the\n"
"--     resulting path set\n"
"local function gen_path_builder(basepath_fn, templates, need_traverse)\n"
"    assert(type(basepath_fn) == 'function', 'basepath_fn must be a function')\n"
"    assert(type(templates) == 'table', 'templates must be a table')\n"
"\n"
"    return function()\n"
"        local base = basepath_fn() or '.'\n"
"        local dirs = need_traverse and traverse_path(base) or { base }\n"
"\n"
"        local searchpaths = {}\n"
"\n"
"        for _, dir in ipairs(dirs) do\n"
"            for _, template in pairs(templates) do\n"
"                table.insert(searchpaths, minifio.pathjoin(dir, template))\n"
"            end\n"
"        end\n"
"\n"
"        return table.concat(searchpaths, ';')\n"
"    end\n"
"end\n"
"\n"
"local path = {\n"
"    package = function() return package.path end,\n"
"    cwd = {\n"
"        dot = gen_path_builder(searchroot, LUA_TEMPLATES),\n"
"        rocks = gen_path_builder(searchroot, ROCKS_LUA_TEMPLATES, true),\n"
"    }\n"
"}\n"
"\n"
"local cpath = {\n"
"    package = function() return package.cpath end,\n"
"    cwd = {\n"
"        dot = gen_path_builder(searchroot, LIB_TEMPLATES),\n"
"        rocks = gen_path_builder(searchroot, ROCKS_LIB_TEMPLATES, true),\n"
"    }\n"
"}\n"
"\n"
"local function gen_file_searcher(loader, pathogen)\n"
"    assert(type(loader) == 'function', '<loader> must be a function')\n"
"    assert(type(pathogen) == 'function', '<pathogen> must be a function')\n"
"\n"
"    return function(name)\n"
"        local path, errmsg = package.searchpath(name, pathogen())\n"
"        if not path then\n"
"            return errmsg\n"
"        end\n"
"        return loader, path\n"
"    end\n"
"end\n"
"\n"
"local function gen_croot_searcher(searcher)\n"
"    assert(type(searcher) == 'function', '<searcher> must be a function')\n"
"\n"
"    return function(name)\n"
"        local path = name:match('[^.]+')\n"
"        local loader, data = searcher(path)\n"
"        -- XXX: <loader> (i.e. the first return value) contains\n"
"        -- error message in this case. Just propagate it to the\n"
"        -- <require> frame...\n"
"        if not data then\n"
"           return loader\n"
"        end\n"
"        -- XXX: ... Otherwise, there is a croot module found.\n"
"        -- Yield the C module loader with the found croot module.\n"
"        assert(loader == load_lib, 'Invalid searcher provided')\n"
"        return loader, data\n"
"    end\n"
"end\n"
"\n"
"local function yield_builtin_loader(loader, sentinel)\n"
"    return function(data)\n"
"        if data == sentinel then\n"
"            return loader\n"
"        end\n"
"        return ('%s loader is used to load %s module'):format(sentinel, data)\n"
"    end, sentinel\n"
"\n"
"end\n"
"\n"
"-- Lua table with \"package searchers\".\n"
"local searchers = debug.getregistry()._TARANTOOL_PACKAGE_SEARCHERS\n"
"\n"
"rawset(searchers, 'preload', function(name)\n"
"    assert(type(package.preload) == 'table',\n"
"           \"'package.preload' must be a table\")\n"
"    local loader = package.preload[name]\n"
"    if loader ~= nil then\n"
"        return yield_builtin_loader(loader, ':preload:')\n"
"    end\n"
"    return (\"\\n\\tno field package.preload['%s']\"):format(name)\n"
"end)\n"
"\n"
"-- Tarantool's builtin modules.\n"
"--\n"
"-- Similar to _LOADED (package.loaded).\n"
"local builtin_modules = debug.getregistry()._TARANTOOL_BUILTIN\n"
"\n"
"-- Searcher for builtin modules.\n"
"rawset(searchers, 'builtin', function(name)\n"
"    local module = builtin_modules[name]\n"
"    if module ~= nil then\n"
"        return yield_builtin_loader(function(_) return module end, ':builtin:')\n"
"    end\n"
"    return (\"\\n\\tno field loaders.builtin['%s']\"):format(name)\n"
"end)\n"
"\n"
"rawset(searchers, 'path.package', gen_file_searcher(load_lua, path.package))\n"
"rawset(searchers, 'cpath.package', gen_file_searcher(load_lib, cpath.package))\n"
"rawset(searchers, 'path.cwd.dot', gen_file_searcher(load_lua, path.cwd.dot))\n"
"rawset(searchers, 'cpath.cwd.dot', gen_file_searcher(load_lib, cpath.cwd.dot))\n"
"rawset(searchers, 'path.cwd.rocks', gen_file_searcher(load_lua, path.cwd.rocks))\n"
"rawset(searchers, 'cpath.cwd.rocks', gen_file_searcher(load_lib, cpath.cwd.rocks))\n"
"\n"
"-- FIXME: There are two problems with croot problems in Tarantool:\n"
"--   1. Since Tarantool uses original croot loader from LuaJIT, it\n"
"--      ignores all paths except ones specified in <package.path>.\n"
"--      However, there might be some croot-like modules in the\n"
"--      Tarantool-specific cpaths (cwd or application root).\n"
"--      It looks like croot machinery is not widely used, if this\n"
"--      bug bothers nobody, but it's still need to be fixed.\n"
"--   2. Similar situation relates to the overriding mechanism,\n"
"--      that was recently introduced: there is no \"prefixed\" croot\n"
"--      loader and all libraries using croot loading machinery are\n"
"--      ignored for the overridden paths as a result.\n"
"rawset(searchers, 'croot.cwd', gen_croot_searcher(searchers['cpath.package']))\n"
"\n"
"-- Indices in <searchers> table must represent the order of the\n"
"-- functions in <package.loaders> below.\n"
"rawset(searchers, 2, searchers['path.cwd.dot'])\n"
"rawset(searchers, 3, searchers['cpath.cwd.dot'])\n"
"rawset(searchers, 4, searchers['path.cwd.rocks'])\n"
"rawset(searchers, 5, searchers['cpath.cwd.rocks'])\n"
"rawset(searchers, 6, searchers['path.package'])\n"
"rawset(searchers, 7, searchers['cpath.package'])\n"
"rawset(searchers, 8, searchers['croot.cwd'])\n"
"\n"
"-- Compose a loader function from options. The function accepts\n"
"-- the table with the searchers and the index, so the user can\n"
"-- override the particular searcher.\n"
"--\n"
"-- @param searchers table with the searcher function to be used\n"
"--     to search a file from path template\n"
"-- @param index in searchers table to be used in the loader to be\n"
"--     generated\n"
"-- @return function a loader, which first search for the file and\n"
"--     then loads it\n"
"local function gen_legacy_loader(searchers, index)\n"
"    assert(type(searchers) == 'table', '<searchers> must be defined')\n"
"    assert(type(searchers[index]) == 'function',\n"
"           'searchers[index] must be a function')\n"
"\n"
"    return function(name)\n"
"        if not name then\n"
"            return \"empty name of module\"\n"
"        end\n"
"        local loader, data = searchers[index](name)\n"
"        -- XXX: <loader> (i.e. the first return value) contains\n"
"        -- error message in this case. Just propagate it to the\n"
"        -- <require> frame...\n"
"        if not data then\n"
"           return loader\n"
"        end\n"
"        -- XXX: ... Otherwise, this is a valid module loader.\n"
"        -- Load the given <data> and return the result.\n"
"        local loaded, err = loader(data, name)\n"
"        local message = (\"error loading module '%s' from file '%s':\\n\\t%s\")\n"
"                        :format(name, data, err)\n"
"        return assert(loaded, message)\n"
"    end\n"
"end\n"
"\n"
"-- XXX: Though <package.search> looks like a handy public\n"
"-- function, it's only the helper used for searching Tarantool\n"
"-- C modules. Hence, despite it uses the same searchers as loaders\n"
"-- for Lua modules do, it follows its own semantics ignoring\n"
"-- preload, override, builtin modules, and application specific\n"
"-- locations.\n"
"-- FIXME: Lua paths look really odd in the list, since the\n"
"-- function is used to load only C modules. However, these paths\n"
"-- are left to avoid breaking change.\n"
"local search_funcs = {\n"
"    searchers['path.cwd.dot'],\n"
"    searchers['cpath.cwd.dot'],\n"
"    searchers['path.cwd.rocks'],\n"
"    searchers['cpath.cwd.rocks'],\n"
"    searchers['path.package'],\n"
"    searchers['cpath.package'],\n"
"}\n"
"\n"
"local function search(name)\n"
"    if not name then\n"
"        return \"empty name of module\"\n"
"    end\n"
"    for _, searcher in ipairs(search_funcs) do\n"
"        local loader, file = searcher(name)\n"
"        if type(loader) == 'function' then\n"
"            return file\n"
"        end\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"-- Accept a searcher and return a searcher, which search by a\n"
"-- prefixed module name.\n"
"--\n"
"-- The module receives the original (unprefixed) module name in\n"
"-- the argument (three dots).\n"
"local function prefix_searcher(prefix, subsearcher)\n"
"    return function(name)\n"
"        local prefixed_name = prefix .. '.' .. name\n"
"        -- On success the return value is a function, which\n"
"        -- implements loading mechanism for particular file.\n"
"        -- If the file is loaded successfully, the loader yields\n"
"        -- module initialization function that is called by\n"
"        -- <require> with one argument: the module name (it can be\n"
"        -- received in the module using three dots). Since\n"
"        -- <require> knows nothing about our prefixing, it passes\n"
"        -- the original name there.\n"
"        --\n"
"        -- It is expected behavior in our case. The prefixed\n"
"        -- loaders are added to enable extra search paths: like\n"
"        -- we would add more package.{path,cpath} entries. It\n"
"        -- shouldn't change the string passed to the module's\n"
"        -- initialization code.\n"
"        return subsearcher(prefixed_name)\n"
"    end\n"
"end\n"
"\n"
"-- Accept a searcher and return the same searcher, but enabled\n"
"-- only when given condition (a function return value) is true.\n"
"local function conditional_searcher(subsearcher, onoff)\n"
"    assert(type(onoff) == 'function')\n"
"    return function(name)\n"
"        if onoff(name) then\n"
"            return subsearcher(name)\n"
"        end\n"
"        -- It is okay to return nothing, <require> ignores it.\n"
"    end\n"
"end\n"
"\n"
"-- Accept an array of searchers and return a searcher, whose\n"
"-- effect is equivalent to calling the searchers in a row.\n"
"local function chain_searchers(subsearchers)\n"
"    return function(name)\n"
"        -- Error accumulator.\n"
"        local err = ''\n"
"\n"
"        for _, searcher in ipairs(subsearchers) do\n"
"            local loader, data = searcher(name)\n"
"            -- Whether the module found\? Let's return it.\n"
"            --\n"
"            -- <loader> is a function, which implements loading\n"
"            -- routine for particular <data> module.\n"
"            if type(loader) == 'function' then\n"
"                return loader, data\n"
"            end\n"
"            -- If the module is not found and the searcher\n"
"            -- function returns an error, add the error into the\n"
"            -- accumulator.\n"
"            if type(loader) == 'string' then\n"
"                err = err .. loader\n"
"            end\n"
"            -- Ignore any other return value: require() does the\n"
"            -- same.\n"
"        end\n"
"\n"
"        return err\n"
"    end\n"
"end\n"
"\n"
"\n"
"-- Search for modules next to the main script.\n"
"--\n"
"-- If the script is not provided (script == nil) or provided as\n"
"-- stdin (script == '-'), there is no script directory, so nothing\n"
"-- to do here.\n"
"local script = minifio.script()\n"
"if script ~= nil and script ~= '-' then\n"
"    -- It is important to obtain the directory at initialization,\n"
"    -- before any cwd change may occur. The script path may be\n"
"    -- passed as relative to the current directory.\n"
"    local script_dir = minifio.dirname(minifio.abspath(script))\n"
"    local function script_dir_fn()\n"
"        return script_dir\n"
"    end\n"
"\n"
"    -- Search non-recursively, only next to the script.\n"
"    path.app = {\n"
"        dot = gen_path_builder(script_dir_fn, LUA_TEMPLATES),\n"
"        rocks = gen_path_builder(script_dir_fn, ROCKS_LUA_TEMPLATES),\n"
"    }\n"
"    cpath.app = {\n"
"        dot = gen_path_builder(script_dir_fn, LIB_TEMPLATES),\n"
"        rocks = gen_path_builder(script_dir_fn, ROCKS_LIB_TEMPLATES)\n"
"    }\n"
"\n"
"    rawset(searchers, 'path.app.dot', gen_file_searcher(load_lua, path.app.dot))\n"
"    rawset(searchers, 'cpath.app.dot', gen_file_searcher(load_lib, cpath.app.dot))\n"
"    rawset(searchers, 'path.app.rocks', gen_file_searcher(load_lua, path.app.rocks))\n"
"    rawset(searchers, 'cpath.app.rocks', gen_file_searcher(load_lib, cpath.app.rocks))\n"
"\n"
"    -- Mix the script directory searchers into corresponding\n"
"    -- searchroot based searchers. It allows to avoid changing\n"
"    -- ordinals of the searchers (and hence package.loaders) and\n"
"    -- also makes the override loader search here.\n"
"    --\n"
"    -- We can just add more paths to package.path/package.cpath,\n"
"    -- but:\n"
"    --\n"
"    -- * Search for override modules is implemented as a loader.\n"
"    -- * Search inside .rocks directory is implemented as a\n"
"    --   separate searchers.\n"
"    --\n"
"    -- And it is simpler to wrap this logic rather than repeat.\n"
"    -- It is possible (and maybe even desirable) to reimplement\n"
"    -- all the searchers logic as paths generation, but we should\n"
"    -- do that for all the logic at once.\n"
"    rawset(searchers, 2, chain_searchers({\n"
"        searchers[2],\n"
"        searchers['path.app.dot'],\n"
"    }))\n"
"    rawset(searchers, 3, chain_searchers({\n"
"        searchers[3],\n"
"        searchers['cpath.app.dot'],\n"
"    }))\n"
"    rawset(searchers, 4, chain_searchers({\n"
"        searchers[4],\n"
"        searchers['path.app.rocks'],\n"
"    }))\n"
"    rawset(searchers, 5, chain_searchers({\n"
"        searchers[5],\n"
"        searchers['cpath.app.rocks'],\n"
"    }))\n"
"end\n"
"\n"
"local function getenv_boolean(varname, default)\n"
"    local envvar = os.getenv(varname)\n"
"\n"
"    -- If unset or empty, use the default.\n"
"    if envvar == nil or envvar == '' then\n"
"        return default\n"
"    end\n"
"\n"
"    -- Explicitly enabled or disabled.\n"
"    --\n"
"    -- Accept false/true case insensitively.\n"
"    --\n"
"    -- Accept 0/1 as boolean values.\n"
"    if envvar:lower() == 'false' or envvar == '0' then\n"
"        return false\n"
"    end\n"
"    if envvar:lower() == 'true' or envvar == '1' then\n"
"        return true\n"
"    end\n"
"\n"
"    -- Can't parse the value, let's use the default.\n"
"    return default\n"
"end\n"
"\n"
"-- true/false if explicitly enabled or disabled, nil otherwise.\n"
"local override_searcher_is_enabled\n"
"\n"
"-- Whether the override loader is enabled.\n"
"local function override_searcher_onoff(_name)\n"
"    -- Follow the switch if it is explicitly enabled or disabled.\n"
"    if override_searcher_is_enabled ~= nil then\n"
"        return override_searcher_is_enabled\n"
"    end\n"
"\n"
"    -- Follow the environment variable otherwise.\n"
"    return getenv_boolean('TT_OVERRIDE_BUILTIN', true)\n"
"end\n"
"\n"
"rawset(searchers, 'override', conditional_searcher(chain_searchers({\n"
"    prefix_searcher('override', searchers[2]),\n"
"    prefix_searcher('override', searchers[3]),\n"
"    prefix_searcher('override', searchers[4]),\n"
"    prefix_searcher('override', searchers[5]),\n"
"    prefix_searcher('override', searchers[6]),\n"
"    prefix_searcher('override', searchers[7]),\n"
"}), override_searcher_onoff))\n"
"\n"
"-- Join three searchers:\n"
"--\n"
"-- - Searcher for preload LuaJIT modules.\n"
"-- - Searcher for override.<module_name> module. It is necessary\n"
"--   for overriding built-in modules.\n"
"-- - Searcher for a built-in module (compiled into tarantool's\n"
"--   executable).\n"
"--\n"
"-- Those three searchers are mixed into the first searcher to\n"
"-- don't change ordinals of the searchers (and hence, loaders)\n"
"-- 2-8. It is possible that someone has a logic based on those\n"
"-- loader positions.\n"
"rawset(searchers, 1, chain_searchers({\n"
"    searchers['preload'],\n"
"    searchers['override'],\n"
"    searchers['builtin'],\n"
"}))\n"
"\n"
"-- XXX: The default LuaJIT loaders are:\n"
"-- [1]: preload\n"
"-- [2]: package.path\n"
"-- [3]: package.cpath\n"
"-- [4]: croot\n"
"--\n"
"-- Tarantool loaders just maps the <searchers> table:\n"
"-- [1]: preload -> override -> builtin\n"
"-- [2]: path.cwd.dot -> [path.app.dot]\n"
"-- [3]: cpath.cwd.dot -> [cpath.app.dot]\n"
"-- [4]: path.cwd.rocks -> [path.app.rocks]\n"
"-- [5]: cpath.cwd.rocks -> [cpath.app.rocks]\n"
"-- [6]: package.path\n"
"-- [7]: package.cpath\n"
"-- [8]: croot\n"
"for index, _ in ipairs(searchers) do\n"
"    rawset(package.loaders, index, gen_legacy_loader(searchers, index))\n"
"end\n"
"\n"
"rawset(package, \"search\", search)\n"
"rawset(package, \"searchroot\", searchroot)\n"
"rawset(package, \"setsearchroot\", setsearchroot)\n"
"\n"
"local no_package_loaded = {}\n"
"local raw_require = require\n"
"\n"
"-- Allow an override module to refuse caching in package.loaded.\n"
"--\n"
"-- It may be useful to return a built-in module into the platform,\n"
"-- but the override module into application's code.\n"
"_G.require = function(modname)\n"
"    if no_package_loaded[modname] then\n"
"        no_package_loaded[modname] = nil\n"
"        package.loaded[modname] = nil\n"
"    end\n"
"    -- NB: This call is a tail call and it is important to some\n"
"    -- extent. At least luajit's test suites expect certain error\n"
"    -- messages from `require()` calls and from the `-l` option.\n"
"    --\n"
"    -- A non-tail call changes the filename at the beginning of\n"
"    -- the error message.\n"
"    return raw_require(modname)\n"
"end\n"
"\n"
"return {\n"
"    ROCKS_LIB_PATH = ROCKS_LIB_PATH,\n"
"    ROCKS_LUA_PATH = ROCKS_LUA_PATH,\n"
"    builtin = builtin_modules,\n"
"    searchers = searchers,\n"
"    override_builtin_enable = function()\n"
"        override_searcher_is_enabled = true\n"
"    end,\n"
"    override_builtin_disable = function()\n"
"        override_searcher_is_enabled = false\n"
"    end,\n"
"    -- It is `true` during tarantool initialization, but once all\n"
"    -- the built-in modules are ready, will be set to `nil`.\n"
"    initializing = true,\n"
"    raw_require = raw_require,\n"
"    -- Add a module name here to ignore package.loaded[modname] at\n"
"    -- next require(). The flag is dropped at the require() call.\n"
"    --\n"
"    -- This is intended to be used from inside a module to refuse\n"
"    -- caching in package.loaded. Any other usage has unspecified\n"
"    -- behavior (just manipulate package.loaded directly if you\n"
"    -- need it).\n"
"    --\n"
"    -- Usage: loaders.no_package_loaded[modname] = true\n"
"    no_package_loaded = no_package_loaded,\n"
"}\n"
""
;
